/** \file appconfig.c	Configuration module.
 * \author Lionel ULMER
 * \author Darryl Luff
 * \ingroup config
 *
 * The appconfig module manages the gtksql configuration files.  At present
 * there are three different types of config files:
 * - The resource file ${HOME}/.gtksqlrc
 * - The config file ${HOME}/.gtksql/gtksql.conf
 * - Script files.
 *
 * The resource file is in the standard gtk resource format.
 * gtksql.conf is loaded every time the application runs.
 * Other scripts are loaded either from the command line, or from menus inside the application.
 */
/* GtkSQL -- an interactive graphical query tool for PostgreSQL
 * Copyright (C) 1998-2006  Lionel ULMER, Darryl Luff
 * $Id: appconfig.c,v 1.54 2006/03/17 13:57:32 djluff Exp $
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Some config items are stored in a gtk resource file int he standard gtk format.
 * But gtk doesnt seem to allow you to put arbitrary things in there. So the 
 * remainder of things are stored in a simple text config file.
 *
 * This module handles the differences transparently, so the application doesn't need
 * to know about these seperate files.
 */

#include "../config.h"

#include <errno.h>
#include <regex.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <unistd.h>

#include "appconfig.h"
#include "guiapi.h"
#include "scriptapi.h"
#include "utils.h"

/** Maximum length of a config line	*/
#define CFG_LINE_SIZE	1024

/** Directory name for GtkSQL config files.	*/
#define CONFIG_DIRNAME ".gtksql"

/** Name of the main configuration file.	*/
#define CONFIG_FILENAME	"gtksql.conf"

/** Contact URL.	*/
#define CONTACT	"http://gtksql.sourceforge.net/"

/* Config strings	*/
/** The workspace.autosave string for the config file.	*/
#define WORKSPACE_AUTOSAVE	"workspace.autosave"

/* Default settings	*/
/** Default autosave setting.	*/
#define DEF_WORKSPACE_AUTOSAVE	1

/** Data structure containing config items.	*/
typedef struct _config {
	/** Whether to automatically save changed workspace	*/
	int autosave;

	/** The last accessed directory.	*/
	char *currentDir;

	/** The path to the configuration directory.	*/
	char *dirname;

	/** The actual file.	*/
	FILE *file;
} Config;

/** Usage string to be displayed to the user.	*/
#define USAGE	"gtksql [-h] \n\n\t-h              Show usage help\n"

#define ARGSPEC	"h"

/** Global app config.	*/
Config cfg;

/** Filename of config file.	*/
char *cfgname = NULL;

static void cfg_save_int(const char *key, int value);

/** \brief Get the autosave status.	*/
int cfg_get_autosave(void)
{
	return cfg.autosave;
}

/** Show a usage string.
 * \todo Build the driver list from the list structure instead of it 
 * being static. */
void cfg_show_banner(void)
{
	printf("%s Version %s\n%s\n", PACKAGE, VERSION, CONTACT);
	printf("Support for:\n");
	printf("\t      MySQL:\t");
#ifdef USE_MYSQL
	printf("Yes\n");
#else
	printf("No\n");
#endif
	printf("\t PostgreSQL:\t");
#ifdef USE_POSTGRESQL
	printf("Yes\n");
#else
	printf("No\n");
#endif
	printf("\t SQLite:\t");
#ifdef USE_SQLITE
	printf("Yes\n");
#else
	printf("No\n");
#endif
	printf("\t unixODBC:\t");
#ifdef USE_UNIXODBC
	printf("Yes\n");
#else
	printf("No\n");
#endif
	printf("\t      XPath:\t");
#ifdef USE_XPATH
	printf("Yes\n");
#else
	printf("No\n");
#endif
}

/** Show a usage string	*/
void cfg_show_usage(void)
{
	cfg_show_banner();
	printf("\nUSAGE: %s\n", USAGE);
}

/** Set the default configuration values.	*/
static void cfg_set_defaults(void)
{
	cfg.autosave = DEF_WORKSPACE_AUTOSAVE;
	cfg.currentDir = NULL;
	cfg.dirname = NULL;
	cfg.file = NULL;
}

/** Close the file if it's open.	*/
void cfg_close_file(void)
{
	if (cfg.file) {
		fclose(cfg.file);
		cfg.file = NULL;
	}
}

/** Get the name of the configuration directory.
 * This is normally ($HOME/.gtksql).
 * \return	The name of the configuration directory.	*/
char *cfg_get_config_dir(void)
{
	int len;
	char *homedir = getenv("HOME");

	if (!homedir) {
		printf(_("Warning: Could not resolve home directory name"));
		return NULL;
	}

	/* Set the current dir if it's still NULL. */
	if (cfg.currentDir == NULL) {
		cfg_set_current_dir(homedir);
	}

	len = strlen(homedir) + 1 + strlen(CONFIG_DIRNAME) + 1;
	if (cfg.dirname)
		free(cfg.dirname);

	cfg.dirname = malloc(len);
	if (cfg.dirname) {
		snprintf(cfg.dirname, len, "%s/%s", homedir, CONFIG_DIRNAME);
		cfg.dirname[len - 1] = '\0';
	}
	printf(_("Using configuration directory %s\n"), cfg.dirname);

	if (chdir(cfg.dirname) == -1) {	   /* couldnt change to it */
		if (errno == ENOENT) {		   /* didnt exist, try and create it */
			if (mkdir(cfg.dirname, 0755) == -1) {
				perror(_("Could not create config directory"));
				free(cfg.dirname);
				cfg.dirname = NULL;
				return NULL;
			}
		}
		else {
			perror(_("trying to change to configuration directory"));
			free(cfg.dirname);
			cfg.dirname = NULL;
			return NULL;
		}
	}
	return cfg.dirname;
}

/** Get the name of the config file.
 * @return	The file name.	*/
static char *cfg_get_filename(void)
{
	char *dir = NULL;
	int len;
	if (!cfgname) {
		dir = cfg_get_config_dir();
		len = strlen(dir) + strlen(CONFIG_FILENAME) + 2;
		cfgname = malloc(len);
		if (!cfgname)
			return NULL;

		snprintf(cfgname, len, "%s/%s", dir, CONFIG_FILENAME);
	}
	return cfgname;
}

/** Load the next line form the config file and parse it.
 * \param key	A buffer to fillin with the key.
 * \param val	A buffer to fillin with the value.
 * \return 1 if the values are valid, 0 otherwise.	*/
int cfg_load_line(char *key, char *val)
{
	char buff[CFG_LINE_SIZE];
	regex_t commentregex;
	regex_t configregex;
	char errstr[CFG_LINE_SIZE];
	regmatch_t matches[3];
	int done = 0;
	int err;
	int len;

	if (!cfg.file)
		return 0;

	if (!key)
		return 0;

	if (!val)
		return 0;

	strcpy(key, "");
	strcpy(val, "");

	if ((err = regcomp(&commentregex, "^\\s*#.*$", REG_EXTENDED | REG_NOSUB))) {
		regerror(err, &commentregex, errstr, sizeof(errstr));
		fprintf(stderr, "comment regex error: %s\n", errstr);
		regfree(&commentregex);
		return 0;
	}

	if ((err =
		 regcomp(&configregex,
				 "^\\s*([a-zA-Z\\.]+)\\s*=\\s*(.*)\\s*$",
				 REG_EXTENDED | REG_NEWLINE))) {
		regerror(err, &configregex, errstr, sizeof(errstr));
		fprintf(stderr, "config regex error: %s\n", errstr);
		regfree(&commentregex);
		regfree(&configregex);
		return 0;
	}

	while (!done && !feof(cfg.file)) {
		if (fgets(buff, sizeof(buff), cfg.file)) {
			if (regexec(&commentregex, buff, 0, NULL, 0)) {
				/* no match, try config line */
				if (!regexec(&configregex, buff, 3, matches, 0)) {
					/* matched. Extract substrings */

					/* copy key */
					len = matches[1].rm_eo - matches[1].rm_so;
					strncpy(key, &buff[matches[1].rm_so], len);
					key[len] = '\0';

					/* copy value */
					len = matches[2].rm_eo - matches[2].rm_so;
					strncpy(val, &buff[matches[2].rm_so], len);
					val[len] = '\0';
					done = 1;
				}
			}
		}
	}
	regfree(&commentregex);
	regfree(&configregex);
	return done;
}

/** Open the config file in the given mode.
 * \param mode	The 'fopen' file mode.	*/
void cfg_open_file(char *mode)
{
	char *fn = NULL;

	cfg_close_file();
	fn = cfg_get_filename();

	if (fn == NULL) {
		fprintf(stderr, "ERROR: Configuration filename is null.\n");
		return;
	}

	printf("Opening config file %s\n", fn);
	cfg.file = fopen(fn, mode);
	if (!cfg.file) {
		perror("opening config file");
		return;
	}
}

/** Set the autosave value.
 * \param as	The new value.	*/
void cfg_set_autosave(int as)
{
	if (as == 0)
		cfg.autosave = 0;
	else
		cfg.autosave = 1;
}

/** Process a variable and value from the config file. Ignores any unsupported
 * values.
 * \param key	The variable name.
 * \param val	The variable value.	*/
void cfg_config_value(char *key, char *val)
{
	if (!strncmp(WORKSPACE_AUTOSAVE, key, strlen(WORKSPACE_AUTOSAVE)))
		cfg_set_autosave(atoi(val));
}

void cfg_set_value(char *key, char *val)
{
	script_config_value(key, val);
	ui_config_value(key, val);
	db_config_value(key, val);
}

/** Dump the config settings.	*/
static void cfg_dump_config(void)
{
	printf("%s: %d\n", WORKSPACE_AUTOSAVE, cfg.autosave);
	ui_dump_config();
	script_dump_config();
}

/** Load the configuration from a file. Sets up the workspace variable, and loads the
 * queries, but does NOT disconnect the old database.	*/
void cfg_load_file(void)
{
	char key[CFG_LINE_SIZE];
	char val[CFG_LINE_SIZE];

	cfg_open_file("r");
	while (cfg_load_line(key, val)) {
		cfg_config_value(key, val);
		cfg_set_value(key, val);
	}
	cfg_close_file();
	// cfg_dump_config();
}

/** Parses the command line.
 * \param	argc	As passed to main()
 * \param	argv	As passed to main()
 * \return	Always returns zero.	*/
int cfg_parse_cmdline(int argc, char *argv[])
{
	char c;
	if ((argc < 1) || (! argv))
		return 0;

	while ((c = getopt(argc, argv, ARGSPEC)) != -1) {
		switch (c) {
		case 'c':
			/* Connect to a URL. */
			db_connect_url(optarg, ui_show_error);
			break;
		case 'f':
			/* Run a script file. */
			script_run_file(optarg);
			break;
		case 's':
			/* Run a script directly. */
			script_run_string(optarg);
			break;
		case 'h':
			cfg_show_usage();
			break;
		}
	}
	return 0;
}

/** Initialise the configuration.
 * Performs these steps:
 *  - Sets default values.	*/
void cfg_init(void)
{
	cfg_set_defaults();
}

/** Save a boolean value to the config file.
 * \param key	The variable name.
 * \param value	The variable value.	*/
void cfg_save_bool(const char *key, int value)
{
	if ((! key) || (! value))
		return;

	if (cfg.file) {
		if (value)
			fprintf(cfg.file, "%s=%d\n", key, 1);
		else
			fprintf(cfg.file, "%s=%d\n", key, 0);
	}
	else
		fprintf(stderr, "ERROR: Cannot save config: File not open\n");
}

/** Save an integer value to the config file.
 * \param key	The variable name.
 * \param value	The variable value.	*/
static void cfg_save_int(const char *key, int value)
{
	if (! key)
		return;

	if (cfg.file)
		fprintf(cfg.file, "%s=%d\n", key, value);
}

/** Save a string value to the config file.
 * \param key	The variable name.
 * \param value	The variable value.	*/
void cfg_save_string(const char *key, const char *value)
{
	if ((! key) || (! value))
		return;

	if (cfg.file)
		fprintf(cfg.file, "%s=%s\n", key, value);
}

/** Save the config settings to a file.	*/
void cfg_save_file(void)
{
	cfg_open_file("w");
	if (!cfg.file)
		return;

	/* Save the settings from cfg */
	cfg_save_int(WORKSPACE_AUTOSAVE, cfg.autosave);

	/* Save module settings.  */
	ui_save_config();
	script_save_config();

	cfg_close_file();
	cfg_dump_config();
}

/** Set the current working directory.
 * \param dir The directory to use.	*/
void cfg_set_current_dir(const char *dir)
{
	if (cfg.currentDir != NULL) {
		free(cfg.currentDir);
		cfg.currentDir = NULL;
	}

	if (dir)
		cfg.currentDir = strdup(dir);
}
