/** \file buf.c General-purpose dynamic buffer that adjusts it's size as things are written to it.
 * \author Darryl Luff
 * \addtogroup misc
 *
 * A buffer pointer is passed as the first argument to every function call.
 * All functions return a pointer to a buffer. Note that this may NOT be the 
 * same pointer that went in. For example:
 *
 * \code
 * ABuf *buffer = buf_new(4);
 * buffer = buf_cat(buffer, "123456");
 * \endcode
 *
 * Here, 'buffer's value after the call will be different to the 'buffer's 
 * value before the call, because the buffer is reallocated to fit in the new 
 * contents. So for any function call that can cause the buffer to grow, always 
 * save the return value into the buffer.
 * \par
 * Functions are all NULL-safe. If a null buffer is passed in, a new buffer will
 * be allocated and returned. So instead of the above, could just say:
 * \code
 * ABuf *buffer = buf_cpy(NULL, "123456");
 * \endcode
 * To minimise reallocating memory, it is a good idea to use buf_new to allocate 
 * a buffer with a reasonable size though.
 * \par
 * When finished with the buffer, call buf_free on it. buf_free always returns a
 * \c NULL buffer pointer. This is to let you do:
 * \code
 * buffer = buf_free(buffer);
 * \endcode
 * So the buffer pointer is set to NULL at the same time the memory is freed.
 */
/* GtkSQL -- an interactive graphical query tool for PostgreSQL
 * Copyright (C) 2002-2005 Darryl Luff
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "buf.h"

/** Minimum size buffer allocated. Any buffer allocation will return a buffer 
 * that is at least this size, even \c buf_new(0).	*/
#define MINSIZE	10

/** Returns the size of buffer to allocate, given the requested size.
 * \param n	The requested size.
 * \return	Either the requested size or \ref MINSIZE, whichever is smaller.	*/
unsigned int allocsize(unsigned int n)
{
	if (n < MINSIZE)
		return MINSIZE;
	else
		return n;
}

ABuf *buf_map_chars(ABuf * buf, char c1, char c2)
{
	char *c;

	if (buf_strlen(buf) < 1)
		return buf;

	c = (char *)buf->b_dat;

	while (*c) {
		if (*c == c1)
			*c = c2;
		c++;
	}
	return buf;
}

ABuf *buf_new(unsigned int len)
{
	ABuf *b;
	if (len < 0)
		return NULL;

	b = calloc(1, sizeof(ABuf));
	if (b) {
		b->b_len = allocsize(len);
		b->b_dat = calloc(1, b->b_len);
	}
	return b;
}

/** Free the data for the passed buffer, but not the buffer itself.	*/
void buf_free_data(ABuf * buf)
{
	if (buf) {
		if (buf->b_dat) {
			free(buf->b_dat);
			buf->b_dat = NULL;
		}
		buf->b_len = 0;
	}
}

/** Moves the existing buffer from b1 into b2.
 * The original buffer (if any) in b2 is freed.
 * \returns b2	*/
ABuf *buf_move(ABuf * b1, ABuf * b2)
{
	if (!b2)
		return b2;

	buf_free_data(b2);

	if (b1) {
		b2->b_dat = b1->b_dat;
		b2->b_len = b1->b_len;
		b1->b_dat = NULL;
		b1->b_len = 0;
	}
	return b2;
}

ABuf *buf_escape_quotes(ABuf * buf)
{
	const char *b;
	const char *c;

	if (!buf)
		return buf;

	if (!buf->b_dat)
		return buf;

	/* Make sure buffer is terminated. */
	buf_terminate(buf);

	/* Now steal the buffer */
	b = (const char *)buf->b_dat;
	buf->b_dat = NULL;
	buf->b_len = 0;

	buf = buf_check(buf, strlen(b) + 10);

	c = b;
	while (*c) {
		if ((*c == '"') || (*c == '\''))
			buf = buf_append_char(buf, '\\');

		buf = buf_append_char(buf, *c);
		c++;
	}

	/* Now, move the new buffer into the original ABuf to avoid memory leaks */
	free((void *)b);
	return buf;
}

ABuf *buf_free(ABuf * buf)
{
	if (buf) {
		if (buf->b_dat)
			free(buf->b_dat);

		free(buf);
	}
	return NULL;
}

/** Count the characters left. Leaves room for a trailing NULL.
 * \param	buf	The buffer to check.
 * \return	How many more characters can fit.	*/
int buf_chars_left(ABuf * buf)
{
	if (!buf)
		return 0;

	if ((!buf->b_dat) || (buf->b_len < 2))
		return 0;

	return buf->b_len - strlen((char *)buf->b_dat) - 1;
}

ABuf *buf_append_char(ABuf * buf, char c)
{
	buf = buf_check(buf, 1000);

	if (buf_chars_left(buf) < 1)
		/* Assume will probably append more chars */
		buf = buf_check(buf, buf_strlen(buf) + 10);

	if (buf) {
		buf->b_dat[strlen((char *)buf->b_dat)] = c;
		buf->b_dat[strlen((char *)buf->b_dat) + 1] = '\0';
	}
	return buf;
}

ABuf *buf_check(ABuf * buf, unsigned int len)
{
	if (!buf)
		return buf_new(len);

	if (buf->b_len < len) {
		buf->b_dat = realloc(buf->b_dat, len);
		buf_terminate(buf);
		if (! buf->b_dat)
			buf->b_len = 0;
		else
			buf->b_len = len;
	}
	return buf;
}

/** Copies the contents of b1 into b2. Returns either b2 or a new allocated buffer.
 * \param b1	The source buffer.
 * \param b2	The destination buffer.
 * \return The result buffer.	*/
ABuf *buf_copy(ABuf * b1, ABuf * b2)
{
	ABuf *b;

	if (!b1)
		return b2;

	if (b2)
		b = b2;
	else
		b = buf_new(b1->b_len);

	if (b)
		memmove(b->b_dat, b1->b_dat, b1->b_len);

	return b;
}

ABuf *buf_strcat(ABuf * buf, const char *s)
{
	ABuf *b = NULL;
	int requiredLen;

	if ((! s) || (strlen(s) < 1)) {
		return buf;
	}

	if (!buf)
		b = buf_new(strlen(s) + 1);
	else {
		if (buf->b_dat != NULL)
			requiredLen = strlen((char *)buf->b_dat) + strlen(s) + 1;
		else
			requiredLen = strlen(s) + 1;

		b = buf_check(buf, requiredLen);
	}

	if (b) {
		strncat((char *)b->b_dat, s, b->b_len - strlen((char *)b->b_dat) - 1);
		b = buf_terminate(b);
	}

	return b;
}

ABuf *buf_strcpy(ABuf * buf, const char *s)
{
	ABuf *b = NULL;
	int len;

	if (!s)
		return buf;

	len = strlen(s) + 1;
	if (!buf)
		b = buf_new(len);
	else
		b = buf_check(buf, len);

	if ((b) && (b->b_dat)) {
		if ((s) && (strlen(s) > 0)) {
			strncpy((char *)b->b_dat, s, b->b_len);
			buf_terminate(b);
		}
		else
			*(b->b_dat) = '\0';
	}
	return b;
}

ABuf *buf_str_shrinkbyone(ABuf * buf)
{
	if ((buf == NULL) || (buf->b_dat == NULL) || (strlen((char *)buf->b_dat) < 1))
		return buf;

	buf->b_dat[strlen((char *)buf->b_dat) - 1] = '\0';
	return buf;
}

int buf_strlen(ABuf * buf)
{
	if (!buf)
		return 0;

	if (!buf->b_dat)
		return 0;

	return strnlen((const char *)buf->b_dat, (size_t)buf->b_len);
}

ABuf *buf_terminate(ABuf * buf)
{
	if (! buf)
		return buf;

	if ((buf->b_dat) && (buf->b_len)) {
		buf->b_dat[buf->b_len - 1] = '\0';
	}
	return buf;
}
