/** \file buf.h General-purpose dynamic buffer include file.
 * \author Darryl Luff
 * \addtogroup misc */
/* GtkSQL -- an interactive graphical query tool for PostgreSQL
 * Copyright (C) 2002-2005 Darryl Luff
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifndef __BUF_H
#define __BUF_H

/** Buffer structure.	*/
typedef struct _a_buf {
	unsigned char *b_dat;	/**< Pointer to the buffer data.	*/
	unsigned int b_len;		/**< Length of the allocated buffer.	*/
} ABuf;

/** Always use:
 * \code
 * buf = buf_append_char(buf, 'c');
 * \endcode
 * with this function, and never rely on the returned pointer being the same as
 * the passed pointer.
 * \param	buf	The original pointer. If \c NULL, a new buffer will be allocated.
 * \param	c	The character to add.
 * \return	The new buffer.	*/
ABuf *buf_append_char(ABuf * buf, char c);

/** Make sure that the passed buffer is large enough. If it is not at least
 * 'len' bytes long, a new buffer is allocated. Returns a pointer to either
 * the original or the new buffer. Always use:
 * \code
 * buf = buf_check(buf, len);
 * \endcode
 * with this function, and never rely on the returned pointer being the same as
 * the passed pointer.
 * \param	buf	The original pointer. If \c NULL, a new buffer will be allocated.
 * \param	len	The requested buffer size.
 * \return	Either the original buffer if it was large enough, or a new larger buffer.	*/
ABuf *buf_check(ABuf * buf, unsigned int len);

/** Escapes any quotes (", ') in the buffer with '\'.
 * \param buf	The original buffer.
 * \return	The new buffer.	*/
ABuf *buf_escape_quotes(ABuf * buf);

/** Free the passed buffer.
 * \return NULL pointer so you can use
 * \code
 * 'buf = buf_free(buf);
 * \endcode
 * to free the buffer and clear the variable in one line.	*/
ABuf *buf_free(ABuf * buf);

/** Replace all of the 'c1' chars in txt to 'c2'.	*/
ABuf *buf_map_chars(ABuf * buf, char c1, char c2);

/** Allocate a new ABuf. \ref allocsize() is used to calculate the buffer size,
 * so the returned buffer may be larger than the requested size.
 * \param len	If a length of zero or negative is passed, a buffer is still allocated.
 * \return	A pointer to the new buffer.
 * \see	MINSIZE	*/
ABuf *buf_new(unsigned int len);

/** Sets the last non-NULL byte of the string to NULL, effectively shortening the string by one
 * character.
 * \param	buf	The original buffer.
 * \return	The buffer.	*/
ABuf *buf_str_shrinkbyone(ABuf * buf);

/** Appends the string 's' to the buffer 'buf'. The buffer is reallocated if
 * necessary.
 * \code
 * buf = buf_strcat(buf, "Hello");
 * \endcode
 * \param	buf	The original buffer
 * \param	s	The string to append
 * \return	The new buffer	*/
ABuf *buf_strcat(ABuf * buf, const char *s);

/** Copies the string 's' into the buffer. The buffer is reallocated
 * if required.
 * \code
 * buf = buf_strcpy(buf, "Hello");
 * \endcode
 * \param	buf	The original buffer
 * \param	s	The string to copy
 * \return	The new buffer.	*/
ABuf *buf_strcpy(ABuf * buf, const char *s);

/** Find the length of the string in the buffer. Make sure the
 * buffer does contain a null-terminated string before calling!
 * in it.
 * \param	buf	The buffer.
 * \return	The string length. If buffer is NULL, returns 0.	*/
int buf_strlen(ABuf * buf);

/** Sets the last byte of the buffer to a null byte, overwriting whatever's
 * in it.
 * \param	buf	The original buffer.
 * \return	The buffer.	*/
ABuf *buf_terminate(ABuf * buf);

#endif
