
#include "../config.h"

#include <check.h>
#include <stdlib.h>
#include <string.h>

#include "buf.h"

/** Always use:
 * \code
 * buf = buf_append_char(buf, 'c');
 * \endcode
 * with this function, and never rely on the returned pointer being the same as
 * the passed pointer.
 * \param	buf	The original pointer. If \c NULL, a new buffer will be allocated.
 * \param	c	The character to add.
 * \return	The new buffer.	*/
//ABuf *buf_append_char(ABuf * buf, char c);

/** Make sure that the passed buffer is large enough. If it is not at least
 * 'len' bytes long, a new buffer is allocated. Returns a pointer to either
 * the original or the new buffer. Always use:
 * \code
 * buf = buf_check(buf, len);
 * \endcode
 * with this function, and never rely on the returned pointer being the same as
 * the passed pointer.
 * \param	buf	The original pointer. If \c NULL, a new buffer will be allocated.
 * \param	len	The requested buffer size.
 * \return	Either the original buffer if it was large enough, or a new larger buffer.	*/
//ABuf *buf_check(ABuf * buf, unsigned int len);

/** Escapes any quotes (", ') in the buffer with '\'.
 * \param buf	The original buffer.
 * \return	The new buffer.	*/
//ABuf *buf_escape_quotes(ABuf * buf);

/** Replace all of the 'c1' chars in txt to 'c2'.	*/
//ABuf *buf_map_chars(ABuf * buf, char c1, char c2);

/** Sets the last non-NULL byte of the string to NULL, effectively shortening the string by one
 * character.
 * \param	buf	The original buffer.
 * \return	The buffer.	*/
//ABuf *buf_str_shrinkbyone(ABuf * buf);

#define TESTSTRING "TestString"
#define TESTSTRING2 "TestString"
#define TESTALL TESTSTRING TESTSTRING2

START_TEST(buf_test_new_free)
{
	ABuf *buf = NULL;

	/* tests buf_strcat, buf_strlen, buf_strcpy	*/
	buf = buf_new(-100);
	fail_if(buf == NULL, "buf_new(-100) returned NULL");
	buf = buf_free(buf);
	fail_unless(buf == NULL, "buf_free() did not return NULL");
	buf = buf_new(0);
	fail_if(buf == NULL, "buf_new(0) returned NULL");
	buf = buf_free(buf);
}
END_TEST

START_TEST(buf_test_strings)
{
	ABuf *buf = NULL;
	int err;

	/* tests buf_strcat, buf_strlen, buf_strcpy	*/
	buf = buf_strcat(NULL, NULL);
	fail_unless(buf == NULL, "buf_strcat(NULL, NULL) did not return NULL");
	buf = buf_strcpy(NULL, NULL);
	fail_unless(buf == NULL, "buf_strcpy(NULL, NULL) did not return NULL");
	err = buf_strlen(NULL);
	fail_unless(err ==0, "buf_strlen(NULL) did not return 0");

	buf = buf_new(0);
	fail_if(buf == NULL, "Could not allocate buffer");
	buf = buf_strcpy(buf, TESTSTRING);
	err = strcmp(TESTSTRING, buf->b_dat);
	fail_unless(err == 0, "buf_cpy() did not copy string");
	fail_unless(buf_strlen(buf) == strlen(TESTSTRING), "buf_strlen() returned %d", strlen(TESTSTRING));
	buf = buf_strcat(buf, TESTSTRING2);
	err = strcmp(TESTALL, buf->b_dat);
	fail_unless(err == 0, "buf_cat() did not copy string properly");
	buf = buf_free(buf);
	fail_unless(buf == NULL, "buf_free() did not return NULL");
}
END_TEST

START_TEST(buf_test_terminate)
{
	ABuf *buf = NULL;
	int len;

	buf = buf_terminate(buf);
	fail_if(buf != NULL, "buf_terminate(NULL) did not return NULL");
	buf = buf_new(4);
	fail_if(buf == NULL, "Could not allocate buffer");
	len = buf->b_len;
	fail_if(len < 4, "Allocated buffer is too small");
	buf->b_dat[len - 1] = (char)9;
	buf = buf_terminate(buf);
	fail_unless(buf->b_dat[len - 1] == 0, "buf_terminate did not terminate the buffer");
	buf = buf_free(buf);
	fail_unless(buf == NULL, "buf_free() did not return NULL");
}
END_TEST

START_TEST(appconfig_test_cfg_parse_cmdline)
{
	/* int cfg_parse_cmdline(int argc, char *argv[]);
	 * Just checking for null pointer expections.	*/
	cfg_parse_cmdline(-1, NULL);
	cfg_parse_cmdline(0, NULL);
	cfg_parse_cmdline(1, NULL);
}
END_TEST

START_TEST(appconfig_test_cfg_save_bool)
{
	/* void cfg_save_bool(const char *key, int value);
	 * Just checking for null pointer expections.	*/
	cfg_save_bool(NULL, 0);
	cfg_save_bool("JunkBool", -1);
	cfg_save_bool("JunkBool", 0);
	cfg_save_bool("JunkBool", 1);
}
END_TEST

START_TEST(appconfig_test_cfg_save_string)
{
	/* void cfg_save_string(const char *key, const char *value);
	 * Just checking for null pointer expections.	*/
	cfg_save_string(NULL, NULL);
	cfg_save_string("Junk", NULL);
	cfg_save_string(NULL, "Junk");
	cfg_save_string("Junk", "Junk");
}
END_TEST

START_TEST(appconfig_test_cfg_set_current_dir)
{
	/* void cfg_set_current_dir(const char *dir);
	 * Just checking for null pointer expections.	*/
	cfg_set_current_dir(NULL);
}
END_TEST

static void	add_appconfig_tests(Suite *s) {
	TCase *tc;
	tc = tcase_create("appconfig");
	suite_add_tcase(s, tc);
	tcase_add_test(tc, appconfig_test_cfg_parse_cmdline);
	tcase_add_test(tc, appconfig_test_cfg_save_bool);
	tcase_add_test(tc, appconfig_test_cfg_save_string);
	tcase_add_test(tc, appconfig_test_cfg_set_current_dir);
}

static void	add_buf_tests(Suite *s) {
	TCase *tc;

	tc = tcase_create("Buf");
	suite_add_tcase(s, tc);
	tcase_add_test(tc, buf_test_new_free);
	tcase_add_test(tc, buf_test_strings);
	tcase_add_test(tc, buf_test_terminate);
}

static void	add_util_tests(Suite *s) {
	TCase *tc;

	tc = tcase_create("Utils");
	suite_add_tcase(s, tc);
	//tcase_add_test(tc, utils_test_XXX);
}

Suite * db_suite(void)
{
	Suite *s = suite_create("Various");
	add_appconfig_tests(s);
	add_buf_tests(s);
	add_util_tests(s);

	return s;
}

int main(void)
{
	int nf;

	Suite *s = db_suite();
	SRunner *sr = srunner_create(s);
	srunner_run_all(sr, CK_VERBOSE);
	nf = srunner_ntests_failed(sr);
	srunner_free(sr);
	return (nf == 0) ? EXIT_SUCCESS : EXIT_FAILURE;
}
