
#include "../config.h"

#include <check.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "dbapi.h"

#ifdef USE_MYSQL
#include "d_mysql_f.h"
#endif

#ifdef USE_POSTGRESQL
#include "d_postgres_f.h"
#endif

#ifdef USE_SQLITE
#include "d_sqlite.h"
#endif

#ifdef USE_UNIXODBC
#include "d_unixodbc.h"
#endif

#include "forms.h"

DBConnector *dbc = NULL;


void check_url_parse(DBConnector * dbc, const char *url, const char *host, const char *port, const char *db,
						 const char *name, const char *pass)
{
	Form *frm;
	int dbMatch = 0;
	int hostMatch = 0;
	int portMatch = 0;
	int nameMatch = 0;
	int passMatch = 0;
	const char *s;

	frm = frm_new(NULL);
	fail_if(frm == NULL, "could not initialise form");

	dbc->url_to_form(frm, url);
	s = frm_get_value_by_name(frm, "Host");
	if (host == NULL) {
		if (s == NULL)
			hostMatch = 1;
	}
	else if ((s != NULL) && (strcmp(s, host) == 0))
		hostMatch = 1;

	s = frm_get_value_by_name(frm, "Database");
	if (db == NULL) {
		if (s == NULL)
			dbMatch = 1;
	}
	else if (strcmp(s, db) == 0)
		dbMatch = 1;

	s = frm_get_value_by_name(frm, "Port");
	if (port == NULL) {
		if (s == NULL)
			portMatch = 1;
	}
	else if (strcmp(s, port) == 0)
		portMatch = 1;

	s = frm_get_value_by_name(frm, "Username");
	if (name == NULL) {
		if (s == NULL)
			nameMatch = 1;
	}
	else if (strcmp(s, name) == 0)
		nameMatch = 1;

	s = frm_get_value_by_name(frm, "Password");
	if (pass == NULL) {
		if (s == NULL)
			passMatch = 1;
	}
	else if (strcmp(s, pass) == 0)
		passMatch = 1;

	fail_unless(dbMatch, "Database parsed incorrectly: %s", url);
	fail_unless(hostMatch, "Host parsed incorrectly: %s", url);
	fail_unless(portMatch, "Port parsed incorrectly: %s", url);
	fail_unless(nameMatch, "Username parsed incorrectly: %s", url);
	fail_unless(passMatch, "Password parsed incorrectly: %s", url);
}

START_TEST(test_init_db_frm)
{
	fail_if(dbc == NULL, "TEST FAILURE: dbc not initialised");
	fail_if(dbc->db_frm == NULL, "Did not initialise db_frm");
}
END_TEST

START_TEST(test_init_db_connect_callback)
{
	fail_if(dbc == NULL, "TEST FAILURE: dbc not initialised");
	fail_if(dbc->db_connect_callback == NULL, "db_connect_callback not initialised");
}
END_TEST

START_TEST(test_init_dbtype)
{
	fail_if(dbc == NULL, "TEST FAILURE: dbc not initialised");
	fail_if(dbc->dbtype == NULL, "Did not initialise dbtype string");
}
END_TEST

START_TEST(test_init_errmsg)
{
	fail_if(dbc == NULL, "TEST FAILURE: dbc not initialised");
	fail_if(dbc->errmsg != NULL, dbc->errmsg);
}
END_TEST

void teardown(void) {
	/* Huge memory leak - there's no defined way to deallocate a driver at the moment	*/
	dbc = NULL;
}

/* ================
 * MYSQL tests
 * ================ */

#ifdef USE_MYSQL
START_TEST(mysql_test_url_to_form)
{
	fail_if(dbc == NULL, "TEST ERROR: dbc not initialised");
	fail_if(dbc->url_to_form == NULL, "url_to_form not initialised");

	/* dbc, url, host, port, db, name, pass)	*/
	check_url_parse(dbc, NULL, NULL, NULL, NULL, NULL, NULL);
	check_url_parse(dbc, "", NULL, NULL, NULL, NULL, NULL);
	check_url_parse(dbc, "junk:/stuff", NULL, NULL, NULL, NULL, NULL);
	check_url_parse(dbc, "junk//stuff", NULL, NULL, NULL, NULL, NULL);
	check_url_parse(dbc, "junk://stuff", NULL, NULL, NULL, NULL, NULL);
	check_url_parse(dbc, "mysql://dbserver", "dbserver", NULL, NULL, NULL, NULL);
	check_url_parse(dbc, "mysql://dbserver:99", "dbserver", "99", NULL, NULL, NULL);
	check_url_parse(dbc, "mysql://UserX@dbserver", "dbserver", NULL, NULL, "UserX", NULL);
	check_url_parse(dbc, "mysql://UserX:PassY@dbserver", "dbserver", NULL, NULL, "UserX", "PassY");
	check_url_parse(dbc, "mysql://UserX:PassY@dbserver:99", "dbserver", "99", NULL, "UserX", "PassY");
	check_url_parse(dbc, "mysql://UserX:PassY@dbserver/thedb", "dbserver", NULL, "thedb", "UserX", "PassY");
	check_url_parse(dbc, "mysql://UserX:PassY@dbserver:99/thedb", "dbserver", "99", "thedb", "UserX", "PassY");
}
END_TEST

void setup_mysql(void) {
	dbc = register_MySQL();
}

static void	add_mysql_tests(Suite *s) {
	TCase *tc;

	tc = tcase_create("MySQL");
	tcase_add_checked_fixture(tc, setup_mysql, teardown);
	tcase_add_test(tc, mysql_test_url_to_form);
	tcase_add_test(tc, test_init_dbtype);
	tcase_add_test(tc, test_init_errmsg);
	tcase_add_test(tc, test_init_db_frm);
	tcase_add_test(tc, test_init_db_connect_callback);
	suite_add_tcase(s, tc);
}
#endif

/* ================
 * POSTGRESQL tests
 * ================ */

#ifdef USE_POSTGRESQL
START_TEST(postgresql_test_url_to_form)
{
	fail_if(dbc == NULL, "TEST ERROR: dbc not initialised");
	fail_if(dbc->url_to_form == NULL, "url_to_form not initialised");

	/* dbc, url, host, port, db, name, pass)	*/
	check_url_parse(dbc, NULL, NULL, NULL, NULL, NULL, NULL);
	check_url_parse(dbc, "", NULL, NULL, NULL, NULL, NULL);
	check_url_parse(dbc, "junk:/stuff", NULL, NULL, NULL, NULL, NULL);
	check_url_parse(dbc, "junk//stuff", NULL, NULL, NULL, NULL, NULL);
	check_url_parse(dbc, "junk://dbserver", "dbserver", NULL, NULL, NULL, NULL);
	check_url_parse(dbc, "postgresql://dbserver", "dbserver", NULL, NULL, NULL, NULL);
	check_url_parse(dbc, "postgresql://dbserver:99", "dbserver", "99", NULL, NULL, NULL);
	check_url_parse(dbc, "postgresql://UserX@dbserver", "dbserver", NULL, NULL, "UserX", NULL);
	check_url_parse(dbc, "postgresql://UserX:PassY@dbserver", "dbserver", NULL, NULL, "UserX", "PassY");
	check_url_parse(dbc, "postgresql://UserX:PassY@dbserver:99", "dbserver", "99", NULL, "UserX", "PassY");
	check_url_parse(dbc, "postgresql://UserX:PassY@dbserver/thedb", "dbserver", NULL, "thedb", "UserX", "PassY");
	check_url_parse(dbc, "postgresql://UserX:PassY@dbserver:99/thedb", "dbserver", "99", "thedb", "UserX", "PassY");
}
END_TEST

void setup_postgresql(void) {
	dbc = register_PostgreSQL();
}

static void	add_postgresql_tests(Suite *s) {
	TCase *tc;

	tc = tcase_create("PostgreSQL");
	tcase_add_checked_fixture(tc, setup_postgresql, teardown);
	tcase_add_test(tc, postgresql_test_url_to_form);
	tcase_add_test(tc, test_init_dbtype);
	tcase_add_test(tc, test_init_errmsg);
	tcase_add_test(tc, test_init_db_frm);
	tcase_add_test(tc, test_init_db_connect_callback);
	suite_add_tcase(s, tc);
}
#endif

/* ================
 * SQLite tests
 * ================ */

#ifdef USE_SQLITE
void check_url_parse_sqlite(DBConnector * dbc, const char *url, const char *filename)
{
	Form *frm;
	int fnMatch = 0;
	const char *s;

	frm = frm_new(NULL);
	fail_if(frm == NULL, "could not initialise form");

	dbc->url_to_form(frm, url);
	s = frm_get_value_by_name(frm, "Database File");
	if (filename == NULL) {
		if (s == NULL)
			fnMatch = 1;
	}
	else if (strcmp(s, filename) == 0)
		fnMatch = 1;

	fail_unless(fnMatch, "Database file parsed incorrectly");
}

START_TEST(sqlite_test_url_to_form)
{
	fail_if(dbc == NULL, "TEST ERROR: dbc not initialised");
	fail_if(dbc->url_to_form == NULL, "url_to_form not initialised");

	/* dbc, url, host, port, db, name, pass)	*/
	check_url_parse_sqlite(dbc, NULL, NULL);
	check_url_parse_sqlite(dbc, "", NULL);
	check_url_parse_sqlite(dbc, "junk:/stuff", NULL);
	check_url_parse_sqlite(dbc, "junk//stuff", NULL);
	check_url_parse_sqlite(dbc, "sqlite:///file/name", "/file/name");
}
END_TEST

void setup_sqlite(void) {
	dbc = register_SQLite();
}

static void	add_sqlite_tests(Suite *s) {
	TCase *tc;

	tc = tcase_create("SQLite");
	tcase_add_checked_fixture(tc, setup_sqlite, teardown);
	tcase_add_test(tc, sqlite_test_url_to_form);
	tcase_add_test(tc, test_init_dbtype);
	tcase_add_test(tc, test_init_errmsg);
	tcase_add_test(tc, test_init_db_frm);
	tcase_add_test(tc, test_init_db_connect_callback);
	suite_add_tcase(s, tc);
}
#endif

/* ================
 * UNIXODBC tests
 * ================ */

#ifdef USE_UNIXODBC
void check_url_parse_unixodbc(DBConnector * dbc, const char *url, const char *datasource, const char *user, const char *pass )
{
	Form *frm;
	int dsMatch = 0;
	int passMatch = 0;
	int userMatch = 0;
	const char *s;

	frm = frm_new(NULL);
	fail_if(frm == NULL, "could not initialise form");

	dbc->url_to_form(frm, url);
	s = frm_get_value_by_name(frm, "Datasource");
	if (datasource == NULL) {
		if (s == NULL)
			dsMatch = 1;
	}
	else if ((s != NULL) && (strcmp(s, datasource) == 0))
		dsMatch = 1;

	s = frm_get_value_by_name(frm, "Username");
	if (user == NULL) {
		if (s == NULL)
			userMatch = 1;
	}
	else if ((s != NULL) && (strcmp(s, user) == 0))
		userMatch = 1;

	s = frm_get_value_by_name(frm, "Password");
	if (pass == NULL) {
		if (s == NULL)
			passMatch = 1;
	}
	else if ((s != NULL) && (strcmp(s, pass) == 0))
		passMatch = 1;

	fail_unless(dsMatch, "Datasource parsed incorrectly: %s (%s)", url, datasource);
	fail_unless(userMatch, "Username parsed incorrectly: %s", url);
	fail_unless(passMatch, "Password parsed incorrectly: %s", url);
}

START_TEST(unixodbc_test_url_to_form)
{
	fail_if(dbc == NULL, "TEST ERROR: dbc not initialised");
	fail_if(dbc->url_to_form == NULL, "url_to_form not initialised");

	/* dbc, url, host, port, db, name, pass)	*/
	check_url_parse_unixodbc(dbc, NULL, NULL, NULL, NULL);
	check_url_parse_unixodbc(dbc, "", NULL, NULL, NULL);
	check_url_parse_unixodbc(dbc, "junk:/stuff", NULL, NULL, NULL);
	check_url_parse_unixodbc(dbc, "junk//stuff", NULL, NULL, NULL);
	check_url_parse_unixodbc(dbc, "junk://dbserver", NULL, NULL, NULL);
	check_url_parse_unixodbc(dbc, "unixodbc://dbserver", "dbserver", NULL, NULL);
	check_url_parse_unixodbc(dbc, "unixodbc://UserX@dbserver", "dbserver", "UserX", NULL);
	check_url_parse_unixodbc(dbc, "unixodbc://UserX:PassY@dbserver", "dbserver", "UserX", "PassY");
	check_url_parse_unixodbc(dbc, "unixodbc://UserX:PassY@dbserver/", "dbserver", "UserX", "PassY");
}
END_TEST

void setup_unixodbc(void) {
	dbc = register_unixODBC();
}

static void	add_unixodbc_tests(Suite *s) {
	TCase *tc;

	tc = tcase_create("unixODBC");
	tcase_add_checked_fixture(tc, setup_unixodbc, teardown);
	tcase_add_test(tc, unixodbc_test_url_to_form);
	tcase_add_test(tc, test_init_dbtype);
	tcase_add_test(tc, test_init_errmsg);
	tcase_add_test(tc, test_init_db_frm);
	tcase_add_test(tc, test_init_db_connect_callback);
	suite_add_tcase(s, tc);
}
#endif

Suite * db_suite(void)
{
	Suite *s = suite_create("Database");
	TCase *tc;

	tc = NULL;

#ifdef USE_MYSQL
	add_mysql_tests(s);
#endif

#ifdef USE_POSTGRESQL
	add_postgresql_tests(s);
#endif

#ifdef USE_SQLITE
	add_sqlite_tests(s);
#endif

#ifdef USE_UNIXODBC
	add_unixodbc_tests(s);
#endif

	return s;
}

int main(void)
{
	int nf;

	Suite *s = db_suite();
	SRunner *sr = srunner_create(s);
	srunner_run_all(sr, CK_VERBOSE);
	nf = srunner_ntests_failed(sr);
	srunner_free(sr);
	return (nf == 0) ? EXIT_SUCCESS : EXIT_FAILURE;
}
