/** \file d_mysql_f.c	Forms version of MySQL driver.	
 * \author Des Herriot
 * \author Lionel ULMER
 * \author Darryl Luff
 * \ingroup db
 *
 * This is a user-interface independent database driver linking MySQL with 
 * the GtkSQL application.
 */
/* GtkSQL -- an interactive graphical query tool for PostgreSQL and MySQL.
 * Copyright (C) 1998-2005 Lionel ULMER, Darryl Luff
 * MySQL "driver" (C) 1998 Des Herriott (des@ops.netcom.net.uk)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "../config.h"

#ifdef USE_MYSQL
#include <stdlib.h>
#include <string.h>
#include <mysql.h>

#include "dbapi.h"
#include "utils.h"

#define MSG(a,b) if (display_error) display_error(a,b); else fprintf(stderr, "%s: %s\n", a, b);

/** URL for generic help	*/
#define HELPURL	"http://www.mysql.com/"
#define HELPSEARCHURL "http://dev.mysql.com/doc/mysql/search.php?q=%s"

/** Service type for zeroconf/avahi	*/
#define SERVICETYPE "_mysql_._tcp"

/* The text tags to use as the form field names.	*/

/** Database name.	*/
#define BASENAME	_("Database")

/** Database host. */
#define BASEHOST	_("Host")

/** Database port number.	*/
#define BASEPORT	_("Port")

/** Database user name.	*/
#define BASELOGIN	_("Username")

/** Database password.	*/
#define BASEPASS	_("Password")

/** The MySQL-specific database connection structure.	*/
typedef struct
{
	MYSQL *conn;			/**< The MySQL connection.	*/
} DBConnection_MYSQL;

/** A nasty hack: if a MySQL query succeeded, but returned a NULL result 
 * (e.g. after an UPDATE), we return the address of dummy_res from 
 * DBexecute_query().  This is so we can distinguish a successful query
 * which returned nothing from a failed query in DBget_query_status(). */
MYSQL_RES dummyRes;

/** The MySQL keywords to be added to the keyword-completion table.	*/
static char *sqlKeywords[] = {
	"ABS", "AND", "ALTER", "ASC", "ASCII",
	"BEGIN", "BETWEEN", "BIN", "BINARY", "BIT_LENGTH",
	"CASE", "CEILING", "CHAR", "CHAR_LENGTH", "CHARACTER_LENGTH",
	"COALESCE", "COMMIT", "CONCAT", "CONCAT_WS", "CONV", "COUNT",
	"CREATE",
	"DATABASE", "DELETE", "DESC", "DESCRIBE", "DISTINCT", "DISTINCTROW",
	"DROP",
	"ELT", "END", "EXP", "EXPLAIN", "EXPORT_SET",
	"FIELD", "FIND_IN_SET", "FLOOR", "FROM",
	"GRANT", "GROUP BY",
	"HAVING", "HEX",
	"IFNULL", "INDEX", "INNER", "INSERT", "INSTR", "INTERVAL", "INTO",
	"ISNULL",
	"JOIN",
	"LCASE", "LEFT", "LENGTH", "LIKE", "LIMIT", "LOAD DATA", "LOAD_FILE", "LOCATE",
	"LOCK", "LOG", "LOG2", "LOG10", "LOWER", "LPAD", "LTRIM",
	"MAKE_SET", "MATCH", "MAX", "MID", "MIN", "MEAN", "MOD",
	"NATURAL", "NOT", "NULL", "NULLIF",
	"OCT", "OCTET_LENGTH", "OR", "ORD", "ORDER BY", "OUTER",
	"POSITION", "POW", "POWER",
	"QUOTE",
	"REGEXP", "REPEAT", "REPLACE", "RESET", "REVERSE", "REVOKE", "RIGHT",
	"RLIKE", "ROLLBACK", "ROUND", "RPAD", "RTRIM",
	"SELECT", "SET", "SHOW", "SIGN", "SOUNDEX", "SPACE", "SQRT", "STRCMP",
	"SUBSTRING", "SUBSTRING_INDEX", "SUM",
	"TABLE", "THEN", "TRANSACTION",
	"UCASE", "UNION", "UPDATE", "UPPER",
	"VIEW",
	"WHEN", "WHERE", "WORK",
	NULL
};

/** Free a DBConnection and it's related DBConnection_MYSQL.
 * \param	dbconn	The DBConnection to free.	*/
static void
mysql_dbconnection_free (DBConnection * dbconn)
{
	DBConnection_MYSQL *dbmysql;

	if (! dbconn)
		return;

	dbmysql = (DBConnection_MYSQL *) (dbconn->pvt);
	if (! dbmysql)
		return;

	/* Clear Keywords memory */
	keywords_remove_all (dbconn);

	/* Free structure memory */
	free (dbmysql);
	free (dbconn);
}

/** Disconnect from database.
 * \param	c	The \ref DBConnection to disconnect.
 * \return	Always returns zero.	*/
static int
my_DBdisconnect (DBConnection * dbconn)
{
	DBConnection_MYSQL *dbmysql;

	if (! dbconn)
		return 0;

	dbmysql = (DBConnection_MYSQL *) (dbconn->pvt);
	if (! dbmysql)
		return 0;

	/* End connection */
	if (dbmysql->conn)
		mysql_close (dbmysql->conn);

	mysql_dbconnection_free (dbconn);

	return 0;
}

/** Return the URL for MySQL documentation search.	*/
const char *my_DBget_driver_info(const DBConnector *dbc)
{
	return mysql_get_client_info();
}

/** Return the error message from the connection.
 * \param	c	The \ref DBConnection to reference.
 * \return	The last error message on this connection.	*/
static const char *
my_DBget_error_message (const DBConnection * dbconn)
{
	DBConnection_MYSQL *dbmysql;

	if (! dbconn)
		return NULL;

	dbmysql = (DBConnection_MYSQL *) (dbconn->pvt);
	if (! dbmysql)
		return "No connection";

	return mysql_error (dbmysql->conn);
}

/** Send a query to mysql. If the returned result is \c NULL, and the number of
 * fields is zero, returns a pointer to the dummy result \ref dummy_res.
 * \param	c	The connection to send the query on.
 * \param	query	The query text.	*/
static DBQResult *
my_DBexecute_query (DBConnection * dbconn, const char *query)
{
	int fields;
	DBConnection_MYSQL *dbmysql;
	MYSQL_RES *res;
	DBQResult *result;

	if (! dbconn)
		return NULL;

	dbmysql = (DBConnection_MYSQL *) (dbconn->pvt);
	if (! dbmysql)
		return NULL;

	if (mysql_query (dbmysql->conn, query))
		return NULL;

	res = mysql_store_result (dbmysql->conn);

	if (res == NULL) {
#ifdef HAVE_MYSQL_FIELD_COUNT
		/* This is for versions 3.22.24 and after */
		fields = mysql_field_count (dbmysql->conn);
#else
		fields = mysql_num_fields (dbmysql->conn);
#endif
		if (fields == 0)
			res = &dummyRes;
	}
	result = dbresult_new(res);

	return result;
}

static ABuf *f2ubuf = NULL;

/**
 * \param frm	The form to copy to.
 * \return The buffer containing the URL.	*/
const char *
my_DBform_to_url (const Form * frm, int showPwd)
{
	const char *s;
	char *at = "";

	if (!frm)
		return NULL;

	f2ubuf = buf_check (f2ubuf, 100);
	f2ubuf = buf_strcpy (f2ubuf, frm_get_title (frm));
	f2ubuf = buf_strcat (f2ubuf, "://");
	s = frm_get_value_by_name (frm, BASELOGIN);
	if ((s != NULL) && (strlen(s) > 0)) {
		f2ubuf = buf_strcat (f2ubuf, s);
		at = "@";
	}

	if (showPwd) {
		s = frm_get_value_by_name (frm, BASEPASS);
		if (s != NULL) {
			f2ubuf = buf_strcat (f2ubuf, ":");
			f2ubuf = buf_strcat (f2ubuf, s);
		}
	}
	f2ubuf = buf_strcat (f2ubuf, at);

	s = frm_get_value_by_name (frm, BASEHOST);
	if (s == NULL)
		s = "localhost";

	f2ubuf = buf_strcat (f2ubuf, s);
	f2ubuf = buf_strcat (f2ubuf, "/");
	s = frm_get_value_by_name (frm, BASENAME);
	if (s != NULL)
		f2ubuf = buf_strcat (f2ubuf, s);

	return (const char *)f2ubuf->b_dat;
}

/** Return the URL for the MySQL home page.	*/
const char *my_DBget_help_url(const DBConnector *dbc)
{
	return HELPURL;
}

/** Return the service type for zeroconf.	*/
const char *my_DBget_service_type(const DBConnector *dbc)
{
	return SERVICETYPE;
}

/** Return the URL for MySQL documentation search.	*/
const char *my_DBget_help_search_url(const DBConnector *dbc, const char *term)
{
#define MAXURLLEN 2048
	static char buff[MAXURLLEN];

	if ((! term) || ((strlen(HELPSEARCHURL) + strlen(term)) >= MAXURLLEN))
		return HELPURL;

	snprintf(buff, MAXURLLEN-1, HELPSEARCHURL, term);
	return buff;
}

/** Get the status from the returned result.
 * \param	c	The connection to operate on.
 * \param	result	The result pointer.
 * \return	The status code.
 * \see QueryStatus.	*/
static int
my_DBget_query_status (const DBConnection * dbconn, const DBQResult *result)
{
	DBConnection_MYSQL *dbmysql;
	MYSQL_RES *res;
	const char *err;

	if (! dbconn)
		return DB_QUERY_ERROR;

	dbmysql = (DBConnection_MYSQL *) (dbconn->pvt);
	if (! dbmysql)
		return DB_QUERY_ERROR;

	res = dbresult_get_data(result);
	if (! res)
		return DB_QUERY_ERROR;

	err = mysql_error (dbmysql->conn);
	if (err[0])
	{
		return DB_QUERY_ERROR;
	}
	else if (res == &dummyRes)
	{
		return DB_QUERY_COMMAND_OK;
	}
	else if (mysql_num_rows (res) == 0)
	{
		return DB_QUERY_EMPTY;
	}
	else
	{
		return DB_QUERY_RECORDS_OK;
	}
}

/** Return the number of rows in the result.
 * \param	c	The connection to use.
 * \param	result	The result data.
 * \return	The number of result rows.	*/
static int
my_DBget_record_count(const DBConnection * dbconn, const DBQResult *result)
{
	MYSQL_RES *res;
	if (! result)
		return 0;

	res = dbresult_get_data(result);
	if (! res)
		return 0;

	return mysql_num_rows (res);
}

/** Returns the number of fields (columns) in the result.
 * \param	c	The connection to use.
 * \param	result	The result data.
 * \return	The number of result columns.	*/
static int
my_DBget_field_count(const DBConnection * dbconn, const DBQResult *result)
{
	MYSQL_RES *res;
	res = dbresult_get_data(result);
	if (! res)
		return 0;

	return mysql_num_fields (res);
}

/** Return the field-name for the selected field.
 * \param	c	The connection to use.
 * \param	result	The query results.
 * \param	field	The field number. This is a zero-based index.	*/
static const char *
my_DBget_field_name (const DBConnection * dbconn, const DBQResult *result, int field)
{
	MYSQL_RES *res;
	MYSQL_FIELD *f;

	res = dbresult_get_data(result);
	if (! res)
		return NULL;

	mysql_field_seek (res, field);
	if ((f = mysql_fetch_field (res)) == NULL)
		return NULL;

	return f->name;
}

/** Returns the value of the selected field.
 * \param	c	The connection to use.
 * \param	result	The query results.
 * \param	record	The record number to access (zero-based).
 * \param	field	The field number to access (zero-based).
 * \return	The field value as a string.	*/
static const char *
my_DBget_field_value (const DBConnection * dbconn, const DBQResult *result,
		      int record, int field)
{
	MYSQL_RES *res;
	MYSQL_ROW row;

	res = (MYSQL_RES *)dbresult_get_data(result);
	if (! res)
		return "";

	mysql_data_seek (res, record);
	if ((row = mysql_fetch_row (res)) == NULL)
	{
		return NULL;
	}
	return (const char *) row[field];
}

/** Clears the result.
 * \param	c	The connection to use.
 * \param	result	The result set to clear.	*/
static void
my_DBclear_result (DBConnection * dbconn, DBQResult *result)
{
	MYSQL_RES *res;

	if (! dbconn)
		return;

	res = dbresult_get_data(result);
	if ((res) && (res != &dummyRes))
		mysql_free_result (res);
}

/** Adds a field definition.
 * \param	c	The connection to use.
 * \param	tbf	The field definition to add.
 * \return	Error code. 0 = no error.	*/
static int
my_add_field_def (DBConnection * dbconn, DBTableDef * tbl)
{
/** \def DESCRIBE_S SQL template string for the DESCRIBE query.	*/
#define DESCRIBE_S	"describe %s"
	DBFieldDef *fld;
	DBConnection_MYSQL *dbmysql;
	MYSQL *co;
	MYSQL_RES *table_info;
	ABuf *query = NULL;
	ABuf *type_str = NULL;
	int i;
	int fieldCount;
	char *fname;
	char *rtype, *rnotnull, *rhasdef;

	if (dbconn == NULL)
	{
		fprintf (stderr,
			 "WARNING: my_add_field_def(): NULL connection\n");
		return -1;
	}

	if (dbconn->pvt == NULL)
	{
		fprintf (stderr,
			 "WARNING: my_add_field_def(): NULL connection\n");
		return -1;
	}

	dbmysql = (DBConnection_MYSQL *) (dbconn->pvt);
	co = dbmysql->conn;
	if (co == NULL)
	{
		fprintf (stderr,
			 "WARNING: my_add_field_def(): NULL connection\n");
		return -1;
	}

	query = buf_check (query,
			   strlen (DESCRIBE_S) +
			   strlen (tabledef_getname (tbl)) + 1);
	if (!query)
	{
		fprintf (stderr,
			 _
			 ("add_field_def() cannot allocate memory for query string\n"));
		return -1;
	}

	snprintf ((char *)query->b_dat, query->b_len, DESCRIBE_S,
		  tabledef_getname (tbl));
	buf_terminate(query);

	if (mysql_query (co, (char *)query->b_dat))
		return -1;

	if ((table_info = mysql_store_result (co)) == NULL)
		return -1;

	fieldCount = mysql_num_rows (table_info);
	tabledef_setfields (tbl, fieldCount);

	for (i = 0; i < fieldCount; i++)
	{
		MYSQL_ROW row;

		mysql_data_seek (table_info, i);
		row = mysql_fetch_row (table_info);

		fname = row[0];
		dbconnection_add_keyword (dbconn, KEYWORD_FIELD, fname);

		rtype = row[1];
		rnotnull = row[2];
		rhasdef = row[4];

		/* allocate type_str */
		type_str = buf_new (256);

		if (rhasdef)
			type_str = buf_check (type_str, strlen (rhasdef));

		type_str = buf_strcpy (type_str, rtype);

		if (strncmp (rnotnull, "YES", 3) == 0)
			type_str = buf_strcat (type_str, " not null");

		if (rhasdef)
			type_str = buf_strcat (type_str, " default");

		fld = (DBFieldDef *)tabledef_getfield (tbl, i);
		fielddef_setname (fld, fname);
		fielddef_settype (fld, (char *)type_str->b_dat);
	}

	buf_free (query);
	buf_free (type_str);

	return 0;
}

/** Gets the table definitions for the passed connection. Stores them in the 
 * connection structure.
 * \param	c	The connection to use.
 * \return	Error code. 0 = no error.	*/
static int
my_DBget_table_def (DBConnection * dbconn)
{
	DBConnection_MYSQL *dbmysql;
	char query[] = "show tables";
	MYSQL_RES *res;

	if (dbconn == NULL)
	{
		fprintf (stderr,
			 _
			 ("WARNING: Could not retrieve table definitions.\n"));
		return -1;
	}

	dbmysql = (DBConnection_MYSQL *) (dbconn->pvt);
	if (dbmysql == NULL)
	{
		fprintf (stderr,
			 "WARNING : Could not get table definitions.\n");
		return -1;
	}

	if (mysql_query (dbmysql->conn, query))
		return -1;

	if ((res = mysql_store_result (dbmysql->conn)) == NULL)
		return -1;
	else
	{
		int i;
		int tables_in_list = mysql_num_rows (res);

		/* Fill the table */
		for (i = 0; i < tables_in_list; i++)
		{
			char *table_name;
			MYSQL_ROW row;
			DBTableDef *tbl;

			tbl = tabledef_new ();
			if (dbconn->_tableDef == NULL)
				dbconn->_tableDef = tbl;
			else
				tabledef_add (dbconn->_tableDef, tbl);

			mysql_data_seek (res, i);
			row = mysql_fetch_row (res);
			table_name = row[0];

			/* Table name */
			tabledef_setname (tbl, table_name);

			/* Table fields */
			if (my_add_field_def (dbconn, tbl) < 0)
			{
				tabledef_freefields (tbl);
				return -1;
			}

			/* Keywords */
			dbconnection_add_keyword (dbconn, KEYWORD_TABLE, table_name);
		}

		mysql_free_result (res);
	}
	return 0;
}

/** Frees the table definition.
 * \param	c	The connection to use.	*/
static void
my_DBfree_table_def (DBConnection * dbconn)
{
	/* Removing Keywords */
	keywords_remove_type (dbconn, KEYWORD_TABLE);
	keywords_remove_type (dbconn, KEYWORD_FIELD);
}

/** Copy the login variables from a text URL to a form.
 * URL format is "mysql://[user[:pass]@]host[:port]/dbname"
 * \param frm	The form to copy to.
 * \param url	The database URL to copy from.	*/
static void
my_DBurl_to_form (Form * frm, const char *url)
{
	char *u;
	char *p;
	char *start;
	char *end;
	char *username = NULL;
	char *password = NULL;
	char *host = NULL;
	char *port = NULL;
	char *database = NULL;

	/* Clear the login parms        */
	frm_clear_fields (frm);

	if (url == NULL)
		return;

	u = strdup (url);
	start = u;

	/* Look for 'mysql://'       */
	p = strstr (start, "mysql://");
	if (p != NULL)
	{
		p += 8;

		/* auth data specified? */
		start = p;
		p = strchr (start, '@');
		if (p != NULL)
		{		/* Yes, set auth data.  */
			*p = '\0';
			end = p;
			username = strdup (start);
			p++;
			start = p;
		}

		/* host part    */
		p = strchr (start, '/');
		if (p == NULL)
		{		/* Rest is all host     */
			host = strdup (start);
		}
		else
		{
			*p = '\0';
			host = strdup (start);
			p++;
			if (strlen (p) > 0)
				database = strdup (p);
		}

		/* Check if password specified  */
		if (username != NULL)
		{
			p = strchr (username, ':');
			if (p != NULL)
			{
				*p = '\0';
				p++;
				password = strdup (p);
			}
		}

		/* Check if port in host part   */
		if (host != NULL)
		{
			p = strchr (host, ':');
			if (p != NULL)
			{
				*p = '\0';
				p++;
				port = strdup (p);
			}
		}
	}

	/* Now, set everything  */
	if (username != NULL)
	{
		frm_set_value_by_name (frm, BASELOGIN, username);
		free (username);
	}

	if (password != NULL)
	{
		frm_set_value_by_name (frm, BASEPASS, password);
		free (password);
	}

	if (host != NULL)
	{
		frm_set_value_by_name (frm, BASEHOST, host);
		free (host);
	}

	if (port != NULL)
	{
		frm_set_value_by_name (frm, BASEPORT, port);
		free (port);
	}

	if (database != NULL)
	{
		frm_set_value_by_name (frm, BASENAME, database);
		free (database);
	}
	free (u);
}

/** Create a new DBConnection, including the MySQL driver private section.
 * \return	The allocated connection.	*/
static DBConnection *
mysql_dbconnection_new ()
{
	DBConnection *dbc = calloc (1, sizeof (DBConnection));
	if (dbc)
		dbc->pvt = calloc (1, sizeof (DBConnection_MYSQL));

	return dbc;
}

/** Template string for database connection error.	*/
#define DB_CONNECT_ERROR_S	"Database connection error : \n %s"
/** Connection callback called by the application to create a new connection.
 * Uses the data in 'dbc' to open a database connection, and returns a new
 * DBConnection * in 'ret'.
 * \param	dbc	The DBConnector to use.
 * \param	ret	The connected DBConnection.
 * \return	Error code.	*/
static int
my_connect_callback (const DBConnector * dbc, DBConnection ** ret, void(*display_error) (const char *, const char *))
{
	const char *bname, *bhost, *bport, *blogin, *bpass;
	ABuf *buf = NULL;
	const char *db_err = NULL;
	MYSQL *my_conn = NULL;
	unsigned int portNo = 0;
	DBConnection *ret_my = NULL;
	DBConnection_MYSQL *ret_my_mysql = NULL;
	int i;

	*ret = NULL;

	if (! dbc)
		return CONNECT_ERROR;

	/* Gets the results */
	bname = frm_get_value_by_name (dbc->db_frm, BASENAME);
	bhost = frm_get_value_by_name (dbc->db_frm, BASEHOST);
	if ((bhost != NULL) && (strlen (bhost) == 0))
		bhost = NULL;

	bport = frm_get_value_by_name (dbc->db_frm, BASEPORT);
	if ((bport != NULL) && (strlen (bport) == 0))
		bport = NULL;

	blogin = frm_get_value_by_name (dbc->db_frm, BASELOGIN);

	if ((blogin != NULL) && (strlen (blogin) == 0))
		blogin = NULL;

	bpass = frm_get_value_by_name (dbc->db_frm, BASEPASS);
	if ((bpass != NULL) && (strlen (bpass) == 0))
		bpass = NULL;

	/* Tests the results */
	if ((my_conn = calloc (1, sizeof (MYSQL))) == NULL)
		return CONNECT_ERROR;

/* NOTE: This symbol is defined in the mysql headers, so no need for a configure check.	*/
#ifdef HAVE_MYSQL_REAL_CONNECT
	/* Thanks to Eike Lange for this code.	*/
	my_conn = mysql_init (my_conn);
	if (bport != NULL)
		portNo = atoi (bport);

	if (!mysql_real_connect
	    (my_conn, bhost, blogin, bpass, NULL, portNo, NULL, 0))
	{
#else
	if (!mysql_connect (my_conn, bhost, blogin, bpass))
	{
#endif
		db_err = mysql_error (my_conn);
		buf = buf_new (strlen (DB_CONNECT_ERROR_S) + strlen (db_err) +
			       1);
		if (!buf) {
			fprintf (stderr, "connect_callback() could not allocate memory for buffer\n");
			return CONNECT_ERROR;
		}

		snprintf ((char *)buf->b_dat, buf->b_len, DB_CONNECT_ERROR_S, db_err);
		buf_terminate(buf);
		MSG("Connection Error", (char *)buf->b_dat);

		buf = buf_free (buf);
		return CONNECT_ERROR;
	}

	if (mysql_select_db (my_conn, bname)) {
		db_err = mysql_error (my_conn);
		buf = buf_check (buf,
				 strlen (DB_CONNECT_ERROR_S) +
				 strlen (db_err) + 1);
		if (!buf)
		{
			fprintf (stderr,
				 _
				 ("connect_callback() couldnt allocate memory for buffer\n"));
			return CONNECT_ERROR;
		}

		snprintf ((char *)buf->b_dat, buf->b_len, DB_CONNECT_ERROR_S, db_err);
		buf_terminate(buf);

		MSG("Connection Error", (char *)buf->b_dat);

		buf = buf_free (buf);
		return CONNECT_ERROR;
	}

	ret_my = mysql_dbconnection_new ();
	*ret = ret_my;

	if (!ret_my)
		return CONNECT_ERROR;

	ret_my->_connector = (DBConnector *)dbc;

	ret_my_mysql = (DBConnection_MYSQL *) (ret_my->pvt);

	/* The connection (private) */
	ret_my_mysql->conn = my_conn;

	/* The database name	*/
	dbconnection_set_dbname(ret_my, bname);

	/* The functions (public) */
	ret_my->DBclear_result = my_DBclear_result;
	ret_my->DBdisconnect = my_DBdisconnect;
	ret_my->DBexecute_query = my_DBexecute_query;
	ret_my->DBfree_table_def = my_DBfree_table_def;
	ret_my->DBget_error_message = my_DBget_error_message;
	ret_my->DBget_field_name = my_DBget_field_name;
	ret_my->DBget_field_count = my_DBget_field_count;
	ret_my->DBget_field_value = my_DBget_field_value;
	ret_my->DBget_query_status = my_DBget_query_status;
	ret_my->DBget_record_count = my_DBget_record_count;
	ret_my->DBget_table_def = my_DBget_table_def;

	/* The keywords (public) */
	ret_my->_keywords = NULL;
	i = 0;
	while (sqlKeywords[i] != NULL)
		dbconnection_add_keyword ((DBConnection *) ret_my, KEYWORD_SQL,
			     sqlKeywords[i++]);

	*ret = ret_my;
	buf = buf_free (buf);
	return CONNECT_OK;
}

/** The name that identifies the MySQL driver.	*/
#define MYSQL_TYPESTR	"MySQL"

/** Register the MySQL driver with GtkSQL.
 * \return	The allocated DBConnector structure.	*/
DBConnector *
register_MySQL (void)
{
	DBConnector *dbconn;
	FormField *fld;

	dbconn = dbconnector_new ();
	if (dbconn == NULL)
		return NULL;

	dbconn->dbtype = MYSQL_TYPESTR;

	/* Prepares all the structures */
	dbconn->db_connect_callback = my_connect_callback;
	dbconn->url_to_form = my_DBurl_to_form;
	dbconn->form_to_url = my_DBform_to_url;
	dbconn->get_driver_info = my_DBget_driver_info;
	dbconn->get_help_url = my_DBget_help_url;
	dbconn->get_help_search_url = my_DBget_help_search_url;
	dbconn->get_service_type = my_DBget_service_type;

	/* Creates the data-structure */
	dbconn->db_frm = frm_new (NULL);
	frm_set_title (dbconn->db_frm, "MySQL");
	fld = frm_add_field (dbconn->db_frm, BASENAME, NULL);
	fld_set_tip (fld, _("Database name (required)"));
	fld_set_relevance (fld, FR_REQUIRED);

	fld = frm_add_field (dbconn->db_frm, BASEHOST, NULL);
	fld_set_tip (fld, _("Host to connect to (optional)"));

	fld = frm_add_field (dbconn->db_frm, BASEPORT, NULL);
	fld_set_tip (fld, _("Port database listens on (optional)"));

	fld = frm_add_field (dbconn->db_frm, BASELOGIN, NULL);
	fld_set_tip (fld, _("Username for connection (optional)"));

	fld = frm_add_field (dbconn->db_frm, BASEPASS, NULL);
	fld_set_type (fld, FT_PWD);
	fld_set_tip (fld, _("Password for connection (optional)"));

	return dbconn;
}

#endif
