/** \file d_sqlite.c	SQLite driver.
 * \author Darryl Luff
 * \ingroup db
 *
 * SQLite is an embedded SQL database. It lets you have an SQL database embedded in
 * your application, so you can create SQL databases without a seperate server. See the
 * software's homepage at http://www.hwaci.com/sw/sqlite/.
 *
 * \par SQLite Connections
 * When you open the connection dialog, you see an SQLite tab listed with the
 * other database tabs. When you select this tab, it has a single input field, a "Browse" button,
 * and a check box. The Browse button lets you connect to a database file.
 * The checkbox lets you select an in-memory database. 
 * This field is for the filename of the database file to open. You can use the
 * 'Browse' button to find it. When you fill in the filename and click Connect,
 * GtkSQL will initialise the Tables list the same as with any other database type.
 * 
 * If you select the "Memory" button, an in-memory database is created. This is a temporary database.
 *
 * \par URL Syntax
 * The SQLite url is of the format: sqlite://filename. To select an in-memory database, use "sqlite://:memory:".
 * So to refer to the file /tmp/report.db, the url would be sqlite:///tmp/report.db
 * Note the three '/'s, two from the 'sqlite://' path, and one from '/tmp...'.
 * 
 * \todo
 * - Implement and test.
 */

/* GtkSQL -- an interactive graphical query tool for PostgreSQL, MySQL and SQLite.
 * Copyright (C) 2003-2005 Darryl Luff
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "../config.h"

#ifdef USE_SQLITE

#include <stdlib.h>

#ifdef SQLITE3
#include <sqlite3.h>
#define sqlite_close sqlite3_close
#define sqlite_exec sqlite3_exec
#define sqlite_get_table sqlite3_get_table
#else
#include <sqlite.h>
#endif

#include <string.h>
#include "d_sqlite_result.h"
#include "guiapi.h"
#include "utils.h"

/** Text for filename input field.	*/
#define SQLITEFILE	"Database File"

#define HELPURL	"http://www.sqlite.org/"

#define MSG(a,b) if (display_error) display_error(a,b); else fprintf(stderr, "%s: %s\n", a, b);

/** The driver-specific database connection structure. This holds and driver-specific data required to manage the connection.	*/
typedef struct {
	/* table fields. These should be in the common part. */
#ifdef SQLITE3
	sqlite3 *db;
#else
	sqlite *db;					/**< The sqlite connection structure.	*/
#endif
	SQLiteResult *result;		/**< The last query result.	*/
	int status;					/**< Status code	*/
	char *errmessage;			/**< Last error message.	*/
} DBConnection_SQLite;

/** The keywords to be added to the keyword-completion table.	*/
static char *sqliteKeywords[] = {
	"ATTACH",
	"BEGIN", "BY",
	"COMMIT", "CONFLICT", "COPY", "CREATE",
	"DATABASE", "DELETE", "DETACH", "DROP",
	"END", "EXPLAIN",
	"FROM",
	"INDEX", "INSERT", "INTO", "IS",
	"NOT", "NULL",
	"ON", "ORDER",
	"PRAGMA",
	"REPLACE", "ROLLBACK",
	"SELECT",
	"TABLE", "TRANSACTION", "TRIGGER",
	"UPDATE",
	"VACUUM", "VALUES", "VIEW",
	"WHERE",
	NULL
};

/** The lastr seen error message on any connection. Dont ever free()
 * this. It is just a copy.	*/
static char *lasterrmsg = NULL;

/** The callback that receives the results from SQLite.
 * \param pArg	The DBConnection pointer.
 * \param argc	The number of columns returned.
 * \param argv	An array of pointers to the result strings.
 * \param columnNames	The names of each column.
 * \return Return non-zero to report an error and abort the query. Otherwise
 * return 0.	*/
static int query_callback(void *pArg, int argc, char **argv, char **columnNames)
{
	DBConnection *dbc;
	DBConnection_SQLite *dbsqlite;
	ResultRow *row;
	int i;

	if (!pArg) {
		fprintf(stderr, "query_callback(). pArg is NULL\n");
		return 1;
	}

	dbc = (DBConnection *) pArg;
	dbsqlite = (DBConnection_SQLite *) dbc->pvt;

	if (!dbsqlite) {
		fprintf(stderr, "query_callback(). dbc->pvt is NULL\n");
		return 1;
	}

	if (!dbsqlite->result)
		return 1;

	if (argv) {
		row = sqliteresult_append_row(dbsqlite->result);
		for (i = 0; i < argc; i++)
			if (argv[i])
				resultrow_append_field(row, argv[i]);
			else
				resultrow_append_field(row, "");
	}
	if (columnNames) {
		sqliteresult_clear_headers(dbsqlite->result);
		for (i = 0; i < argc; i++) {
			if (columnNames[i])
				sqliteresult_append_header(dbsqlite->result, columnNames[i]);
			else
				sqliteresult_append_header(dbsqlite->result, "");
		}
	}

	return 0;
}

/** Set the current error message.
 * \param dbsqlite	The SQLite connection.
 * \param msg	The new error message.	*/
static void set_errmsg(DBConnection_SQLite * dbsqlite, char *msg)
{
	if (!dbsqlite)
		return;

	if (dbsqlite->errmessage) {
		free(dbsqlite->errmessage);
		dbsqlite->errmessage = NULL;
	}

	if (msg)
		dbsqlite->errmessage = strdup(msg);

	lasterrmsg = dbsqlite->errmessage;
}

/** Run a query against the database.
 * \param	dbc	The connection.
 * \param	querystr	The query string. This is in the SQLite query format.	*/
static DBQResult *sqlite_query(DBConnection * dbc, const char *querystr)
{
	DBConnection_SQLite *dbsqlite;
	DBQResult *result;
	int err;
	if (!dbc) {
		fprintf(stderr, "WARNING: sqlite_query() dbc is NULL\n");
		return NULL;
	}

	if (!dbc->pvt) {
		fprintf(stderr, "WARNING: sqlite_query() dbc->pvt is NULL\n");
		return NULL;
	}

	dbsqlite = (DBConnection_SQLite *) (dbc->pvt);
	dbsqlite->status = DB_QUERY_RECORDS_OK;

	if (!dbsqlite->db) {
		dbsqlite->status = DB_QUERY_ERROR;
		return NULL;
	}

	set_errmsg(dbsqlite, NULL);
	dbsqlite->result = sqliteresult_new();
	result = dbresult_new(dbsqlite->result);

	err = sqlite_exec(dbsqlite->db, querystr, query_callback, dbc, &(dbsqlite->errmessage));
	if (err == SQLITE_OK) {
		dbsqlite->status = DB_QUERY_RECORDS_OK;
	}
	else {
		dbsqlite->status = DB_QUERY_ERROR;
		return NULL;
	}

	return result;
}

/** Free the connection structure.
 * \param	c	The connection to free.
 * \return	Always returns a \c NULL pointer.	*/
static DBConnection *sqlite_dbconnection_free(DBConnection * c)
{
	DBConnection_SQLite *dbsqlite = (DBConnection_SQLite *) (c->pvt);

	/* Clear Keywords memory */
	keywords_remove_all(c);

	/* Free structure memory */
	free(dbsqlite);
	free(c);
	return NULL;
}

/** Return the last sqlite error string.
 * \return The error string.	*/
/*static char *sqlite_error()
{
	return lasterrmsg;
}
*/

/** Clears the passed query result.
 * \param	c	The connection.
 * \param	result	Query result to clear.	*/
static void sqlite_DBclear_result(DBConnection * c, DBQResult *result)
{
	DBConnection_SQLite *dbsqlite;
	SQLiteResult *sqlresult;

	sqlresult = dbresult_get_data(result);
	if (! sqlresult)
		return;

	if (! c)
		return;

	dbsqlite = (DBConnection_SQLite *) (c->pvt);
	if (!dbsqlite)
		return;

	/* Clear the result.  */
	sqliteresult_free(sqlresult);
	dbresult_set_data(result, NULL);
}

/** Disconnect from database.
 * \param	c	The database connection to disconnect.
 * \return	Always returns zero.	*/
static int sqlite_DBdisconnect(DBConnection * c)
{
	DBConnection_SQLite *dbsqlite;

	if (c == NULL) {
		fprintf(stderr, "WARNING: sqlit_DBdisconnect(). c is NULL\n");
		return 0;
	}

	if (c->pvt == NULL) {
		fprintf(stderr, "WARNING: sqlit_DBdisconnect(). c->pvt is NULL\n");
		return 0;
	}

	dbsqlite = (DBConnection_SQLite *) (c->pvt);

	/* End connection */
	if (dbsqlite->result)
		dbsqlite->result = sqliteresult_free(dbsqlite->result);

	if (dbsqlite->db) {
		sqlite_close(dbsqlite->db);
		dbsqlite->db = NULL;
	}
	sqlite_dbconnection_free(c);
	return 0;
}

/** Return the error message from the connection.
 * \param	c	The connection.
 * \return	The error string.	*/
static const char *sqlite_DBget_error_message(const DBConnection * c)
{
	DBConnection_SQLite *dbsqlite;

	if (c == NULL) {
		fprintf(stderr, "WARNING: sqlite_DBget_error_message(). c is NULL\n");
		return NULL;
	}

	if (c->pvt == NULL) {
		fprintf(stderr,
				"WARNING: sqlite_DBget_error_message(). c->pvt is NULL\n");
		return NULL;
	}

	dbsqlite = (DBConnection_SQLite *) (c->pvt);

	return dbsqlite->errmessage;
}

/** Send a query to sqlite. The result is stored in the connection.
 * \param	c	The conenction.
 * \param	query	The query string.
 * \return	A pointer to the 'result' structure, or NULL on error.	*/
static DBQResult *sqlite_DBexecute_query(DBConnection * c, const char *query)
{
	return sqlite_query(c, query);
}

/** Return the URL for the SQLite home page.	*/
const char *sqlite_DBget_help_url(const DBConnector *dbc)
{
	return HELPURL;
}

/** Get the status from the returned result.
 * \param	c	The connection.
 * \param	result	The query result.
 * \return	The status code.
 * \see QueryStatus.	*/
static int sqlite_DBget_query_status(const DBConnection * c, const DBQResult *result)
{
	DBConnection_SQLite *dbsqlite;

	if (c == NULL) {
		fprintf(stderr, "WARNING: sqlite_DBget_error_message(). c is NULL\n");
		return DB_QUERY_ERROR;
	}

	if (c->pvt == NULL) {
		fprintf(stderr,
				"WARNING: sqlite_DBget_error_message(). c->pvt is NULL\n");
		return DB_QUERY_ERROR;
	}

	dbsqlite = (DBConnection_SQLite *) (c->pvt);

	return dbsqlite->status;
}

/** Return the number of rows in the result.
 * \param	c	The connection.
 * \param	result	The query result.
 * \return	The number of rows in the result.	*/
static int sqlite_DBget_record_count(const DBConnection * c, const DBQResult *result)
{
	DBConnection_SQLite *dbsqlite;
	SQLiteResult *sqlresult;

	if (c == NULL) {
		fprintf(stderr, "WARNING: sqlite_DBget_record_number(). c is NULL\n");
		return 0;
	}

	if (c->pvt == NULL) {
		fprintf(stderr,
				"WARNING: sqlite_DBget_record_number(). c->pvt is NULL\n");
		return 0;
	}

	dbsqlite = (DBConnection_SQLite *) (c->pvt);
	sqlresult = dbresult_get_data(result);

	if (! sqlresult)
		return 0;

	return sqliteresult_get_row_count(sqlresult);
}

/** Returns the number of fields (columns) in the result.
 * \param	c	The connection.
 * \param	result	The query result to access.
 * \return	The number of columns in the result.	*/
static int sqlite_DBget_field_count(const DBConnection * c, const DBQResult *result)
{
	DBConnection_SQLite *dbsqlite;
	SQLiteResult *sqlresult;

	if (c == NULL) {
		fprintf(stderr, "WARNING: sqlite_DBget_field_number(). c is NULL\n");
		return 0;
	}

	if (c->pvt == NULL) {
		fprintf(stderr,
				"WARNING: sqlite_DBget_field_number(). c->pvt is NULL\n");
		return 0;
	}

	dbsqlite = (DBConnection_SQLite *) (c->pvt);
	sqlresult = dbresult_get_data(result);

	if (!sqlresult)
		return 0;

	return sqliteresult_get_column_count(sqlresult);
}

/** Gets the field-name with the selected index.
 * \param	c	The connection.
 * \param	result	The results to look at.
 * \param	field	Field to return. The index is zero-based.	*/
static const char *sqlite_DBget_field_name(const DBConnection * c, const DBQResult *result, int field)
{
	char *name;
	DBConnection_SQLite *dbsqlite;
	SQLiteResult *sqlresult;

	if (c == NULL) {
		fprintf(stderr, "WARNING: sqlite_DBget_field_name(). c is NULL\n");
		return "";
	}

	if (c->pvt == NULL) {
		fprintf(stderr,
				"WARNING: sqlite_DBget_field_name(). c->pvt is NULL\n");
		return "";
	}

	dbsqlite = (DBConnection_SQLite *) (c->pvt);
	sqlresult = dbresult_get_data(result);
	if (!sqlresult)
		return "";

	name = sqliteresult_get_header_text(sqlresult, field);
	if (name)
		return name;

	return "";
}

/** Returns the value of the passed field.
 * \param	c	The connection.
 * \param	result	The query result to access.
 * \param	record	The record number (zero-based).
 * \param	field	The field number (zero-based).
 * \return	The field value as a text string.	*/
static const char *sqlite_DBget_field_value(const DBConnection * c, const DBQResult *result, int record, int field)
{
	char *value = sqliteresult_get_field((SQLiteResult *)dbresult_get_data(result), record, field);

	if (value == NULL)
		return "";
	else 
		return value;
}

/** Copy the login variables from a text URL to a form.
 * URL format is "sqlite://filename"
 * \param frm	The form to copy to.
 * \param url	The database URL to copy from.	*/
static void sqlite_DBurl_to_form(Form * frm, const char *url)
{
	char *u;
	char *p;
	char *start;

	/* Clear the login parms */
	frm_clear_fields(frm);

	if (url == NULL)
		return;

	u = strdup(url);
	start = u;

	/* Look for '://' */
	p = strstr(start, "://");
	if (p != NULL) {
		p += 3;

		/* Set filename */
		frm_set_value_by_name(frm, SQLITEFILE, p);
	}
	free(u);
}

/** Callback to collect the field list.
 * \param pArg x
 * \param argc x
 * \param argv x
 * \param columnNames	The list to fillin
 * \return Always returns 0.	*/
static int field_def_callback(void *pArg, int argc, char **argv, char **columnNames)
{
	int i;
	DBFieldDef *fld;
	DBTableDef *tbl = (DBTableDef *) pArg;

	if (tbl == NULL) {
		fprintf(stderr, "WARNING: field_def_callback() tbf is NULL\n");
		return 0;
	}

	if (columnNames == NULL) {
		fprintf(stderr, "WARNING: field_def_callback() No column names\n");
		return 0;
	}

	tabledef_freefields(tbl);

	for (i = 0; i < argc; i++) {
		if (columnNames[i] != NULL) {
			fld = (DBFieldDef *)tabledef_addfield(tbl, fielddef_new());
			fielddef_setname(fld, columnNames[i]);
			if (columnNames[argc] != 0)
				fielddef_settype(fld, columnNames[argc + i]);
			else
				fielddef_settype(fld, "? ?");
		}
	}
	return 0;
}

/** Adds the field definitions for a table.
 * \param	c	The connection.
 * \param	tbf	The table definition to add the fields to. The table name is already filled in.
 * \return	Always returns zero.	*/
static int sqlite_add_field_def(DBConnection * dbconn, DBTableDef * tbl)
{
	const DBFieldDef *fld;
	DBConnection_SQLite *dbsqlite;
	int fieldCount;
	ABuf *query = NULL;
	int i;
	int err;

	if (dbconn == NULL) {
		fprintf(stderr,
				"WARNING: sqlite_add_field_def(). DBConnection is NULL\n");
		return -1;
	}

	if (tbl == NULL) {
		fprintf(stderr, "WARNING: sqlite_add_field_def(). tbl is NULL\n");
		return -1;
	}

	tabledef_freefields(tbl);

	dbsqlite = (DBConnection_SQLite *) (dbconn->pvt);
	if (dbsqlite == NULL) {
		fprintf(stderr,
				"WARNING: sqlite_add_field_def(). dbsqlite is NULL\n");
		return -1;
	}

	if (tbl == NULL) {
		fprintf(stderr,
				"WARNING: sqlite_add_field_def(). tbl is NULL\n");
		return -1;
	}

	if (dbsqlite->db == NULL) {
		fprintf(stderr,
				"WARNING: sqlite_add_field_def(). dbsqlite->db is NULL\n");
		return -1;
	}

	/* Make sure we get answers even if the table has no entries */
	err = sqlite_exec(dbsqlite->db,
					  "PRAGMA empty_result_callbacks = ON;", NULL, NULL,
					  NULL);

	/* Get table schema.  */
	query = buf_strcpy(query, "SELECT * from ");
	query = buf_strcat(query, tbl->name);
	query = buf_strcat(query, " LIMIT 1;");

	set_errmsg(dbsqlite, NULL);

	err = sqlite_exec(dbsqlite->db, (char *)query->b_dat, field_def_callback, tbl, &(dbsqlite->errmessage));

	if (err != SQLITE_OK) {
		fprintf(stderr,
				"WARNING: sqlite_add_field_def could not query table schema\n");
		query = buf_free(query);
		return -1;
	}

	/* Set the keyword entries */
	fieldCount = tabledef_fieldcount(tbl);
	for (i = 0; i < fieldCount; i++) {
		fld = tabledef_getfield(tbl, i);
		dbconnection_add_keyword(dbconn, KEYWORD_FIELD, fielddef_getname(fld));
	}

	buf_free(query);
	return 0;
}

/** Gets the table definitions for the passed connection. Table definitions 
 * are stored in the connection structure.
 * \param	c	The conenction.
 * \return	Error code.	*/
static int sqlite_DBget_table_def(DBConnection * dbconn)
{
	int tables_in_list;
	DBConnection_SQLite *dbsqlite;
	DBTableDef *tbl;
	char **result;
	int rows;
	int columns;
	int err;
	int i;
	SQLiteResult *res = NULL;
	char *errmsg = NULL;

	if (dbconn == NULL) {
		fprintf(stderr, "WARNING : Could not get table definitions.\n");
		return -1;
	}

	dbsqlite = (DBConnection_SQLite *) (dbconn->pvt);
	if (dbsqlite == NULL) {
		fprintf(stderr, "WARNING : Could not get table definitions.\n");
		return -1;
	}

	dbconnection_free_table_def(dbconn);

	/* Get list of all tables */
	err = sqlite_get_table(dbsqlite->db,
						   "SELECT name FROM sqlite_master WHERE type = 'table' ORDER BY name ",
						   &result, &rows, &columns, &errmsg);

	if (err != SQLITE_OK)
		return -1;

	/* Allocating table array */
	tables_in_list = rows;

	/* Fill the table */
	for (i = 1; i <= tables_in_list; i++) {
		tbl = tabledef_new();
		if (dbconn->_tableDef == NULL)
			dbconn->_tableDef = tbl;
		else
			tabledef_add(dbconn->_tableDef, tbl);

		/* Table name */
		tabledef_setname(tbl, result[i]);

		/* Table fields */
		if (sqlite_add_field_def(dbconn, tbl) < 0) {
			dbconn->_tableDef = tabledef_free(dbconn->_tableDef);
			return -1;
		}

		/* Keywords */
		dbconnection_add_keyword(dbconn, KEYWORD_TABLE, result[i]);
	}

	if (res)
		sqliteresult_free(res);

	return 0;
}

static ABuf *f2ubuf = NULL;

/**
 * \param frm	The form to copy to.
 * \return The buffer containing the URL.	*/
static const char *sqlite_DBform_to_url(const Form * frm, int showPwd)
{
	const char *s;

	if (frm == NULL)
		return NULL;

	s = frm_get_value_by_name(frm, SQLITEFILE);
	if (s != NULL) {
		f2ubuf = buf_check(f2ubuf, 100);
		f2ubuf = buf_strcat(f2ubuf, "sqlite://");
		f2ubuf = buf_strcat(f2ubuf, s);
	}
	else {
		return NULL;
	}

	return (const char *)f2ubuf->b_dat;
}

/** Frees the table definition associated with the connection.
 * \param	c	The connection.	*/
static void sqlite_DBfree_table_def(DBConnection * dbconn)
{
	if (dbconn == NULL)
		return;

	/* Removing Keywords */
	keywords_remove_type(dbconn, KEYWORD_TABLE);
	keywords_remove_type(dbconn, KEYWORD_FIELD);
}

/** Return the table definition for the requested table.
 * \param	c	The connection.
 * \param	table	The table index (zero-based).
 * \return	A pointer to the table definition.	*/
/*static const DBTableDef *sqlite_DBget_table(const DBConnection * dbconn, int table)
{
	if (dbconn == NULL) {
		fprintf(stderr, "WARNING: sqlite_DBget_table() Connection is NULL\n");
		return NULL;
	}

	return tabledef_gettable(dbconn->tableDef, table);
}
*/

/** Allocate a new SQLite DBConnection.
 * \return	The allocated connection.	*/
static DBConnection *sqlite_dbconnection_new()
{
	DBConnection *dbc = calloc(1, sizeof(DBConnection));
	DBConnection_SQLite *pvtdata;
	int i;
	if (dbc) {
		dbc->DBclear_result = sqlite_DBclear_result;
		dbc->DBdisconnect = sqlite_DBdisconnect;
		dbc->DBexecute_query = sqlite_DBexecute_query;
		dbc->DBfree_table_def = sqlite_DBfree_table_def;
		dbc->DBget_error_message = sqlite_DBget_error_message;
		dbc->DBget_field_name = sqlite_DBget_field_name;
		dbc->DBget_field_count = sqlite_DBget_field_count;
		dbc->DBget_field_value = sqlite_DBget_field_value;
		dbc->DBget_query_status = sqlite_DBget_query_status;
		dbc->DBget_record_count = sqlite_DBget_record_count;
		dbc->DBget_table_def = sqlite_DBget_table_def;

		/* Set up keywords */
		dbc->_keywords = NULL;
		i = 0;
		while (sqliteKeywords[i] != NULL)
			dbconnection_add_keyword((DBConnection *) dbc, KEYWORD_SQL,
						sqliteKeywords[i++]);

		/* Setup private data. */
		pvtdata = calloc(1, sizeof(DBConnection_SQLite));
		dbc->pvt = pvtdata;
	}
	return dbc;
}

/** Set the filename for the connection.
 * \param	dbc	The connection.
 * \param	fn	Filename to set.	*/
static void dbconnection_sqlite_set_filename(DBConnection * dbc, const char *fn)
{
	dbconnection_set_dbname(dbc, fn);
}

/** Open a connection to an XML document.
 * \param	dbc	Thc database connection.
 * \param	fn	Filename to open.
 * \return	Connection status.	*/
static int dbconnection_sqlite_open_file(DBConnection * dbc, const char *fn)
{
	DBConnection_SQLite *dbsqlite;

	if (!dbc)
		return CONNECT_ERROR;

	dbsqlite = (DBConnection_SQLite *) (dbc->pvt);
	if (!dbsqlite)
		return CONNECT_ERROR;

	dbconnection_sqlite_set_filename(dbc, fn);
#ifdef SQLITE3
	if (sqlite3_open(fn, &(dbsqlite->db)) != SQLITE_OK) {
		set_errmsg(dbsqlite, (char *)sqlite3_errmsg(dbsqlite->db));
		if (sqlite3_close(dbsqlite->db) != SQLITE_OK) {
			// XXX should warn here. It means that an SQL statement is still unreleased somewhere
		}
		dbsqlite->db = NULL;
		return CONNECT_ERROR;
	};
#else
	dbsqlite->db = sqlite_open(fn, 0, &(dbsqlite->errmessage));
	
	if (dbsqlite->db == NULL) {
		set_errmsg(dbsqlite, "Database could not be opened.");
		return CONNECT_ERROR;
	}
#endif

	return CONNECT_OK;
}

/** Template for connection error string.	*/
#define DB_CONNECT_ERROR_S	"Database connection error : \n %s"
/** Connection callback called by the application. For the SQLite driver,
 * 'connecting' is really just opening the datbase file.
 * Uses the data in 'dbc' to open the file, and returns a new
 * DBConnection * in 'ret'.
 * \param	dbc	The connection.
 * \param	ret	This is the return value set by the function.
 * \return	Error code.	*/
static int sqlite_connect_callback(const DBConnector * dbc, DBConnection ** ret, void(*display_error) (const char *, const char *))
{
	const char *dbname;
	DBConnection *ret_my;

	/* Gets the database filename */
	dbname = frm_get_value_by_name(dbc->db_frm, SQLITEFILE);
	if ((dbname == NULL) || (strlen(dbname) == 0)) {
		MSG(_("Connection Error"), _("You need to provide a database filename"));

		return CONNECT_ERROR;
	}

	/* Allocate a new connection structure. */
	ret_my = sqlite_dbconnection_new();
	if (!ret_my)
		return CONNECT_ERROR;

	ret_my->_connector = (DBConnector *)dbc;

	/* All OK - set things up */
	if (dbconnection_sqlite_open_file(ret_my, dbname) != CONNECT_OK)
		return CONNECT_ERROR;

	*ret = ret_my;
	return CONNECT_OK;
}

/** Label that identifies the SQLite driver.	*/
#define SQLITE_TYPEKEY	"SQLite"
/** Register the SQLite driver with the application.
 * \return	The SQLite driver DBConnector.	*/
DBConnector *register_SQLite(void)
{
	DBConnector *ret;
	FormField *fld;
	ret = dbconnector_new();
	if (ret == NULL)
		return NULL;
	ret->dbtype = SQLITE_TYPEKEY;

	/* Prepares all the structures */
	ret->db_connect_callback = sqlite_connect_callback;
	ret->url_to_form = sqlite_DBurl_to_form;
	ret->form_to_url = sqlite_DBform_to_url;
	ret->get_help_url = sqlite_DBget_help_url;
	/* ret->get_help_search_url = sqlite_DBget_help_search_url;	*/

	/* Creates the connection form */
	ret->db_frm = frm_new(NULL);
	frm_set_title(ret->db_frm, "SQLite");
	fld = frm_add_field(ret->db_frm, SQLITEFILE, NULL);
	fld_set_relevance(fld, FR_REQUIRED);
	fld_set_type(fld, FT_FILENAME);
	fld_set_tip(fld, "The filename for the SQLite database (required)");
	return ret;
}

#endif
