/** \file d_sqlite_result.c	SQLite driver result structures.
 * \author Darryl Luff
 * \ingroup db
 *
 * SQLite doesnt supply a result structure we can use directly, so we
 * need to build our own.
 *
 * The result structure passed back to GtkSQL is a SQLiteResult. This
 * includes one ResultRow for header names, and another ResultRow
 * pointer to the row data. The rows form a linked list of ResultRow's.
 * Each ResultRow has a pointer to a list of ResultValue's. Each 
 * ResultValue holds a text string that is the field contents.
 */

/* GtkSQL -- an interactive graphical query tool for PostgreSQL, MySQL and SQLite.
 * Copyright (C) 2003-2005 Darryl Luff
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "../config.h"

#ifdef USE_SQLITE

#include <stdlib.h>
#include <string.h>
#include "d_sqlite_result.h"


static ResultValue *resultvalue_new(void)
{
	return calloc(1, sizeof(ResultValue));
}

static ResultValue *resultvalue_free(ResultValue * val)
{
	if (!val)
		return NULL;

	if (val->text)
		free(val->text);

	if (val->next)
		val->next = resultvalue_free(val->next);

	free(val);
	return NULL;
}

static int resultvalue_get_count(ResultValue * val)
{
	if (!val)
		return 0;

	if (!val->next)
		return 1;

	return resultvalue_get_count(val->next) + 1;
}

static ResultValue *resultvalue_append_value(ResultValue * val, ResultValue * n)
{
	if (!val)
		return NULL;

	if (!n)
		return NULL;

	if (val->next)
		resultvalue_append_value(val->next, n);
	else
		val->next = n;

	return n;
}

static ResultValue *resultvalue_get_value(ResultValue * val, int i)
{
	if (!val)
		return NULL;

	if (i < 0)
		return NULL;

	if (i == 0)
		return val;

	if (val->next)
		return resultvalue_get_value(val->next, i - 1);

	return NULL;
}

static void resultvalue_set_value(ResultValue * val, char *s)
{
	if (!val)
		return;

	if (val->text)
		free(val->text);

	if (s)
		val->text = strdup(s);
	else
		val->text = NULL;
}

/*
static void resultrow_set_value(ResultRow * row, int n, char *s)
{
	ResultValue *value;

	if (!row)
		return;

	if (n < 0)
		return;

	if (resultvalue_get_count(row->val) <= n)
		return;

	value = resultvalue_get_value(row->val, n);
	if (value)
		resultvalue_set_value(value, s);
}
*/

static char *resultvalue_get_text(ResultValue * val)
{
	if (!val)
		return NULL;

	return val->text;
}

static ResultRow *resultrow_new()
{
	return calloc(1, sizeof(ResultRow));
}

ResultValue *resultrow_append_field(ResultRow * row, char *txt) {
	ResultValue *val;

	if (!row)
		return NULL;

	val = resultvalue_new();
	if (! val)
		return NULL;

	if (txt)
		resultvalue_set_value(val, txt);	

	if (row->val)
		resultvalue_append_value(row->val, val);
	else
		row->val = val;

	return val;
}

static ResultRow *resultrow_free(ResultRow * row)
{
	if (!row)
		return NULL;

	row->next = resultrow_free(row->next);
	row->val = resultvalue_free(row->val);

	free(row);
	return NULL;
}

static void resultrow_append_row(ResultRow * row, ResultRow * n)
{
	if (!row)
		return;

	if (!n)
		return;

	if (row->next)
		resultrow_append_row(row->next, n);
	else
		row->next = n;
}

static int resultrow_get_count(ResultRow * row)
{
	if (!row)
		return 0;

	if (!row->next)
		return 1;

	return resultrow_get_count(row->next) + 1;
}

static int resultrow_get_field_count(ResultRow * row)
{
	if (!row)
		return 0;

	return resultvalue_get_count(row->val);
}

static ResultRow *resultrow_get_row(ResultRow * row, int n)
{
	if (!row)
		return NULL;

	if (n < 0)
		return NULL;

	if (n == 0)
		return row;

	if (row->next)
		return resultrow_get_row(row->next, n - 1);

	return NULL;
}

static char *resultrow_get_field(ResultRow * row, int n)
{
	ResultValue *value;

	if (!row)
		return NULL;

	if (n < 0)
		return NULL;

	value = resultvalue_get_value(row->val, n);
	return resultvalue_get_text(value);
}

int sqliteresult_get_column_count(SQLiteResult *result) {
	if (! result)
		return 0;

	if (result->rows)
		return resultrow_get_field_count(result->rows);

	if (result->headers)
		return resultrow_get_field_count(result->headers);

	return 0;
}

char *sqliteresult_get_field(SQLiteResult *result, int rowno, int fieldno) {
	ResultRow *row;
	char *value;

	if (! result)
		return NULL;

	row = resultrow_get_row(result->rows, rowno);
	if (! row)
		return NULL;

	value = resultrow_get_field(row, fieldno);

	return value;
}

char *sqliteresult_get_header_text(SQLiteResult *result, int field) {
	if (! result)
		return NULL;

	return resultrow_get_field(result->headers, field);
}

int sqliteresult_get_row_count(SQLiteResult *result) {
	if (! result)
		return 0;

	return resultrow_get_count(result->rows);
}
		
SQLiteResult *sqliteresult_new(void)
{
	return calloc(1, sizeof(SQLiteResult));
}

void sqliteresult_clear_headers(SQLiteResult * res)
{
	if (! res)
		return;

	res->headers = resultrow_free(res->headers);
}

SQLiteResult *sqliteresult_free(SQLiteResult * res)
{
	if (!res)
		return NULL;

	res->headers = resultrow_free(res->headers);
	res->rows = resultrow_free(res->rows);

	free(res);
	return NULL;
}

ResultRow *sqliteresult_append_row(SQLiteResult * res)
{
	ResultRow *row;
	if (!res)
		return NULL;

	row = resultrow_new();
	if (res->rows)
		resultrow_append_row(res->rows, row);
	else
		res->rows = row;

	return row;
}

ResultValue *sqliteresult_append_header(SQLiteResult * res, char *s)
{
	if (!res)
		return NULL;

	if (!(res->headers))
		res->headers = resultrow_new();

	return resultrow_append_field(res->headers, s);
}

#endif
