/** \file d_unixodbc.c	Forms version of unixODBC driver.
 * \author Darryl Luff
 * \ingroup db
 *
 * This is the database driver linking unixODBC with the GtkSQL application.
 *
 * The ODBC API is very wordy. More lines of code are needed than with other
 * interfaces. The driver also needs to keep track of a lot more handles and
 * things than with other API's. This of course means that it is more vulnerable 
 * to bugs and other problems.
 *
 * The URL form at used is unixodbc://[user[:password]]@datasource/
 *
 * \section unixodbc_overview ODBC Overview
 *
 * ODBC forces the application to keep track of a lot of data handles. First is
 * an environment handle, which is global and only one is needed.
 *
 * Using the environment handle, a connection handle is allocated. This is used 
 * to create the connection.
 *
 * To send a query, a statement handle is allocated using the connection handle.
 * The result data is associated with the statement handle.
 *
 * \section unixodbc_implementation ODBC Driver Implementation
 * The driver will leave the data in the statement handle for now. A statement 
 * handle is allocated and returned as the 'result' parameter. 
 *
 * \section unixodbc_todo ODBC Driver TODO
 * - Method to extract table list from database on connection.
 * - Extract table metadata
 * - Make sure random access to data rows/columns can be done. If not, copy data to local structure. 
 * 
 */
/* GtkSQL -- an interactive graphical query tool.
 * Copyright (C) 1998-2005 Lionel ULMER, Darryl Luff
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "../config.h"

#ifdef USE_UNIXODBC
#include <sql.h>
#include <sqlext.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "buf.h"
#include "db_connectors.h"
#include "db_tabledef.h"
#include "dbapi.h"
#include "guiapi.h"
#include "utils.h"

/* The text tags to use as the form field names.	*/

/** Database name.	*/
#define BASEDSN	_("Datasource")

/** Database user name.	*/
#define BASELOGIN	_("Username")

/** Database password.	*/
#define BASEPASS	_("Password")

#define MAX_DATA 4096

/** The unixODBC-specific database connection structure.	*/
typedef struct {
	RETCODE rc;		/**< ODBC return code	*/
	HENV env;		/**< Environment handle	*/
	HDBC dbc;		/**< Connection handle	*/
	char *dsn;		/**< Data source name	*/
	char *errmsg;	/**< Last error message	*/
} DBConnection_UNIXODBC;

/** A nasty hack: if a unixODBC query succeeded, but returned a NULL result 
 * (e.g. after an UPDATE), we return the address of dummy_res from 
 * DBexecute_query().  This is so we can distinguish a successful query
 * which returned nothing from a failed query in DBget_query_status(). */
//UNIXODBC_RES dummyRes;

/** The unixODBC keywords to be added to the keyword-completion table.	*/
/* XXX need to update these for the ODBC driver capabilities. Which is a bit hard, 
 * because there is no way to know what the actual driver is. */
static char *sqlKeywords[] = {
	"ABS", "AND", "ALTER", "ASC", "ASCII",
	"BEGIN", "BETWEEN", "BIN", "BINARY", "BIT_LENGTH", "BY",
	"CASE", "CEILING", "CHAR", "CONCAT", "COUNT", "CREATE",
	"DATABASE", "DELETE", "DESCRIBE", "DISTINCT", "DISTINCTROW",
	"DROP",
	"END", "EXP", "EXPLAIN",
	"FLOOR", "FROM",
	"GRANT", "GROUP",
	"HAVING", "HEX",
	"IFNULL", "INDEX", "INNER", "INSERT", "INSTR", "INTERVAL", "INTO", "ISNULL",
	"JOIN",
	"LCASE", "LEFT", "LENGTH", "LIKE", "LOCATE",
	"LOCK", "LOG", "LOG2", "LOG10", "LOWER", "LPAD", "LTRIM",
	"MATCH", "MAX", "MID", "MIN", "MEAN", "MOD",
	"NATURAL", "NOT", "NULL", "NULLIF",
	"OCT", "OR", "ORD", "ORDER", "OUTER",
	"REPEAT", "REPLACE", "RESET", "REVERSE", "REVOKE", "RIGHT",
	"ROUND", "RPAD", "RTRIM",
	"SELECT", "SET", "SHOW", "SIGN", "SOUNDEX", "SPACE", "SQRT", "STRCMP",
	"SUBSTRING", "SUM",
	"TABLE", "THEN", "TRANSACTION",
	"UCASE", "UNION", "UPDATE", "UPPER",
	"WHEN", "WHERE", "WORK",
	NULL
};

static void unixodbc_set_errmsg(DBConnection_UNIXODBC * dbunixodbc, char *msg)
{
	if (!dbunixodbc)
		return;

	if (dbunixodbc->errmsg) {
		free(dbunixodbc->errmsg);
		dbunixodbc->errmsg = NULL;
	}

	if (msg)
		dbunixodbc->errmsg = strdup(msg);
}

static int unixodbc_check_error(DBConnection_UNIXODBC * unixodbc)
{
	char errmsg[1024];

	if (!unixodbc)
		return 1;

	switch (unixodbc->rc) {
	case SQL_NO_DATA:
		unixodbc_set_errmsg(unixodbc, "No data");
		return SQL_NO_DATA;
	case SQL_SUCCESS:
	case SQL_SUCCESS_WITH_INFO:
		unixodbc_set_errmsg(unixodbc, NULL);
		return SQL_SUCCESS;
	case SQL_ERROR:
		SQLError(unixodbc->env, unixodbc->dbc, SQL_NULL_HSTMT, NULL, NULL,
				 errmsg, sizeof(errmsg), NULL);

		printf("ERROR: %s\n", errmsg);
		unixodbc_set_errmsg(unixodbc, errmsg);
		return SQL_ERROR;
	case SQL_INVALID_HANDLE:
		printf("ERROR: SQL_INVALID_HANDLE\n");
		unixodbc_set_errmsg(unixodbc, "ERROR: SQL_INVALID_HANDLE\n");
		return SQL_INVALID_HANDLE;
	}
	unixodbc_set_errmsg(unixodbc, NULL);
	return unixodbc->rc;
}

/** Free a DBConnection and it's related DBConnection_UNIXODBC.
 * \param	dbconn	The DBConnection to free.	*/
static void unixodbc_dbconnection_free(DBConnection * dbconn)
{
	DBConnection_UNIXODBC *dbunixodbc;

	if (!dbconn)
		return;

	dbunixodbc = (DBConnection_UNIXODBC *) (dbconn->pvt);
	if (!dbunixodbc)
		return;

	/* Clear Keywords memory */
	keywords_remove_all(dbconn);

	/* Free structure memory */
	free(dbunixodbc);
	free(dbconn);
}

/** Disconnect from database.
 * \param	c	The \ref DBConnection to disconnect.
 * \return	Always returns zero.	*/
static int unixodbc_DBdisconnect(DBConnection * dbconn)
{
	DBConnection_UNIXODBC *dbunixodbc;
	if (!dbconn)
		return 0;

	dbunixodbc = (DBConnection_UNIXODBC *) (dbconn->pvt);
	if (!dbunixodbc)
		return 0;

	/* Deallocate handles */
	if (dbunixodbc->dbc) {
		SQLDisconnect(dbunixodbc->dbc);
		SQLFreeHandle(SQL_HANDLE_DBC, dbunixodbc->dbc);
		dbunixodbc->dbc = 0;
	}

	if (dbunixodbc->env) {
		SQLFreeHandle(SQL_HANDLE_ENV, dbunixodbc->env);
		dbunixodbc->env = 0;
	}
	unixodbc_dbconnection_free(dbconn);
	return 0;
}

/** Return the error message from the connection.
 * \param	c	The \ref DBConnection to reference.
 * \return	The last error message on this connection.	*/
static const char *unixodbc_DBget_error_message(const DBConnection * dbconn)
{
	DBConnection_UNIXODBC *dbunixodbc;
	static char errmsg[MAX_DATA];
	if (!dbconn)
		return "Invalid connection";

	dbunixodbc = (DBConnection_UNIXODBC *) (dbconn->pvt);
	if (!dbunixodbc)
		return "Private data invalid";

	switch (dbunixodbc->rc) {
	case SQL_ERROR:
		SQLError(dbunixodbc->env, dbunixodbc->dbc, SQL_NULL_HSTMT, NULL, NULL,
				 errmsg, sizeof(errmsg), NULL);
		return errmsg;
		break;
	case SQL_INVALID_HANDLE:
		return "SQL_INVALID_HANDLE\n";
	case SQL_SUCCESS:
	case SQL_SUCCESS_WITH_INFO:
		return SQL_SUCCESS;
	}
	return "";
}

/** Send a query to unixodbc. If the returned result is \c NULL, and the number of
 * fields is zero, returns a pointer to the dummy result \ref dummy_res.
 * \param	c	The connection to send the query on.
 * \param	query	The query text.	*/
static DBQResult *unixodbc_DBexecute_query(DBConnection * dbconn, const char *query)
{
	DBConnection_UNIXODBC *dbunixodbc;
	HSTMT *stmt = NULL;

	fprintf(stderr, "XXX WARNING: unixodbc_DBexecute_query() not properly implemented.\n");
	if (!dbconn)
		return NULL;

	dbunixodbc = (DBConnection_UNIXODBC *) (dbconn->pvt);
	if (!dbunixodbc)
		return NULL;

	/* Allocate a new statement */
	stmt = calloc(1, sizeof(stmt));
	if (! stmt)
		return NULL;

	dbunixodbc->rc = SQLAllocHandle(SQL_HANDLE_STMT, dbunixodbc->dbc, stmt);
	if (unixodbc_check_error(dbunixodbc) != 0) {
		printf("Error allocating statement handle\n");
		return NULL;
	}

	/* Run the statement */
	dbunixodbc->rc = SQLExecDirect(*stmt, (char *)query, SQL_NTS);
	if (unixodbc_check_error(dbunixodbc) != 0) {	/* Error */
		/* Deallocate handles and disconnect. */
		printf("ERROR: %d\n", dbunixodbc->rc);
		SQLFreeStmt(*stmt, SQL_DROP);
		free(stmt);
	}
	else {
		/* Query OK. Process results. */
		/* ODBC doesn't seem to have any real way to store the results, which 
		 * is a pain. */
		char buff[2048];
		SQLINTEGER dataLen;

		while ((dbunixodbc->rc = SQLFetch(*stmt)) == SQL_SUCCESS) {
			SQLGetData(*stmt, 1, SQL_C_CHAR, buff, sizeof(buff), &dataLen);

			/* Normally the SQLBindCol() ODBC API could
			 * be called to bind individual rows of data and assign for a
			 * rowset. */
			printf("[ODBC]%s\n", buff);
			unixodbc_check_error(dbunixodbc);
		}
	}
	return NULL;
}

static ABuf *f2ubuf = NULL;

/**
 * \param frm	The form to copy to.
 * \return The buffer containing the URL.	*/
const char *unixodbc_DBform_to_url(const Form * frm)
{
	const char *s;

	if (frm == NULL)
		return NULL;

	f2ubuf = buf_check(f2ubuf, 100);
	f2ubuf = buf_strcpy(f2ubuf, frm_get_title(frm));
	f2ubuf = buf_strcat(f2ubuf, "://");
	s = frm_get_value_by_name(frm, BASELOGIN);
	if (s != NULL) {
		f2ubuf = buf_strcat(f2ubuf, s);
		s = frm_get_value_by_name(frm, BASEPASS);
		if (s != NULL) {
			f2ubuf = buf_strcat(f2ubuf, ":");
			f2ubuf = buf_strcat(f2ubuf, s);
		}
		f2ubuf = buf_strcat(f2ubuf, "@");
	}
	s = frm_get_value_by_name(frm, BASEDSN);
	if (s == NULL)
		s = "DEFAULT";

	f2ubuf = buf_strcat(f2ubuf, s);
	f2ubuf = buf_strcat(f2ubuf, "/");

	return f2ubuf->b_dat;
}

/** Return the number of rows in the result.
 * \param	c	The connection to use.
 * \param	result	The result data.
 * \return	The number of result rows.	*/
static int unixodbc_DBget_record_count(const DBConnection * dbconn, const DBQResult *result)
{
	DBConnection_UNIXODBC *dbunixodbc;
	HSTMT stmt;
	SQLINTEGER count;

	if (!dbconn)
		return 0;

	dbunixodbc = (DBConnection_UNIXODBC *) dbconn->pvt;
	if (!dbunixodbc)
		return 0;

	if (! result)
		return 0;

	stmt = *((HSTMT *)result);

	/* XXX SQLRowCount is only supposed to return the number of affected rows 
	 * for a query that updates the database, NOT the number of rows in the 
	 * return dataset. So this needs to be redone. ODBC drivers do not all 
	 * support this function.	*/
	dbunixodbc->rc = SQLRowCount(stmt, &count);
	if (unixodbc_check_error(dbunixodbc) != 0) {
		return 0;
	}

	return count;
}

/** Get the status from the returned result.
 * \param	c	The connection to operate on.
 * \param	result	The result pointer.
 * \return	The status code.
 * \see QueryStatus.	*/
static int unixodbc_DBget_query_status(const DBConnection * dbconn, const DBQResult *result)
{
	DBConnection_UNIXODBC *dbunixodbc;

	if (!dbconn)
		return -1;

	dbunixodbc = (DBConnection_UNIXODBC *) (dbconn->pvt);
	if (!dbunixodbc)
		return -1;

	if (unixodbc_check_error(dbunixodbc)) {
		return DB_QUERY_ERROR;
	}
	// else if (res == &dummyRes) {
	// return DB_QUERY_COMMAND_OK;
	// }
	else if (unixodbc_DBget_record_count(dbconn, result) == 0) {
		return DB_QUERY_EMPTY;
	}
	else {
		return DB_QUERY_RECORDS_OK;
	}
}

/** Returns the number of fields (columns) in the result.
 * \param	c	The connection to use.
 * \param	result	The result data.
 * \return	The number of result columns.	*/
static int unixodbc_DBget_field_count(const DBConnection * dbconn, const DBQResult *result)
{
	DBConnection_UNIXODBC *dbunixodbc;
	HSTMT stmt;
	SQLSMALLINT count;

	if (!dbconn)
		return 0;

	dbunixodbc = (DBConnection_UNIXODBC *) dbconn->pvt;
	if (!dbunixodbc)
		return 0;

	if (! result)
		return 0;

	stmt = *((HSTMT*) result);

	dbunixodbc->rc = SQLNumResultCols(stmt, &count);
	if (unixodbc_check_error(dbunixodbc) != 0) {
		return 0;
	}

	return count;
}

/** Return the field-name for the selected field.
 * \param	c	The connection to use.
 * \param	result	The query results.
 * \param	field	The field number. This is a zero-based index.	*/
static const char *unixodbc_DBget_field_name(const DBConnection * dbconn,
											 const DBQResult *result, int field)
{
	DBConnection_UNIXODBC *dbunixodbc;
	HSTMT stmt;
	static SQLCHAR colName[1024];
	SQLSMALLINT nameLength;

	if (!dbconn)
		return NULL;

	dbunixodbc = (DBConnection_UNIXODBC *) dbconn->pvt;
	if (!dbunixodbc)
		return NULL;

	if (! result)
		return NULL;

	stmt = *((HSTMT *) result);

	dbunixodbc->rc =
		SQLColAttribute(stmt, field + 1, SQL_DESC_LABEL, &colName,
						sizeof(colName), &nameLength, NULL);
	if (unixodbc_check_error(dbunixodbc) != 0) {
		return 0;
	}

	return colName;
}

/** Returns the value of the selected field.
 * \param	c	The connection to use.
 * \param	result	The query results.
 * \param	record	The record number to access (zero-based).
 * \param	field	The field number to access (zero-based).
 * \return	The field value as a string.	*/
static const char *unixodbc_DBget_field_value(const DBConnection * dbconn,
											  const DBQResult *result, int record,
											  int field)
{
	DBConnection_UNIXODBC *dbunixodbc;
	HSTMT stmt;
	SQLINTEGER dataLength;
	static SQLCHAR dat[1024];

	if (!dbconn)
		return 0;

	dbunixodbc = (DBConnection_UNIXODBC *) dbconn->pvt;
	if (!dbunixodbc)
		return 0;

	if (! result)
		return NULL;

	stmt = *((HSTMT *) result);

	dbunixodbc->rc = SQLFetchScroll(stmt, SQL_FETCH_ABSOLUTE, record + 1);
	if (unixodbc_check_error(dbunixodbc) != 0)
		return "";

	dbunixodbc->rc =
		SQLGetData(stmt, field + 1, SQL_C_CHAR, &dat, sizeof(dat),
				   &dataLength);
	if (unixodbc_check_error(dbunixodbc) != 0)
		return "";

	return dat;
}

/** Clears the result.
 * \param	c	The connection to use.
 * \param	result	The result set to clear.	*/
static void unixodbc_DBclear_result(DBConnection * dbconn, DBQResult *result)
{
	HSTMT stmt;

	if (result) {
		stmt = *((HSTMT *)result);
		SQLFreeHandle(SQL_HANDLE_STMT, stmt);
		free(result);
	}
}

/** Adds a field definition.
 * \param	c	The connection to use.
 * \param	tbf	The field definition to add.
 * \return	Error code. 0 = no error.	*/
static int unixodbc_add_field_def(DBConnection * dbconn, DBTableDef * tbl)
{
#define STR_LEN 128+1
#define REM_LEN 254+1
	DBConnection_UNIXODBC *dbunixodbc;
	DBFieldDef *fld;
	HSTMT stmt;
	ABuf *type_str = NULL;

	if (!dbconn)
		return 0;

	dbunixodbc = (DBConnection_UNIXODBC *) dbconn->pvt;
	if (!dbunixodbc)
		return 0;

	/* Allocate a statement to use for the query */
	dbunixodbc->rc = SQLAllocHandle(SQL_HANDLE_STMT, dbunixodbc->dbc, &stmt);
	if (unixodbc_check_error(dbunixodbc) != 0) {
		printf("Error allocating statement handle\n");
		return -1;
	}

	/* Declare buffers for result set data */
	SQLCHAR szCatalog[STR_LEN], szSchema[STR_LEN];
	SQLCHAR szTableName[STR_LEN], szColumnName[STR_LEN];
	SQLCHAR szTypeName[STR_LEN], szRemarks[REM_LEN];
	SQLCHAR szColumnDefault[STR_LEN], szIsNullable[STR_LEN];
	SQLINTEGER ColumnSize, BufferLength, CharOctetLength, OrdinalPosition;
	SQLSMALLINT DataType, DecimalDigits, NumPrecRadix, Nullable;
	SQLSMALLINT SQLDataType, DatetimeSubtypeCode;

	/* Declare buffers for bytes available to return */
	SQLINTEGER cbCatalog, cbSchema, cbTableName, cbColumnName;
	SQLINTEGER cbDataType, cbTypeName, cbColumnSize, cbBufferLength;
	SQLINTEGER cbDecimalDigits, cbNumPrecRadix, cbNullable, cbRemarks;
	SQLINTEGER cbColumnDefault, cbSQLDataType, cbDatetimeSubtypeCode,
		cbCharOctetLength;
	SQLINTEGER cbOrdinalPosition, cbIsNullable;

	dbunixodbc->rc = SQLColumns(stmt, NULL, 0,	/* All catalogs */
								NULL, 0,	/* All schemas */
								(char *)tabledef_getname(tbl), SQL_NTS,	/* CUSTOMERS
																 * table */
								NULL, 0);	/* All columns */
	if (unixodbc_check_error(dbunixodbc) != 0)
		return -1;

	/* Bind columns in result set to buffers */
	SQLBindCol(stmt, 1, SQL_C_CHAR, szCatalog, STR_LEN, &cbCatalog);
	SQLBindCol(stmt, 2, SQL_C_CHAR, szSchema, STR_LEN, &cbSchema);
	SQLBindCol(stmt, 3, SQL_C_CHAR, szTableName, STR_LEN, &cbTableName);
	SQLBindCol(stmt, 4, SQL_C_CHAR, szColumnName, STR_LEN, &cbColumnName);
	SQLBindCol(stmt, 5, SQL_C_SSHORT, &DataType, 0, &cbDataType);
	SQLBindCol(stmt, 6, SQL_C_CHAR, szTypeName, STR_LEN, &cbTypeName);
	SQLBindCol(stmt, 7, SQL_C_SLONG, &ColumnSize, 0, &cbColumnSize);
	SQLBindCol(stmt, 8, SQL_C_SLONG, &BufferLength, 0, &cbBufferLength);
	SQLBindCol(stmt, 9, SQL_C_SSHORT, &DecimalDigits, 0, &cbDecimalDigits);
	SQLBindCol(stmt, 10, SQL_C_SSHORT, &NumPrecRadix, 0, &cbNumPrecRadix);
	SQLBindCol(stmt, 11, SQL_C_SSHORT, &Nullable, 0, &cbNullable);
	SQLBindCol(stmt, 12, SQL_C_CHAR, szRemarks, REM_LEN, &cbRemarks);
	SQLBindCol(stmt, 13, SQL_C_CHAR, szColumnDefault, STR_LEN,
			   &cbColumnDefault);
	SQLBindCol(stmt, 14, SQL_C_SSHORT, &SQLDataType, 0, &cbSQLDataType);
	SQLBindCol(stmt, 15, SQL_C_SSHORT, &DatetimeSubtypeCode, 0,
			   &cbDatetimeSubtypeCode);
	SQLBindCol(stmt, 16, SQL_C_SLONG, &CharOctetLength, 0,
			   &cbCharOctetLength);
	SQLBindCol(stmt, 17, SQL_C_SLONG, &OrdinalPosition, 0,
			   &cbOrdinalPosition);
	SQLBindCol(stmt, 18, SQL_C_CHAR, szIsNullable, STR_LEN, &cbIsNullable);

	dbunixodbc->rc = SQLFetch(stmt);
	int i = 0;
	while (unixodbc_check_error(dbunixodbc) == 0) {
		if (szColumnName && (strlen(szColumnName) > 0)) {
			dbconnection_add_keyword(dbconn, KEYWORD_FIELD, szColumnName);

			/* allocate type_str */
			type_str = buf_new(256);
			if (szTypeName)
				type_str = buf_strcpy(type_str, szTypeName);

			if (!Nullable)
				type_str = buf_strcat(type_str, " not null");

			fld = (DBFieldDef *)tabledef_getfield(tbl, i);
			fielddef_setname(fld, szColumnName);
			fielddef_settype(fld, type_str->b_dat);
			i++;
		}

		dbunixodbc->rc = SQLFetch(stmt);
	}

	buf_free(type_str);

	return 0;
}

/** Gets the table definitions for the passed connection. Stores them in the 
 * connection structure.
 * \param	c	The connection to use.
 * \return	Error code. 0 = no error.	*/
static int unixodbc_DBget_table_def(DBConnection * dbconn)
{
#define STR_LEN 128+1
#define REM_LEN 254+1
	DBConnection_UNIXODBC *dbunixodbc;
	HSTMT tablestmt;

	if (dbconn == NULL) {
		fprintf (stderr,  _("WARNING: Could not retrieve table definitions.\n"));
		return -1;
	}

	dbunixodbc = (DBConnection_UNIXODBC *) (dbconn->pvt);
	if (dbunixodbc == NULL)
	{
		fprintf (stderr, "WARNING : Could not get table definitions.\n");
		return -1;
	}

	/* Allocate a statement to use for the query */
	dbunixodbc->rc = SQLAllocHandle(SQL_HANDLE_STMT, dbunixodbc->dbc, &tablestmt);
	if (unixodbc_check_error(dbunixodbc) != 0) {
		printf("Error allocating statement handle\n");
		return -1;
	}

	/* Declare buffers for result set data. Isn't ODBC painfull?	*/
	SQLCHAR catalogName[STR_LEN];
	SQLLEN catalogNameLen = 0;
	SQLCHAR schemaName[STR_LEN];
	SQLLEN schemaNameLen = 0;
	SQLCHAR tableName[STR_LEN];
	SQLLEN tableNameLen = 0;
	SQLCHAR tableType[STR_LEN];
	SQLLEN tableTypeLen = 0;
	SQLCHAR remarks[STR_LEN];
	SQLLEN remarksLen = 0;

	dbunixodbc->rc = SQLTables(tablestmt, NULL, 0, NULL, 0, "%", SQL_NTS, "'TABLE'", SQL_NTS);
	if (unixodbc_check_error(dbunixodbc) != 0)
		return -1;

	SQLSMALLINT count;

	dbunixodbc->rc = SQLNumResultCols(tablestmt, &count);

	/* Bind columns in result set to buffers */
	SQLBindCol(tablestmt, 1, SQL_C_CHAR, catalogName, STR_LEN, &catalogNameLen);
	SQLBindCol(tablestmt, 2, SQL_C_CHAR, schemaName, STR_LEN, &schemaNameLen);
	SQLBindCol(tablestmt, 3, SQL_C_CHAR, tableName, STR_LEN, &tableNameLen);
	SQLBindCol(tablestmt, 4, SQL_C_CHAR, tableType, STR_LEN, &tableTypeLen);
	SQLBindCol(tablestmt, 5, SQL_C_CHAR, &remarks, STR_LEN, &remarksLen);

	dbunixodbc->rc = SQLFetch(tablestmt);
	while (SQL_SUCCEEDED(dbunixodbc->rc)) {
		if (tableNameLen >= 1) {
			DBTableDef *tbl;

			tbl = tabledef_new();
			if (dbconn->_tableDef == NULL)
				dbconn->_tableDef = tbl;
			else
				tabledef_add(dbconn->_tableDef, tbl);

			/* Table name */
			tabledef_setname(tbl, tableName);

			/* Keywords */
			dbconnection_add_keyword(dbconn, KEYWORD_TABLE, tableName);

			/* Table fields */
			if (unixodbc_add_field_def(dbconn, tbl) < 0) {
				tabledef_freefields(tbl);
				return -1;
			}

			/* Now, get field info	*/
			fprintf(stderr, "XXX UnixODBC driver does not yet extract field information!\n");			
		}
		dbunixodbc->rc = SQLFetch(tablestmt);
	}
	dbunixodbc->rc = SQLFreeHandle(SQL_HANDLE_STMT, tablestmt);

	return 0;
}

/** Frees the table definition.
 * \param	c	The connection to use.	*/
static void unixodbc_DBfree_table_def(DBConnection * dbconn)
{
	dbconn->_tableDef  = tabledef_free(dbconn->_tableDef );

	/* Removing Keywords */
	keywords_remove_type(dbconn, KEYWORD_TABLE);
	keywords_remove_type(dbconn, KEYWORD_FIELD);
}

/** Return the number of tables that exist in the open database.
 * \param	c	The connection to use.
 * \return	Error code. 0 = no error.	*/
/*static int unixodbc_DBget_table_count(const DBConnection * dbconn)
{
	if (dbconn == NULL)
		return 0;

	if (dbconn->_tableDef  == 0)
		return 0;

	return tabledef_count(dbconn->_tableDef );
}
*/

/** Copy the login variables from a text URL to a form.
 * URL format is "unixodbc://[user[:pass]@]host[:port]/dbname"
 * \param frm	The form to copy to.
 * \param url	The database URL to copy from.	*/
static void unixodbc_DBurl_to_form(Form * frm, const char *url)
{
	char *u;
	char *p;
	char *start;
	char *end;
	char *username = NULL;
	char *password = NULL;
	char *dsn = NULL;

	frm_clear_fields(frm);

	if ((! frm) || (! url))
		return;

	u = strdup(url);
	start = u;
	if (strncmp(start, "unixodbc", 8) != 0)
		return;

	/* Look for 'unixodbc://' */
	p = strstr(start, "unixodbc://");
	if (p != NULL) {
		p += 11;

		/* auth data specified? */
		start = p;
		p = strchr(start, '@');
		if (p != NULL) {			   /* Yes, set auth data.  */
			*p = '\0';
			end = p;
			username = strdup(start);
			p++;
			start = p;
		}

		/* host part */
		p = strchr(start, '/');
		if (p == NULL) {			   /* Rest is all host */
			dsn = strdup(start);
		}
		else {
			*p = '\0';
			dsn = strdup(start);
			p++;
		}

		/* Check if password specified */
		if (username != NULL) {
			p = strchr(username, ':');
			if (p != NULL) {
				*p = '\0';
				p++;
				password = strdup(p);
			}
		}
	}

	/* Now, set everything */
	if (username != NULL) {
		frm_set_value_by_name(frm, BASELOGIN, username);
		free(username);
	}

	if (password != NULL) {
		frm_set_value_by_name(frm, BASEPASS, password);
		free(password);
	}

	if (dsn != NULL) {
		frm_set_value_by_name(frm, BASEDSN, dsn);
		free(dsn);
	}

	free(u);
}

/** Create a new DBConnection, including the unixODBC driver private section.
 * \return	The allocated connection.	*/
static DBConnection *unixodbc_dbconnection_new()
{
	DBConnection *dbc = calloc(1, sizeof(DBConnection));
	if (dbc)
		dbc->pvt = calloc(1, sizeof(DBConnection_UNIXODBC));

	return dbc;
}

static void unixodbc_dbconnection_set_dsn(DBConnection_UNIXODBC * dbunixodbc,
										  const char *dsn)
{
	if (!dbunixodbc)
		return;

	if (dbunixodbc->dsn) {
		free(dbunixodbc->dsn);
		dbunixodbc->dsn = NULL;
	}

	if (dsn)
		dbunixodbc->dsn = strdup(dsn);
}

/* Allocate environment handle, allocate connection handle,
 * connect to data source, and allocate statement handle. */
static int unixodbc_do_connect(DBConnection_UNIXODBC * dbunixodbc, const char *dsn,
							   const char *user, const char *pwd)
{
	unsigned int dsnlen;
	unsigned int pwdlen;
	unsigned int userlen;

	if (!dbunixodbc)
		return -1;

	if (!dbunixodbc->env) {
		dbunixodbc->rc =
			SQLAllocHandle(SQL_HANDLE_ENV, SQL_NULL_HANDLE, &dbunixodbc->env);
		if (unixodbc_check_error(dbunixodbc) != 0) {
			printf("Error: SQL_HANDLE_ENV: %d\n\n", dbunixodbc->rc);
			return dbunixodbc->rc;
		}

		dbunixodbc->rc =
			SQLSetEnvAttr(dbunixodbc->env, SQL_ATTR_ODBC_VERSION,
						  (SQLPOINTER) SQL_OV_ODBC3, 0);
		if (unixodbc_check_error(dbunixodbc) != 0) {
			printf("Error: SQL_ATTR_ODBC_VERSION: %d\n\n", dbunixodbc->rc);

			return dbunixodbc->rc;
		}
	}

	/* start connection */
	dbunixodbc->rc =
		SQLAllocHandle(SQL_HANDLE_DBC, dbunixodbc->env, &dbunixodbc->dbc);
	if (unixodbc_check_error(dbunixodbc) != 0) {
		printf("Error: SQL_HANDLE_DBC: %d\n", dbunixodbc->rc);

		return dbunixodbc->rc;
	}

	if (dsn)
		dsnlen = strlen(dsn);
	else
		dsnlen = 0;

	if (user)
		userlen = strlen(user);
	else
		userlen = 0;

	if (pwd)
		pwdlen = strlen(pwd);
	else
		pwdlen = 0;

	dbunixodbc->rc =
		SQLConnect(dbunixodbc->dbc, (char *)dsn, dsnlen, (char *)user, userlen, (char *)pwd, pwdlen);
	if (unixodbc_check_error(dbunixodbc) != 0) {
		printf("ERROR: connecting.\n");
		return dbunixodbc->rc;
	}
	else
		unixodbc_dbconnection_set_dsn(dbunixodbc, dsn);

	return dbunixodbc->rc;
}

/** Template string for database connection error.	*/
#define DB_CONNECT_ERROR_S	"Database connection error : \n %s"
/** Connection callback called by the application to create a new connection.
 * Uses the data in 'dbc' to open a database connection, and returns a new
 * DBConnection * in 'ret'.
 * \param	dbc	The DBConnector to use.
 * \param	ret	The connected DBConnection.
 * \return	Error code.	*/
static int unixodbc_connect_callback(const DBConnector * dbc, DBConnection ** ret, void(*display_error) (const char *, const char *))
{
	const char *bdsn, *blogin, *bpass;
	ABuf *buf = NULL;
	DBConnection *ret_my = NULL;
	DBConnection_UNIXODBC *ret_my_unixodbc = NULL;
	int i;

	if (!dbc)
		return CONNECT_ERROR;

	/* Setup connection object */
	ret_my = unixodbc_dbconnection_new();
	ret_my_unixodbc = ret_my->pvt;
	*ret = NULL;

	/* Get connection parameters */
	bdsn = frm_get_value_by_name(dbc->db_frm, BASEDSN);
	blogin = frm_get_value_by_name(dbc->db_frm, BASELOGIN);
	if ((blogin != NULL) && (strlen(blogin) == 0))
		blogin = NULL;

	bpass = frm_get_value_by_name(dbc->db_frm, BASEPASS);
	if ((bpass != NULL) && (strlen(bpass) == 0))
		bpass = NULL;

	unixodbc_do_connect(ret_my_unixodbc, bdsn, blogin, bpass);
	if (ret_my_unixodbc->errmsg) {
		buf =
			buf_new(strlen(DB_CONNECT_ERROR_S) +
					strlen(ret_my_unixodbc->errmsg) + 1);
		if (!buf) {
			fprintf(stderr,
					"connect_callback() could not allocate memory for buffer\n");
			return CONNECT_ERROR;
		}

		snprintf(buf->b_dat, buf->b_len, DB_CONNECT_ERROR_S, ret_my_unixodbc->errmsg);
		if (display_error)
			display_error(_("Connection Error"), buf->b_dat);

		buf = buf_free(buf);
		return CONNECT_ERROR;
	}


	/* The functions (public) */
	ret_my->_connector = (DBConnector *)dbc;
	ret_my->DBclear_result = unixodbc_DBclear_result;
	ret_my->DBdisconnect = unixodbc_DBdisconnect;
	ret_my->DBexecute_query = unixodbc_DBexecute_query;
	ret_my->DBfree_table_def = unixodbc_DBfree_table_def;
	ret_my->DBget_error_message = unixodbc_DBget_error_message;
	ret_my->DBget_field_name = unixodbc_DBget_field_name;
	ret_my->DBget_field_count = unixodbc_DBget_field_count;
	ret_my->DBget_field_value = unixodbc_DBget_field_value;
	ret_my->DBget_query_status = unixodbc_DBget_query_status;
	ret_my->DBget_record_count = unixodbc_DBget_record_count;
	ret_my->DBget_table_def = unixodbc_DBget_table_def;

	/* The keywords (public) */
	ret_my->_keywords = NULL;
	i = 0;
	while (sqlKeywords[i] != NULL)
		dbconnection_add_keyword((DBConnection *) ret_my, KEYWORD_SQL, sqlKeywords[i++]);

	*ret = ret_my;
	buf = buf_free(buf);
	return CONNECT_OK;
}

/** The name that identifies the unixODBC driver.	*/
#define UNIXODBC_TYPESTR	"unixODBC"

/** Register the unixODBC driver with GtkSQL.
 * \return	The allocated DBConnector structure.	*/
DBConnector *register_unixODBC(void)
{
	DBConnector *dbconn;
	FormField *fld;

	dbconn = dbconnector_new();
	if (dbconn == NULL)
		return NULL;

	dbconn->dbtype = UNIXODBC_TYPESTR;

	/* Prepares all the structures */
	dbconn->db_connect_callback = unixodbc_connect_callback;
	dbconn->url_to_form = unixodbc_DBurl_to_form;
	dbconn->form_to_url = unixodbc_DBform_to_url;

	/* Creates the data-structure */
	dbconn->db_frm = frm_new(NULL);
	frm_set_title(dbconn->db_frm, "unixODBC");
	fld = frm_add_field(dbconn->db_frm, BASEDSN, NULL);
	fld_set_tip(fld, _("DSN (required)"));
	fld_set_relevance(fld, FR_REQUIRED);

	fld = frm_add_field(dbconn->db_frm, BASELOGIN, NULL);
	fld_set_tip(fld, _("Username for connection (optional)"));

	fld = frm_add_field(dbconn->db_frm, BASEPASS, NULL);
	fld_set_type(fld, FT_PWD);
	fld_set_tip(fld, _("Password for connection (optional)"));

	return dbconn;
}

#endif
