/** \file db_connect_dlg.c	Database connect dialog.
 * \author Lionel ULMER
 * \author Darryl Luff
 * \ingroup gui	*/
/* GtkSQL -- an interactive graphical query tool for PostgreSQL and MySQL.
 * Copyright (C) 1998-2005  Lionel ULMER, Darryl Luff.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of tdo_he License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "../config.h"

#include <gtk/gtk.h>
#include <string.h>

#include "browser.h"
#include "dburl.h"
#include "gtkui-object.h"
#include "guiapi.h"
#include "register.h"
#include "utils.h"

/* Data keys	*/
#define	KEY_FIELDNAME	"fieldname"	/**< Field name data key. Used to associate a form field with a gtk_widget.	*/

/** Database connection dialog widget.	*/
static GtkWidget *dbConnectDialog = NULL;
GtkWidget *connectorNotebook = NULL;
GtkWidget *browselist = NULL;

static void db_connect_dialog_create(void);
static int db_connect_dialog_reconnect(GtkWidget * widget);

/** Sets the value of a form field from a widget. If the widget has a
 * 'fieldname' data item set in it, the value of the text field is
 * read and store dinto the appropriate named field in the form.
 * \param w	The widget to query.
 * \param data	A pointer to the form.	*/
static void get_text_field(GtkWidget * w, gpointer data)
{
	Form *frm = (Form *) data;
	gchar *fname;
	gchar *s;

	if (!w) {
		return;
	}

	fname = (gchar *) g_object_get_data(G_OBJECT(w), KEY_FIELDNAME);
	if (fname) {
		s = gtk_editable_get_chars(GTK_EDITABLE(w), 0, -1);
		if (s) {
			frm_set_value_by_name(frm, (char *) fname, s);
			g_free(s);
			s = NULL;
		}
	}
}

/** Transfer fields from a the connection pane to a Form. Uses get_text_field()
 * on each GTK widget to extract the data.
 * \param frm	The form pointer.
 * \param pane	The connection pane.	*/
static void load_form_from_pane(Form * frm, GtkWidget * pane)
{
	gtk_container_foreach(GTK_CONTAINER(pane), get_text_field, frm);
}

/** Hide the active dialog.	*/
static void db_connect_dialog_hide(void)
{
	if (dbConnectDialog != NULL) {
		gtk_grab_remove(dbConnectDialog);
		gtk_widget_hide(GTK_WIDGET(dbConnectDialog));
	}
}

/** Reconnect using the login parameters in the login dialog.	*/
static int db_connect_dialog_reconnect(GtkWidget * widget)
{
	GtkWidget *pane = NULL;
	ABuf *errmsg = NULL;
	ABuf *err = NULL;

	int page;
	DBConnection *conn = ui_get_connection();
	DBConnector *dbc = NULL;

	/* Make sure we're not currently connected. */
	if (conn != NULL) {
		ui_disconnect();
	}

	/* First, we get the right connector function. */
	if (!connectorNotebook)
		return CONNECT_ERROR;

	page = gtk_notebook_get_current_page(GTK_NOTEBOOK(connectorNotebook));

	/* Check the connector */
	dbc = dbconnectors_get(page);
	if (dbc == NULL) {
		ui_show_error(_("Internal error"),
					   _("Internal error : open dialog badly created"));
		return CONNECT_ERROR;
	}

	/* Get the notbook pages */
	pane = gtk_notebook_get_nth_page(GTK_NOTEBOOK(connectorNotebook), page);
	if (!pane) {
		g_warning(_("Connection pane is NULL\n"));
		return CONNECT_ERROR;
	}

	/* Load the pane data */
	load_form_from_pane(dbc->db_frm, pane);

	/* Make sure all compulsory fields are complete. */
	errmsg = frm_check_fields(dbc->db_frm);
	if (errmsg == NULL) {
		conn = dbconnector_connect(dbc, ui_show_error);
		if (conn == NULL) {
			g_warning(_("Couldn't finish connecting.\n"));
			return CONNECT_ERROR;
		}
		if (dbconnection_get_status(conn) == CONNECT_OK) {
			ui_set_connection(conn);
		}
		else {
			dbconnection_free(conn);
			return CONNECT_ERROR;
		}

		return dbconnection_get_status(conn);
	}
	else {
		err = buf_new(1024);
		err =
			buf_strcpy(err, _("You did not complete some required fields ("));
		err = buf_strcat(err, (char *)errmsg->b_dat);
		err =
			buf_strcat(err,
					   _("). Please complete them and try " "again. \n"));
		ui_show_error(_("Connection error"), (char *)err->b_dat);
		errmsg = buf_free(errmsg);
		err = buf_free(err);
	}
	return CONNECT_ERROR;
}

/** Sets a widget's text to the value of a form field. Looks up the 'fieldname'
 * data item from the widget to get the field name, then checks for that field
 * in the form. If it exists, its value is copied to the widget.
 * \param w	The widget to query.
 * \param data	A pointer to the form.	*/
static void set_text_field(GtkWidget * w, gpointer data)
{
	Form *frm = (Form *) data;
	gpointer *fname;
	const gchar *s;
	gint pos = 0;

	if (!w) {
		return;
	}

	fname = g_object_get_data(G_OBJECT(w), KEY_FIELDNAME);
	if (fname) {
		s = frm_get_value_by_name(frm, (char *) fname);
		gtk_editable_delete_text(GTK_EDITABLE(w), 0, -1);
		if (s)
			gtk_editable_insert_text(GTK_EDITABLE(w), s, strlen(s), &pos);
	}
}

/** Transfer fields from a Form to the pane. For each field in the form,
 * locate a widget in the container with the same name as the field name,
 * and set it's text to the value of the field.
 * \param pane	The connection pane.
 * \param frm	The form pointer.	*/
static void load_pane_from_form(GtkWidget * pane, Form * frm)
{
	gtk_container_foreach(GTK_CONTAINER(pane), set_text_field, frm);
}

/** Execute the file open dialog for the file 'connect' function.
 * \param widget	The sending object.
 * \param data	???	*/
static void do_browse_file_dialog(GtkWidget * widget, gpointer data)
{
	gtkui_show_file_open_dialog(data);
}

/** Creates a pane from a form.
 * \param	frm	The form data to use.	*/
static GtkWidget *make_form_pane(Form * frm)
{
	GtkWidget *table;
	GtkWidget *lbl;
	GtkWidget *txt;
	GtkWidget *btnLbl;
	GtkWidget *openBtn;
	const FormField *fld;
	const char *tiptext;
	const char *helptext;
	const char *fieldname;
	int i;
	int fieldcount;
	int ftype;
	ABuf *buf = NULL;
	int somerequired = 0;

	/* Creates the notebook pane and label */
	/* The connection pane */
	fieldcount = frm_get_field_count(frm);
	table = gtk_table_new(fieldcount + 1, 3, FALSE);
	gtk_container_set_border_width(GTK_CONTAINER(table), 6);
	for (i = 0; i < fieldcount; i++) {
		fld = frm_get_field(frm, i);
		fieldname = fld_get_name(fld);

		buf = buf_strcpy(buf, fieldname);
		if (fld_get_relevance(fld) == FR_REQUIRED) {
			buf = buf_strcat(buf, "*");
			somerequired++;
		}
		lbl = gtk_label_new((char *)buf->b_dat);
		gtk_misc_set_padding(GTK_MISC(lbl), 4, 0);

		/* this doesnt seem to work? */
		gtk_misc_set_alignment(GTK_MISC(lbl), 1.0, 0.5);
		gtk_widget_show(lbl);

		ftype = fld_get_type(fld);
		tiptext = fld_get_tip(fld);
		helptext = fld_get_help(fld);

		txt = gtk_entry_new();
		if (tiptext)
			gtk_tooltips_set_tip(gtk_tooltips_new(), txt, tiptext, helptext);

		gtk_widget_show(txt);
		switch (ftype) {
		case FT_TEXT:
			break;
		case FT_PWD:
			gtk_entry_set_visibility(GTK_ENTRY(txt), FALSE);
			break;
		case FT_FILENAME:
			openBtn = gtk_button_new();
			btnLbl = gtk_label_new("Browse...");
			gtk_misc_set_padding(GTK_MISC(btnLbl), 4, 0);
			gtk_widget_show(btnLbl);
			gtk_widget_show(openBtn);
			gtk_container_add(GTK_CONTAINER(openBtn), btnLbl);
			g_signal_connect(G_OBJECT(openBtn), "clicked",
							 G_CALLBACK(do_browse_file_dialog), txt);

			gtk_table_attach(GTK_TABLE(table), openBtn, 2, 3, i, i + 1, 0, 0,
							 0, 0);
			break;
		}
		g_object_set_data(G_OBJECT(txt), KEY_FIELDNAME,
						  strdup(frm_get_field_name(frm, i)));
		gtk_table_attach(GTK_TABLE(table), lbl, 0, 1, i, i + 1, GTK_SHRINK,
						 GTK_FILL | GTK_EXPAND, 0, 0);
		gtk_table_attach_defaults(GTK_TABLE(table), txt, 1, 2, i, i + 1);
	}
	if (somerequired > 0) {
		lbl = gtk_label_new(_("* = required fields"));
		gtk_label_set_justify(GTK_LABEL(lbl), GTK_JUSTIFY_LEFT);
		gtk_widget_show(lbl);
		gtk_table_attach_defaults(GTK_TABLE(table), lbl, 1, 2, fieldcount,
								  fieldcount + 1);
	}
	buf_free(buf);
	load_pane_from_form(table, frm);
	return table;
}

/** Callback for 'Connect' dialog Cancel button.	*/
static void db_connect_dialog_cancel_callback(GtkWidget * widget,
											  gpointer data)
{
	GtkWidget *dialog = GTK_WIDGET(data);

	gtk_grab_remove(dialog);
	gtk_widget_hide(dialog);
}

/** Callback for 'Connect' dialog 'OK' button.
 * \param widget	The OK button.
 * \param data	The Connector table.	*/
static void db_connect_dialog_ok_callback(GtkWidget * widget, gpointer data)
{
	if (db_connect_dialog_reconnect(widget) == CONNECT_OK)
		db_connect_dialog_hide();
}

static void browse_changed(GtkComboBox *widget, gpointer user_data) {
	const DBConnector *dbc;
	DBConnector *panedbc;
	const char *url;
	GtkWidget *pane;
	int npages;
	int i;

	url = gtk_combo_box_get_active_text(widget);
	if (! url)
		return;

	dbc = dburl_get_connector(url);
	if (! dbc)
		return;

	dbconnector_url_to_form(dbc, (const char *)url);

	npages = gtk_notebook_get_n_pages(GTK_NOTEBOOK(connectorNotebook));
	for (i=0; i<npages; i++) {
		panedbc = dbconnectors_get(i);
		if (dbc == panedbc) {
			pane = gtk_notebook_get_nth_page(GTK_NOTEBOOK(connectorNotebook), i);
			if (pane) {
				load_pane_from_form(pane, dbc->db_frm);
				gtk_notebook_set_current_page(GTK_NOTEBOOK(connectorNotebook), i);
			}
		}
	}
}

/** Create the connection dialog.	*/
static void db_connect_dialog_create(void)
{
	int i;
	int nb_db;
	GtkWidget *cancel_button;
	GtkWidget *ok_button;

	if (dbConnectDialog != NULL)
		return;

	/* Count the number of databases supported */
	nb_db = registered_drivers_count();

	if (nb_db == 0) {
		ui_show_error(_("Notice"),
					   _("You do not have any database driver "
						 "installed. When GtkSQL was built it could not find any supported database."));
		return;
	}

	/* Creates the top-level dialog */
	dbConnectDialog = gtk_dialog_new();
	gtk_window_set_title(GTK_WINDOW(dbConnectDialog),
						 _("Database connection"));

	/* Browser list	*/
	browselist = gtk_combo_box_entry_new_text ();
	gtk_widget_show (browselist);
	gtk_box_pack_start (GTK_BOX(GTK_DIALOG(dbConnectDialog)->vbox), browselist, FALSE, TRUE, 0);
	g_signal_connect(G_OBJECT(browselist), "changed", G_CALLBACK(browse_changed), NULL);

	/* The fields notebook. */
	connectorNotebook = gtk_notebook_new();
	gtk_notebook_set_tab_pos(GTK_NOTEBOOK(connectorNotebook), GTK_POS_TOP);
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dbConnectDialog)->vbox),
					   connectorNotebook, FALSE, FALSE, 0);
	gtk_widget_show(connectorNotebook);

	/* Buttons... */
	ok_button = gtk_button_new_from_stock(GTK_STOCK_OK);
	g_signal_connect(G_OBJECT(ok_button), "clicked",
					 G_CALLBACK(db_connect_dialog_ok_callback), NULL);
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dbConnectDialog)->action_area),
					   ok_button, FALSE, FALSE, 0);
	GTK_WIDGET_SET_FLAGS(ok_button, GTK_CAN_DEFAULT);
	gtk_widget_grab_default(ok_button);
	gtk_widget_show(ok_button);

	cancel_button = gtk_button_new_from_stock(GTK_STOCK_CANCEL);
	g_signal_connect(G_OBJECT(cancel_button), "clicked",
					 G_CALLBACK(db_connect_dialog_cancel_callback),
					 dbConnectDialog);
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dbConnectDialog)->action_area),
					   cancel_button, FALSE, FALSE, 0);
	gtk_widget_show(cancel_button);

	/* Add the pages to the dialog and the callback functions */
	for (i = 0; i < nb_db; i++) {
		GtkWidget *pane;
		Form *f;

		f = dbconnectors_get_form(i);
		if (f) {
			pane = make_form_pane(f);
			gtk_widget_show(pane);
			gtk_notebook_append_page(GTK_NOTEBOOK(connectorNotebook), pane,
									 gtk_label_new(frm_get_title(f)));
		}
	}
}

void fill_browselist(void) {
	BrowseEntry *b;

	if (! browselist)
		return;

	/* Clear browse list. There must be an easier way?	*/
	GtkTreeModel *model = gtk_combo_box_get_model(GTK_COMBO_BOX(browselist));
	gtk_list_store_clear(GTK_LIST_STORE(model));

	/* And add new entries.	*/
	b = browse_getlist();
	while (b) {
		if (b->url)
			gtk_combo_box_append_text(GTK_COMBO_BOX(browselist), b->url);

		b = b->next;
	}
}

/** Show the connect dialog.	*/
void db_connect_dialog_show(void)
{
	if (dbConnectDialog == NULL) {
		db_connect_dialog_create();
		if (dbConnectDialog == NULL) {
			g_warning("Could not initialise connection dialog.\n");
			return;
		}
	}
	fill_browselist();
	gtk_window_set_resizable(GTK_WINDOW(dbConnectDialog), FALSE);
	gtk_window_set_position(GTK_WINDOW(dbConnectDialog), GTK_WIN_POS_MOUSE);
	gtk_grab_add(dbConnectDialog);
	gtk_widget_show(dbConnectDialog);
}
