/** \addtogroup db 
 *
 *@{
 */
/** \file db_keywords.c	Data structures and functions for managing 
 * the keyword list.
 * \author Lionel ULMER
 * \author Darryl Luff
 *
 */
/* GtkSQL -- an interactive graphical query tool for PostgreSQL and MySQL.
 * Copyright (C) 1998-2005 Lionel ULMER, Darryl Luff.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of tdo_he License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "db_keywords.h"

/** Set the type of a keyword.	*/
static void keyword_set_type(Keyword * key, KeywordType type)
{
	if (!key)
		return;

	key->type = type;
}

static void keyword_set_text(Keyword * key, const char *txt)
{
	if (!key)
		return;

	if (key->key != NULL) {
		free(key->key);
		key->key = NULL;
	}

	if (txt)
		key->key = strdup(txt);
}

const Keyword *keyword_add(Keyword * key, const KeywordType type, const char *txt)
{
	Keyword *k;

	if (!key)
		return NULL;

	/* Check if this record is free */
	if (key->type == KEYWORD_FREE) {
		keyword_set_text(key, txt);
		keyword_set_type(key, type);
		return key;
	}

	/* Any more in the list? */
	if (!key->next) {
		k = keyword_new();
		keyword_set_text(k, txt);
		keyword_set_type(k, type);
		key->next = k;
		return k;
	}

	/* Pass on to next in chain. */
	return keyword_add(key->next, type, txt);
}

int keyword_count(const Keyword * key)
{
	if (key == 0)
		return 0;

	if (key->next)
		return keyword_count(key->next) + 1;

	if (key->type == KEYWORD_FREE)
		return 0;

	return 1;
}

Keyword *keyword_free(Keyword * key)
{
	if (!key)
		return NULL;

	if (key->key) {
		free(key->key);
		key->key = NULL;
	}

	if (key->next)
		key->next = keyword_free(key->next);

	free(key);
	return NULL;
}

/** Mark a keyword structure as 'free'.	*/
static void keyword_mark_free(Keyword * key)
{
	if (!key)
		return;

	if (key->key) {
		free(key->key);
		key->key = NULL;
	}

	key->type = KEYWORD_FREE;
}

const Keyword *keyword_match(const Keyword * key, const char *check)
{
	if (!key)
		return NULL;

	/* A null or empty key matches anything. */
	if ((!check) || (strlen(check) < 1))
		return key;

	if (!strncasecmp(key->key, check, strlen(check)))
		return key;
	else
		return keyword_match(key->next, check);
}

const Keyword *keyword_match_exact(const Keyword * key, const char *check)
{
	if (!key) {
		fprintf(stderr, "keyword_match_exact() key is NULL\n");
		return NULL;
	}

	if (!check) {
		fprintf(stderr, "keyword_match_exact() check is NULL\n");
		return NULL;
	}

	if (!strcasecmp(key->key, check)) {
		return key;
	}
	else if (key->next) {
		return keyword_match_exact(key->next, check);
	}

	return NULL;
}

const Keyword *keyword_match_next(const Keyword * key, const char *check)
{
	if (!key)
		return NULL;

	if (key->next)
		return keyword_match(key->next, check);

	return NULL;
}

Keyword *keyword_new()
{
	Keyword *key = calloc(1, sizeof(Keyword));
	if (key)
		key->type = KEYWORD_FREE;

	return key;
}

void keyword_remove_type(Keyword * key, const KeywordType type)
{
	if (!key)
		return;

	if (key->type == type)
		keyword_mark_free(key);

	if (key->next)
		keyword_remove_type(key->next, type);
}

/*@}*/
