/** \addtogroup db
 *@{
 */
/** \file db_tabledef.c
 * \author Lionel ULMER
 * \author Darryl Luff
 *
 * This module defines and manages the data structures that handle 
 * the database table and field definitions. There are two structures.
 *
 * - DBTableDef provides a definition for a single database table.
 * - DBFieldDef provides a definition for a single table.
 *
 * Each of these are implemented as linked lists.
 * Each database connection structure contains a pointer to a 
 * TableDef list that contains details of the tables in that database.
 * Each DBTableDef has a pointer to a DBFieldDef which is the head of 
 * a list of fields for the table.
 *
 * There are accessor functions provided for each field of the 
 * structures, so you should never have to manipulate the structure
 * fields directly. These functions take care of null pointers etc
 * so should be used for all accesses to the structure data.
 *
 * As a convenience, there are also a few functions provided for
 * the table structure to access the table fields.
 */
/*@}*/
/* GtkSQL -- an interactive graphical query tool for PostgreSQL
 * Copyright (C) 1998-2005	Lionel ULMER, Darryl Luff
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "../config.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "db_tabledef.h"

void fielddef_setname(DBFieldDef *fld, const char *name) {
	if (fld == NULL)
		return;

	if (fld->fldName) {
		free(fld->fldName);
		fld->fldName = NULL;
	}

	if (name != NULL)
		fld->fldName = strdup(name);
}

void fielddef_settype(DBFieldDef *fld, const char *type) {
	if (fld == NULL)
		return;

	if (fld->fldType) {
		free(fld->fldType);
		fld->fldType = NULL;
	}

	if (type != NULL)
		fld->fldType = strdup(type);
}

static const DBFieldDef *fielddef_add(DBFieldDef *fld, const DBFieldDef *fld2) {
	if ((fld == NULL) || (fld2 == NULL))
		return NULL;

	if (fld->next)
		return fielddef_add(fld->next, fld2);
	else
		fld->next = (DBFieldDef *)fld2;

	return fld2;
}

static int fielddef_count(DBFieldDef *fld) {
	if (fld == NULL)
		return 0;

	return (fielddef_count(fld->next) + 1);
}

static DBFieldDef *fielddef_free(DBFieldDef *fld) {
	if (fld == NULL)
		return NULL;
	
	if (fld->next == fld) {
		fprintf(stderr, "WARNING: fielddef->fred(). fld->next == fld\n");
		fld->next = NULL;
	}

	if (fld->next) {
		fielddef_free(fld->next);
	}

	fielddef_setname(fld, NULL);
	fielddef_settype(fld, NULL);
	free(fld);
	return NULL;
}

static DBFieldDef *fielddef_getfield(DBFieldDef *fld, int i) {
	if (fld == NULL)
		return NULL;

	if (i < 0)
		return NULL;

	if (i == 0)
		return fld;

	return fielddef_getfield(fld->next, i-1);
}

const char *fielddef_getname(const DBFieldDef *fld) {
	if (fld == NULL)
		return NULL;

	return fld->fldName;
}

const char *fielddef_gettype(const DBFieldDef *fld) {
	if (fld == NULL)
		return NULL;

	return fld->fldType;
}

DBFieldDef *fielddef_new(void) {
	return calloc(1, sizeof(DBFieldDef));
}

const DBTableDef *tabledef_add(DBTableDef *tbl, const DBTableDef *tbl2) {
	if (tbl == NULL)
		return NULL;

	if (tbl->next == NULL)
		tbl->next = (DBTableDef *)tbl2;
	else
		tabledef_add(tbl->next, tbl2);

	return tbl2;
}

const DBFieldDef *tabledef_addfield(DBTableDef *tbl, const DBFieldDef *fld) {
	if (tbl == NULL)
		return NULL;

	if (tbl->fieldDef == NULL)
		tbl->fieldDef = (DBFieldDef *)fld;
	else
		fielddef_add(tbl->fieldDef, fld);

	return fld;
}

int tabledef_count(const DBTableDef *tbl) {
	if (tbl == NULL)
		return 0;

	if (tbl->next == NULL)
		return 1;
	else
		return tabledef_count(tbl->next) + 1;
}

int tabledef_fieldcount(const DBTableDef *tbl) {
	if (tbl == NULL)
		return 0;

	return fielddef_count(tbl->fieldDef);
}

DBTableDef *tabledef_free(DBTableDef *tbl) {
	if (! tbl)
		return NULL;

	/* Free the following tabledefs first	*/
	if (tbl->next)
		tbl->next = tabledef_free(tbl->next);

	/* Now free fields	*/
	tabledef_freefields(tbl);

	/* And finally, free this tabledef	*/
	tabledef_setname(tbl, NULL);
	free(tbl);
	return NULL;
}

void tabledef_freefields(DBTableDef *tbl) {
	if (! tbl)
		return;

	/* Now free fields	*/
	if (tbl->fieldDef != NULL) {
		tbl->fieldDef = fielddef_free(tbl->fieldDef);
	}
}

const DBFieldDef *tabledef_getfield(const DBTableDef *tbl, int i) {
	if (tbl == NULL)
		return NULL;

	return fielddef_getfield(tbl->fieldDef, i);
}

const char *tabledef_getname(const DBTableDef *tbl) {
	if (tbl == NULL)
		return NULL;

	return tbl->name;
}

const DBTableDef *tabledef_gettable(const DBTableDef *tbl, int i) {
	if ((tbl == NULL) || (i < 0))
		return NULL;
		
	if (i == 0)
		return tbl;

	return tabledef_gettable(tbl->next, i-1);
}

DBTableDef *tabledef_new() {
	return calloc(1, sizeof(DBTableDef));
}

DBFieldDef *tabledef_setfields(DBTableDef *tbl, int n) {
	int i;

	if (tbl == NULL)
		return NULL;

	tabledef_freefields(tbl);
	if (n > 0) {
		tbl->fieldDef = fielddef_new();
		for (i=1; i<n; i++)
			fielddef_add(tbl->fieldDef, fielddef_new());
	}
	return tbl->fieldDef;
}

void tabledef_setname(DBTableDef *tbl, const char *name) {
	if (tbl == NULL)
		return;

	if (tbl->name) {
		free(tbl->name);
		tbl->name = NULL;
	}

	if (name != NULL)
		tbl->name = strdup(name);
}
