/** \addtogroup db Database
 * The functions in the database module provide the interface to the database drivers and database
 * connections. All calls from other parts of the program to the database system should be through 
 * the interface defined in this module.
 *
 *@{
 */
/** \file dbapi.c	Public database API.
 * \author Lionel ULMER
 * \author Darryl Luff
 */
/*@}*/
/* GtkSQL -- an interactive graphical query tool for PostgreSQL
 * Copyright (C) 1998-2005 Lionel ULMER, Darryl Luff
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "../config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "dburl.h"
#include "plugins.h"
#include "plugins_pil.h"
#include "register.h"

#define DBAPI_PLUGINDIR "db.plugindir"
char *plugindir = NULL;

/*static ResultData *resultdata_free(ResultData *data) {
	if (data) {
		free(data);
	}
	return NULL;
}

ResultData *resultdata_new(void) {
	ResultData *data;
	data = calloc(1, sizeof(ResultData));

	return data;
}

ResultData *dbresult_get_data(DBResult *result) {
	if (! result)
		return NULL;

	return result->data;
}

DBConnection *dbresult_get_connection(DBResult *result) {
	ResultData *data = dbresult_get_data(result);
	if (! data)
		return NULL;

	return data->conn;
}

*/

/** WARNING: This does NOT free any internal data. Just frees the
 * structure itself.
 * \todo Free driver data.	*/
DBQResult *dbresult_free(DBQResult *result) {
	if (result)
		free(result);

	return NULL;
}

DBQResult *dbresult_new(void *data) {
	DBQResult *r = calloc(1, sizeof(DBQResult));
	if (r)
		dbresult_set_data(r, data);

	return r;
}

/** Update a configuration value.	*/
void db_config_value(const char *key, const char *val)
{
	if (!strncmp(DBAPI_PLUGINDIR, key, strlen(DBAPI_PLUGINDIR))) {
		if (plugindir) {
			free(plugindir);
			plugindir = NULL;
		}

		if (val && strlen(val) > 0)
			plugindir = strdup(val);
	}
}

/** Connect to the database defined in the current workspace.
 * \param url	The database URL to connect to.
 * \return The DBConnection.	*/
DBConnection *db_connect_url(const char *url, void(*display_error) (const char *, const char *)) {
	DBConnection *conn = NULL;
	const DBConnector *dbc;
	int return_status;

	/* Get a connector for this URL	*/
	dbc = dburl_get_connector(url);
	if (dbc == NULL) {
		printf("Could not find a connector for the URL \"%s\"\n", url);
		return 0;
	}

	/* Set the login parameters	*/
	dbconnector_url_to_form(dbc, url);

	/* Now, connect with it	*/
	return_status = dbc->db_connect_callback(dbc, &conn, display_error);
	if (conn != NULL)
		conn->_connectionStatus = return_status;

	return conn;
}

DBConnection *db_connect_url_simple(const char *url) {
	return db_connect_url(url, NULL);
}

void db_init() {
	dbconnectors_init();
}

/** Allocate a new generic DBConnector. All fields are NULL-filled.
 * @return	A pointer to a new DBConnector.	*/
DBConnector *dbconnector_new(void)
{
	return (DBConnector *) calloc(1, sizeof(DBConnector));
}

/** Connects to the database.
 * \param dbc	The connector to use.
 * \return A pointer to the new DBConnection structure, or NULL on error.	*/
DBConnection *dbconnector_connect(const DBConnector *dbc, void(*display_error) (const char *, const char *)) {
	DBConnection *conn = NULL;
	int return_status = dbc->db_connect_callback(dbc, &conn, display_error);
	if (conn != NULL)
		conn->_connectionStatus = return_status;

	return conn;
}

void dbconnector_url_to_form(const DBConnector * dbc, const char *url) {
	if (! dbc)
		return;

	if (! dbc->url_to_form)
		return;

	dbc->url_to_form(dbc->db_frm, url); 
}

/** Disconnects the current database.
 * \param conn	The connection to disconnect.
 * \param display_error	A pointer to an error 
 * function with this signature: void display_error(char *title, char *msg).
 * If not NULL, this function is called to display error messages if
 * required.	*/
void dbconnection_disconnect(DBConnection *connection, void(*display_error) (char *, char *)) {
	if ((! connection) || (! connection->DBdisconnect)) {
		if (display_error != NULL)
			display_error("Internal error",
					  "Internal error : trying to disconnect "
					  "with no current connection");
		return;
	}
	connection->DBdisconnect(connection);
}

void dbconnection_disconnect_simple(DBConnection *connection) {
	dbconnection_disconnect(connection, NULL);
}

const char *dbconnection_get_error_message(const DBConnection *connection) {
	if ((! connection) || (!connection->DBget_error_message))
		return NULL;

	return connection->DBget_error_message(connection);
}

const char *dbconnection_get_db_name(const DBConnection *dbc) {
	if (! dbc)
		return NULL;

	return dbc->dbname;
}

DBQResult *dbconnection_execute_query(DBConnection *dbc, const char *query) {
	if ((!dbc) || (! dbc->DBexecute_query))
		return NULL;

	return dbc->DBexecute_query(dbc, query);
}

int dbconnection_get_query_status(const DBConnection *dbc, const DBQResult *result) {
	if ((!dbc) || (! dbc->DBget_query_status))
		return 0;

	return dbc->DBget_query_status(dbc, result);
}

int dbresult_get_status(const DBQResult *result) {
	if (!result)
		return 0;

	return result->status;
}

int dbconnection_get_record_count(const DBConnection *dbc, const DBQResult *result) {
	if ((!dbc) || (! dbc->DBget_record_count))
		return 0;

	return dbc->DBget_record_count(dbc, result);
}

int dbconnection_get_field_count(const DBConnection *dbc, const DBQResult *result) {
	if ((! dbc) || (! dbc->DBget_field_count))
		return 0;

	return dbc->DBget_field_count(dbc, result);
}

const char *dbconnection_get_field_name(const DBConnection *dbc, const DBQResult *result, int field) {
	if ((!dbc) || (! dbc->DBget_field_name))
		return NULL;

	return dbc->DBget_field_name(dbc, result, field);
}

const char *dbconnection_get_field_value(const DBConnection *dbc, const DBQResult *result, int record, int field) {
	if ((!dbc) || (! dbc->DBget_field_value))
		return NULL;

	return dbc->DBget_field_value(dbc, result, record, field);
}

/** Add a keyword to the keyword-completion list.
 * \param conn	The database connection.
 * \param type	The type of keyword.
 * \param txt	The keyword.	*/
void dbconnection_add_keyword(DBConnection * conn, const KeywordType type, const char *txt)
{
	if (txt == NULL) {
		fprintf(stderr, "add_keyword(): keyword text is NULL\n");
		return;
	}

	if (conn == NULL) {
		fprintf(stderr, "add_keyword(): conn is NULL\n");
		return;
	}

	if (! conn->_keywords)
		conn->_keywords = keyword_new();

	if (conn->_keywords == NULL) {
		fprintf(stderr, "add_keyword(): keylist is NULL\n");
		return;
	}

	keyword_add(conn->_keywords, type, txt);
}

void dbconnection_clear_result(DBConnection *conn, DBQResult *result) {
	if (! conn) {
		fprintf(stderr, "dbconnection_clear_result(%p, %p): conn is null\n", conn, result);
		return;
	}

	if (! result) {
		fprintf(stderr, "dbconnection_clear_result(%p, %p): result is null\n", conn, result);
		return;
	}
	conn->DBclear_result(conn, result);
}

const DBConnector *dbconnection_get_connector(const DBConnection *connection) {
	if (!connection)
		return NULL;

	return connection->_connector;
}

int dbconnection_get_table_def(DBConnection *dbc) {
	if (! dbc)
		return 0;

	/* Clear existing table list	*/
	dbc->_tableDef = tabledef_free(dbc->_tableDef);

	/* Load the new one	*/
	if (!dbc->DBget_table_def)
		return 0;

	return dbc->DBget_table_def(dbc);
}

DBConnection *dbconnection_new(void) {
	DBConnection *dbc;

	dbc = calloc(1, sizeof(DBConnection));;
	return dbc;
}

void dbconnection_set_dbname(DBConnection * dbc, const char *n)
{
	if (! dbc)
		return;

	if (dbc->dbname)
		free(dbc->dbname);

	dbc->dbname = strdup(n);
}

const char *dbconnector_form_to_url(const DBConnector * dbc, int showPwd) {
	if (! dbc)
		return NULL;

	if (! dbc->form_to_url)
		return NULL;

	return dbc->form_to_url(dbc->db_frm, showPwd);
}

void dbconnection_free_table_def(DBConnection *dbc) {
	if ((!dbc) || (! dbc->DBfree_table_def))
		return;

	dbc->DBfree_table_def(dbc);
	dbc->_tableDef = tabledef_free (dbc->_tableDef);
}

int dbconnection_get_table_count(const DBConnection *dbc) {
	if (dbc == NULL)
		return 0;

	return tabledef_count(dbc->_tableDef);
}

DBConnector *dbconnection_free(DBConnection * dbconn)
{
	if (! dbconn)
		return NULL;

	dbconn->_connector = NULL;	/* No need to free this	*/
	keywords_remove_all(dbconn);
	dbconnection_disconnect(dbconn, NULL);
	free(dbconn);
	return NULL;
};

void *dbresult_get_data(const DBQResult *result) {
	if (! result)
		return NULL;

	return result->data;
}

void dbresult_set_data(DBQResult *result, void *data) {
	if (! result) {
		fprintf(stderr, "dbresult_set_data(%p, %p): result NULL\n", result, data);
		return;
	}

	result->data = data;
}

int dbconnection_get_status(const DBConnection *connection) {
	if (!connection)
		return CONNECT_ERROR;

	return connection->_connectionStatus;
}

const DBTableDef *dbconnection_get_table(const DBConnection *dbc, int table) {
	if (! dbc)
		return NULL;

	return tabledef_gettable(dbc->_tableDef, table);
}

const char *dbconnector_get_help_url(const DBConnector *dbc) {
	if ((!dbc) || (! dbc->get_help_url))
		return NULL;

	return dbc->get_help_url(dbc);
}

const char *dbconnector_get_name(const DBConnector *dbc) {
	if ((!dbc) || (! dbc->dbtype))
		return "Unknown";

	return dbc->dbtype;
}

const char *dbconnector_get_service_type(const DBConnector * dbc) {
	if ((!dbc) || (! dbc->get_service_type))
		return NULL;

	return dbc->get_service_type(dbc);
}

const char *dbconnector_get_driver_info(const DBConnector *dbc) {
	if ((!dbc) || (! dbc->get_driver_info))
		return NULL;

	return dbc->get_driver_info(dbc);
}

const char *dbconnector_get_help_search_url(const DBConnector *dbc, const char *term) {
	if (!dbc)
		return NULL;

	if (! dbc->get_help_search_url) {
		/* If no specific search URL, return the generic help URL.	*/
		return dbconnector_get_help_url(dbc);
	}

	return dbc->get_help_search_url(dbc, term);
}


static DBConnector **connectorTable = NULL;

/** Find a connector for a specified database type.
 * \param key	The type name. This needs to match the 'dbtype' field in the
 * DBConnector structure.
 * \return A pointer to the connector, or NULL.	*/
DBConnector *dbconnectors_find(char *key) {
	int i = 0;
	DBConnector *dbc;
	if (connectorTable == NULL) {
		fprintf(stderr, "WARNING: connectorTable is NULL\n");
		return NULL;
	}
	dbc = connectorTable[i];
	while (dbc != NULL) {
		if (strcasecmp(dbc->dbtype, key) == 0)
			return dbc;

		i++;
		dbc = connectorTable[i];
	}
	return NULL;
}

/** Initialise the connector table.	*/
void dbconnectors_init(void) {
	int i;
	db_register_func driver;
	int nb_db;

	if (connectorTable != NULL)
		return;

	/* Create the connector table */
	registered_drivers_init_static();

#ifdef USE_PILS
	pilplugins_init("/usr/local/gtksql");
#else
	plugins_load(NULL);
#endif
	
	nb_db = registered_drivers_count();
	if (nb_db < 1) {
		printf("No database drivers registered\n");
		return;
	}

	connectorTable = (DBConnector **) calloc(sizeof(DBConnector *), nb_db+1);
	driver = registered_drivers_get_first();
		
	i = 0;
	while (driver) {
		connectorTable[i] = (driver)();
		printf("Connecting %s driver\n", connectorTable[i]->dbtype);
		connectorTable[i + 1] = NULL;
		driver = registered_drivers_get_next(driver);
		i++;
	}

	/* Add the pages to the dialog and the callback functions */
	/*for (i = 0; i < nb_db; i++) {
		connectorTable[i] = registerDB[i]();
		printf("Connecting %s driver\n", connectorTable[i]->dbtype);
		connectorTable[i + 1] = NULL;
	}*/
}

int dbconnectors_count(void) {
	return registered_drivers_count();
}

/** Get the form from a connector.	*/
Form *dbconnectors_get_form(int i) {
	if ((i >= 0) && (i < registered_drivers_count()))
		return connectorTable[i]->db_frm;

	return NULL;
}

/** Get the 'i'th connector.	*/
DBConnector *dbconnectors_get(int i) {
	if ((i >= 0) && (i < registered_drivers_count()))
		return connectorTable[i];

	return NULL;
}

/** Remove all keywords for the given type on a connection.	*/
void keywords_remove_type(DBConnection *conn, KeywordType type) {
	if (! conn)
		return;

	keyword_remove_type(conn->_keywords, type);
}

/** Remove all keywords on a connection.	*/
void keywords_remove_all(DBConnection *conn) {
	if (! conn)
		return;

	conn->_keywords = keyword_free(conn->_keywords);
}
