/** \addtogroup db
 * The functions in the database module provide the interface to the database drivers and database
 * connections. All calls from other parts of the program to the database system should be through 
 * the interface defined in this module.
 *
 * \par
 * There are two main structures provided by the module:
 * - \ref DBConnector "DBConnector". This is an interface to a database driver. To connect to a database, you need to get a 
 * DBConnector for the specific database driver, and then use this connector to connect to the actual database.
 * - \ref DBConnection "DBConnection". When you connect to a database, you get back a DBConnection. This structure includes
 * pointers to functions to operate on the connection. These functions are used to send queries and access
 * the results of the query, among other things.
 *
 *@{
 */
/** \file dbapi.h	Public database API.
 * \author Lionel ULMER
 * \author Darryl Luff
 *
 * This file describes the API functions supported by the database module, and available for use by
 * other modules.
 */
/* GtkSQL -- an interactive graphical query tool for PostgreSQL and MySQL.
 * Copyright (C) 1998-2005  Lionel ULMER, Darryl Luff.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of tdo_he License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifndef _DBAPI_H
#define _DBAPI_H

#include "db_keywords.h"
#include "db_tabledef.h"
#include "forms.h"

/** Query Status codes	*/
enum QueryStatus {
	DB_QUERY_ERROR = 0,
	DB_QUERY_EMPTY = 1,
	DB_QUERY_COMMAND_OK = 2,
	DB_QUERY_RECORDS_OK = 3,
	DB_QUERY_OTHER = 4
};

/** How to connect.	*/
enum ConnectStatus {
	CONNECT_DISCONNECTED = 0,
	CONNECT_ERROR,
	CONNECT_OK
};

/** DBConnector is the type used to refer to the _dbconnector structure.	*/
typedef struct _dbconnector DBConnector;
typedef struct _dbresult DBQResult;

/** Per-connection database structure. This defines the interface the
 * application talks to the driver with. It also holds the keyword list
 * for the connection.	*/
struct db_conn {
	DBConnector *_connector;	/**< A link back to the connector that created this connection.	*/
	int _connectionStatus;	/**< \see ConnectStatus	*/
	char *dbname;	/**< Database name.	*/

	Keyword *_keywords;	/**< Keywords for autocompletion.	*/
	DBTableDef *_tableDef;	/**< Table definitions.	*/
	void *pvt;	/**< Private driver data	*/
	unsigned int _pvtlen;	/**< Length of private driver data	*/

	/* Connection and status functions. */
	int (*DBdisconnect) (struct db_conn *);	/**< Disconnect from database.	*/

	/* Information functions. */
	const char *(*DBget_error_message) (const struct db_conn *);	/**< Return the last error message.	*/

	/* Queries functions */
	DBQResult *(*DBexecute_query) (struct db_conn *, const char *);	/**< Execute a query.	*/
	int (*DBget_query_status) (const struct db_conn *, const DBQResult *);	/**< Return the status of the query.	*/
	int (*DBget_record_count) (const struct db_conn *, const DBQResult *);	/**< Return the record count.	*/
	int (*DBget_field_count) (const struct db_conn *, const DBQResult *);	/**< Get the field (column) count.	*/
	const char *(*DBget_field_name) (const struct db_conn *, const DBQResult *, int field);	/**< Get a field name.	*/
	const char *(*DBget_field_value) (const struct db_conn *, const DBQResult *, int record, int field);	/**< Get a field value.	*/
	void (*DBclear_result) (struct db_conn *, DBQResult *result);	/**< Clear the query result.	*/

	/* Tables functions (definitions, ...) */
	int (*DBget_table_def) (struct db_conn *);			/**< Load table definitions.	*/
	void (*DBfree_table_def) (struct db_conn *);		/**< Free table definitions.	*/
};

/** DBConnection is the type used to refer to the db_conn structure.	*/
typedef struct db_conn DBConnection;

/** Data passed from the application to the driver when initiating a
 * connection. Contains login data, and a pointer to a callback function in the
 * application that can display an error message from the driver. Most fields 
 * in this structure are initialised by the database driver in the registration
 * function.	*/
struct _dbconnector {
	char *dbtype;	/**<- Database driver type string.	*/
	char *errmsg;	/**<- Last error message.	*/

	/** Callback function for the application to call to open a connection.	*/
	int (*db_connect_callback) (const DBConnector *, DBConnection **, void(*display_error) (const char *, const char *));
	const char *(*get_driver_info)(const DBConnector *dbc);
	const char *(*get_help_url) (const DBConnector *);	/**< Get the generic help URL for this driver.	*/
	const char *(*get_help_search_url) (const DBConnector *, const char *term);	/**< Get the URL to search for this term.	*/
	const char *(*get_service_type)(const DBConnector *dbc);
	void (*url_to_form) (Form *, const char *);	/**< Load a URL into form fields.	*/
	const char *(*form_to_url) (const Form *, int showPwd);	/**< Return a URL string from form fields.	*/

	Form *db_frm;	/**< Form listing the fields required to connect.	*/
};

/** The result of a query.	*/
struct _dbresult {
	void *data;	/**< Private data for the driver.	*/
	int	status;	/**< Status from the query.	*/
};

typedef DBConnector *(*db_register_func) (void);

/* Other	*/
void db_config_value(const char *key, const char *val);
void db_init();

/** XXX PILS START Testing PILS interface	*/
struct DBAPIOps {
	void (*set_imports)(void *);	/* kludge until I figure out how you're supposed to do it */
	int (*db_connect) (DBConnection **);
	const char *(*get_help_url) (void);	/**< Get the generic help URL for this driver.	*/
	const char *(*get_help_search_url) (const char *term);	/**< Get the URL to search for this term.	*/
	void (*url_to_form) (Form *, const char *);	/**< Load a URL into form fields.	*/
	const char *(*form_to_url) (const Form *);	/**< Return a URL string from form fields.	*/
	const char *(*get_dbtype) (void);	/**< Return a URL string from form fields.	*/

	Form *db_frm;	/**< Form listing the fields required to connect.	*/
};

struct DBAPIImports {
	void (*display_error)(const char *, const char *);
};

// PILS END

/** Connect to the database defined in the current workspace.
 * \param url	The database URL to connect to.
 * \return The DBConnection.	*/
DBConnection *db_connect_url(const char *url, void(*display_error) (const char *, const char *));
/** Same as db_connect_url, except that the error function pointer is missing.	
 */
DBConnection *db_connect_url_simple(const char *url);

/* dbconnector functions	*/
DBConnection *dbconnector_connect(const DBConnector * dbc, void(*display_error) (const char *, const char *));
DBConnector *dbconnector_new(void);
const char *dbconnector_form_to_url(const DBConnector * dbc, int showPwd);
void dbconnector_url_to_form(const DBConnector * dbc, const char *url);

/** Get the generic help URL for this driver.	*/
const char *dbconnector_get_help_url(const DBConnector *dbc);
/** Free the connector. NOTE: Does not do any disconnections. Just frees the structure.
 * @return	A pointer to a new DBConnector.	*/
DBConnector *dbconnection_free(DBConnection * dbconn);
const char *dbconnector_get_driver_info(const DBConnector *dbc);
/** Get the URL to search for a term in this driver's documentation.	*/
const char *dbconnector_get_help_search_url(const DBConnector *dbc, const char *term);
/** Get the URL to search for a term in this driver's documentation.	*/
const char *dbconnector_get_name(const DBConnector *dbc);
const char *dbconnector_get_service_type(const DBConnector * dbc);

/* dbconnection functions	*/
void dbconnection_add_keyword(DBConnection * conn, const KeywordType type, const char *txt);
void dbconnection_clear_result(DBConnection *conn, DBQResult *result);
void dbconnection_disconnect(DBConnection * conn, void (*display_error) (char *, char *));
/** Disconnects the current database.
 * \param conn	The connection to disconnect.
 * \param display_error	A pointer to an error 
 * function with this signature: void display_error(char *title, char *msg).
 * If not NULL, this function is called to display error messages if
 * required.	*/
void dbconnection_disconnect_simple(DBConnection *connection);
DBQResult *dbconnection_execute_query(DBConnection * dbc, const char *query);
void dbconnection_free_table_def(DBConnection * dbc);
const DBConnector *dbconnection_get_connector(const DBConnection *connection);
int dbconnection_get_field_count(const DBConnection * dbc, const DBQResult * result);
const char *dbconnection_get_db_name(const DBConnection * dbc);
const char *dbconnection_get_error_message(const DBConnection * connection);
const char *dbconnection_get_field_name(const DBConnection * dbc, const DBQResult * result, int field);
const char *dbconnection_get_field_value(const DBConnection * dbc, const DBQResult * result, int record, int field);
int dbconnection_get_query_status(const DBConnection * dbc, const DBQResult * result);
int dbconnection_get_record_count(const DBConnection * dbc, const DBQResult * result);
int dbconnection_get_status(const DBConnection *connection);
const DBTableDef *dbconnection_get_table(const DBConnection * dbc, int table);
int dbconnection_get_table_count(const DBConnection * dbc);
int dbconnection_get_table_def(DBConnection * dbc);
DBConnection *dbconnection_new(void);
/** Set the filename for the connection.
 * \param	dbc	The connection.
 * \param	fn	Filename to set.	*/
void dbconnection_set_dbname(DBConnection * dbc, const char *n);

int dbconnectors_count(void);
void dbconnectors_init(void);
DBConnector *dbconnectors_find(char *key);
DBConnector *dbconnectors_get(int i);
Form *dbconnectors_get_form(int i);

/* DBResult methods	*/
DBQResult *dbresult_new(void *data);
void *dbresult_get_data(const DBQResult *result);
void dbresult_set_data(DBQResult *result, void *data);

void keywords_remove_all(DBConnection * conn);
void keywords_remove_type(DBConnection * conn, KeywordType type);


/*@}*/
#endif
