/** \file forms.c	Forms interface.
 * \author Lionel ULMER
 * \author Darryl Luff
 * \ingroup forms
 *
 * Each form can have a title and a number
 * of fields. Each field has a name and a value. This is used in GtkSQL as
 * the interface for database drivers to popup login dialogs etc. The driver
 * creates a Form object that defines the fileds for the login dialog. GtkSQL
 * then turns this into a GTK dialog.
 *
 * \par FormField's
 * The FormField type has built-in support for a linked list. Some field functions work on a single field,
 * others work on a field list. All functions take a FormField pointer as the first parameter. The ones that work
 * on a list normally take an integer index as a second parameter. This selects the index number in the list
 * of the field you want to work on. So given a function like fld_xyz(FormField *fld, int n), the call frm_xyz(fld, 5) 
 * would operate on the sixth element of the field list based at 'fld'. The index number is zero-based, so '0' refers
 * to the first field in the list, '1' to the second, etc.
 */
/* $Id: forms.c,v 1.18 2007/12/01 12:00:12 djluff Exp $
 * GtkSQL -- an interactive graphical query tool for PostgreSQL
 * Copyright (C) 2003-2004 Darryl Luff
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "forms.h"

/** Longest field name supported.	*/
#define LONGEST_KEY 256

/** Creates and returns a new field.	*/
FormField *fld_new(const char *nm) {
	FormField *fnew = (FormField *)calloc(1, sizeof(FormField));
	if (fnew) {
		fnew->type = FT_TEXT;
		fnew->rel = FR_OPTIONAL;
		if (nm)
			fnew->name = strdup(nm);
	}
	return fnew;
}

/** Add a field to the end of the list.
 * \param fld	If NULL, allocates a new field and returns it as the list.
 * \param nm	The name for the field.
 * \return A pointer to the list. If 'fld' was NULL, the return value will be the newly allocated Field. If
 * 'fld' was NOT NULL, the return value will be 'fld'.	*/
FormField *fld_add(FormField *fld, const char *nm) {
	FormField *fnew = NULL;

	if (! fld) {
		fnew = fld_new(nm);
		return fnew;
	}

	/* Check if the field already exists	*/
	if (fld->next) {
		/* pass it on	*/
		fld_add((FormField *)(fld->next), nm);
		return fld;
	}
	else {
		fnew = fld_new(nm);
		fld->next = fnew;
		return fld;
	}
	return NULL;
}

/** Count the fields in the list.
 * \param fld	The start of the list.
 * \return The field count.	*/
int fld_count(FormField *fld) {
	FormField *f = fld;
	int cnt = 0;

	while (f) {
		cnt++;
		f = f->next;
	}
	return cnt;
}

/** Sets the name of the specified field.	*/
FormField *fld_set_name(FormField *fld, char *nm) {
	if (! fld)
		return NULL;

	if (fld->name)
		free(fld->name);

	if (nm)
		fld->name = strdup(nm);
	else
		fld->name = NULL;

	return fld;
}

/** Sets the value of the specified field.
 * \param fld The field to set.
 * \param val The new value (may be NULL).	*/
void fld_set_value(FormField *fld, const char *val) {
	if (! fld)
		return;

	if (fld->value)
		free(fld->value);

	if (val)
		fld->value = strdup(val);
	else
		fld->value = NULL;
}

/** Frees the field list based at 'fld'.
 * \param fld	The base of the list.
 * \return Always returns NULL. This is so you can do 
 * "fields = fld_free(fields);" to NULL the pointer.	*/
FormField *fld_free(FormField *fld) {
	if (! fld)
		return NULL;

	if (fld->next)
		fld->next = fld_free(fld->next);

	fld_set_name(fld, NULL);
	fld_set_value(fld, NULL);
	free(fld);
	return NULL;
}

/** Retrieve the 'i'th field in the list.
 * \param	fld	The start field.
 * \param	i	The index of the field to find.
 * \return	A pointer to the selected field.	*/
const FormField *fld_get(const FormField *fld, int i) {
	if (! fld)
		return NULL;

	if (i < 0)
		return NULL;
	else if (0 == i)
		return fld;
	else
		return fld_get(fld->next, i-1);
}

/** Retrieve the help text for the field.
 * \param	fld	The field.
 * \return	The help string, or NULL if none specified.	*/
const char *fld_get_help(const FormField *fld) {
	if (! fld)
		return NULL;

	return fld->help;
}

/** Retrieve the tooltip text for the field.
 * \param	fld	The field.
 * \return	The relevance option.	*/
__attribute__((const)) FieldRelevance fld_get_relevance(const FormField *fld) {
	if (! fld)
		return FR_OPTIONAL;

	return fld->rel;
}

/** Retrieve the tooltip text for the field.
 * \param	fld	The field.
 * \return	The text string, or NULL if none specified.	*/
const char *fld_get_tip(const FormField *fld) {
	if (! fld)
		return NULL;

	return fld->tip;
}

/** Returns the edit mode of the selected field in the list based at 'fld'.
 * \param fld	The field.
 * \return The field type.
 * \see FieldType
 * \see FT_NONE
 * \see FT_TEXT
 * \see FT_PWD
 * \see FT_FILENAME	*/
__attribute__((const)) FieldType fld_get_type(const FormField *fld) {
	if (! fld)	/* nothing there	*/
		return FT_NONE;

	return fld->type;
}

/** Get the name of the field.
 * \param fld
 * \return The name pointer. Dont free this pointer.	*/
const char *fld_get_name(const FormField *fld) {
	if (! fld)	/* nothing there	*/
		return NULL;

	return fld->name;
}

/** Get the value of the selected field in the list based at 'fld'.
 * \param fld	The base of the field list.
 * \return The field value string.	*/
static const char *fld_get_value(const FormField *fld) {
	if (! fld)
		return NULL;

	return fld->value;
}

/** Returns the value of field with the name 'nm'.	*/
const char *fld_get_value_by_name(const FormField *fld, const char *nm) {
	if ((! fld) || (! nm))
		return NULL;

	if ((fld->name) && (!strcmp(nm, fld->name)))
		return fld->value;

	return fld_get_value_by_name(fld->next, nm);
}

/** Sets the edit mode for the specified field.
 * \param fld	The field to modify.
 * \param typ The new type.	*/
void fld_set_type(FormField *fld, FieldType typ) {
	if ((! fld) || (typ < FT_MIN) || (typ > FT_MAX))
		return;

	fld->type = typ;
}

/** Set the help text for the field.
 * \param	fld	The field to set the help on.
 * \param	hlp	The help text.	*/
void fld_set_help(FormField *fld, const char *hlp) {
	if (! fld)
		return;

	if (fld->help)
		free(fld->help);

	if (hlp)
		fld->help = strdup(hlp);
	else
		fld->help = NULL;
}

/** Set the relevance of the field.
 * \param	fld	The field to set the help on.
 * \param	r	The relevance to set.	*/
void fld_set_relevance(FormField *fld, FieldRelevance r) {
	if (! fld)
		return;

	if ((r >= FT_MIN) && (r <= FT_MAX))
		fld->rel = r;
}

/** Set the 'tooltip' text for the field.
 * \param	fld	The field to set the help on.
 * \param	txt	The tooltip text.	*/
void fld_set_tip(FormField *fld, const char *txt) {
	if (! fld)
		return;

	if (fld->tip) {
		free(fld->tip);
		fld->tip = NULL;
	}

	if (txt)
		fld->tip = strdup(txt);
}

/** Sets a field value.	*/
FormField *fld_find_by_name(const FormField *fld, const char *nm) {
	if ((! fld) || (! nm))
		return NULL;

	if ((fld->name) && (!strcmp(nm, fld->name)))
		return (FormField *)fld;

	return fld_find_by_name(fld->next, nm);
}

/** Sets a field value.	*/
FormField *fld_set_value_by_name(FormField *fld, const char *nm, const char *val) {
	FormField *foundfield = fld_find_by_name(fld, nm);
	fld_set_value(foundfield, val);

	return foundfield;
}

/** Returns a pointer to the last added field.	*/
const FormField *fld_get_last(const FormField *fld) {
	const FormField *f = fld;

	if (! f)
		return NULL;

	while (f->next)
		f = f->next;

	return f;
}

/** Returns a pointer to the last added field.	*/
const FormField *frm_last_field(const Form *frm) {
	if (! frm)
		return NULL;

	return fld_get_last(frm->fld);
}

/** Adds a form field. Returns a pointer to the added field.	*/
FormField *frm_add_field(Form *frm, const char *nm, const char *val) {
	FormField *f = NULL;

	if (! frm)
		return NULL;

	frm->fld = fld_add(frm->fld, nm);
	f = (FormField *)frm_last_field(frm);
	if (f)
		fld_set_value(f, val);

	return f;
}

/** Deletes all the fields in a form. Internal function.
 * \param frm The form to operate on.	*/
void frm_clear(Form *frm) {
	if (! frm)
		return;

	frm_set_title(frm, NULL);
	frm->fld = fld_free(frm->fld);
}

Form *frm_free(Form *frm) {
	frm_clear(frm);
	return NULL;
}

/** Copies one form to another.
 * \param dst The destination form.
 * \param src The source form.	*/
void frm_copy(Form *dst, const Form *src) {
	const FormField *fld = NULL;
	int fields;
	const char *name = NULL;
	const char *val = NULL;

	if ((! dst) || (! src))
		return;

	frm_clear(dst);
	frm_set_title(dst, frm_get_title(src));

	fields = frm_get_field_count(src);

	fld = frm_get_field(src, 0);
	while (fld) {
		name = fld_get_name(fld);
		if (name) {
			val = fld_get_value_by_name(fld, name);
			frm_add_field(dst, name, val);
		}
		fld = fld->next;
	}
}

/** Checks that all required data is present.
 * \param	frm	The form data to use.
 * \return A Buf containing an error message, or NULL if no error.
 * If a Buf is returned, the caller should free it after use.	*/
ABuf *frm_check_fields(Form * frm)
{
	const FormField *fld = NULL;

	int i;
	int fieldcount;

	const char *value = NULL;
	ABuf *errmsg = NULL;
	FieldRelevance rel;

	/* Creates the notebook pane and label */
	/* The connection pane */
	fieldcount = frm_get_field_count(frm);
	for (i = 0; i < fieldcount; i++) {
		fld = frm_get_field(frm, i);
		rel = fld_get_relevance(fld);
		if (rel == FR_REQUIRED) {
			value = fld_get_value(fld);
			if ((value == NULL) || (strlen(value) == 0)) {
				/* No good.	*/
				errmsg = buf_strcat(errmsg, fld_get_name(fld));
				errmsg = buf_strcat(errmsg, ", ");
			}
		}
	}
	if (errmsg != NULL) {
		/* Remove the trailing ", "	*/
		errmsg = buf_str_shrinkbyone(errmsg);
		errmsg = buf_str_shrinkbyone(errmsg);
	}
	return errmsg;
}

/** Clears the values of all fields, but doesn't free them.
 * \param frm The form to operate on.	*/
void frm_clear_fields(Form *frm) {
	FormField *fld = NULL;

	if (frm == NULL)
		return;

	fld = frm->fld;
	while (fld) {
		fld_set_value(fld, NULL);
		fld = fld->next;
	}
}

/** Returns the edit mode of the field 'n'.	*/
const FormField *frm_get_field(const Form *frm, int n) {
	if (! frm)
		return NULL;

	return fld_get(frm->fld, n);
}

/** Returns the count of active fields.	*/
int frm_get_field_count(const Form *frm) {
	if (! frm)
		return 0;

	return fld_count(frm->fld);
}

/** Returns the name of the selected field.	*/
const char *frm_get_field_name(const Form *frm, int i) {
	const FormField *fld = NULL;

	if (! frm)
		return NULL;

	fld = frm_get_field(frm, i);
	return fld_get_name(fld);
}

/** Returns the form title.	*/
const char *frm_get_title(const Form *frm) {
	if (! frm)
		return NULL;

	return frm->title;
}

/** Returns the value of field with the name 'nm'.	*/
const char *frm_get_value_by_name(const Form *frm, const char *nm) {
	if (! frm)
		return NULL;

	return fld_get_value_by_name(frm->fld, nm);
}

/** Sets a field value from a form file line.
 * \param frm	The form to operate on.
 * \param nm	The field name.
 * \param val	The field value, tip text(optional) and help text (optional).
 * These are in the format:
 * value[|hint[|help]].	*/
void frm_define_field(Form *frm, const char *nm, const char *val) {
	FormField *fld = NULL;
	const char *value = NULL;
	const char *tip = NULL;
	const char *help = NULL;
	char *pipe = NULL;
	const char *start = val;

	/* Find value.	*/
	pipe = strchr(start, '|');
	value = val;
	if (pipe == NULL) {
		/* Whole string must be value.	*/
		goto done;
	}

	/* terminate value string.	*/
	*pipe = '\0';

	/* Look for tip.	*/
	start = pipe;
	start++;

	if (strlen(start) == 0)
		/* No tip text.	*/
		goto done;

	tip = start;

	/* Look for help	*/
	pipe = strchr(start, '|');
	if (pipe == NULL) {
		/* Rest of string Whole string must be tip.	*/
		goto done;	
	}

	/* terminate tip	*/
	*pipe = '\0';
	
	start = pipe;
	start++;

	if (strlen(start) == 0)
		goto done;

	help = start;

done:
	fld = frm_add_field(frm, nm, val);
	if (tip != NULL)
		fld_set_tip(fld, tip);

	if (help != NULL)
		fld_set_help(fld, help);
}

/** Print the form values for debugging.
 * \param frm	The form to print.	*/
void frm_print(const Form *frm) {
	FormField *fld = NULL;
	const char *title = NULL;

	if (frm == NULL)
		return;

	title = frm_get_title(frm);
	if (title)
		printf("Title: %s\n", frm_get_title(frm));
	else
		printf("Title: %s\n", "(none)");
	
	printf("=Fields=\n");
	fld = frm->fld;
	while (fld != NULL) {
		printf(" %s: %s|%s|%s\n", fld_get_name(fld), fld_get_value(fld), fld_get_tip(fld), fld_get_help(fld));
	}
}

/** Load a Form file from disk.
 * \param frm	The form to load to.
 * \param fn	Filename to load. If NULL, returns without doing anything, otherwise
 * clears the current form fields and attempts to load the file. If there's an
 * error, sets the "frm_error" field to an error string.
 * \return Zero on success, non-zero on error.	*/
int frm_load_file(Form *frm, char *fn)
{
	FILE *ff;
	char buf[1024];
	char key[LONGEST_KEY + 1];
	char *val;

	if ((frm == NULL) || (fn == NULL))
		return 1;

	frm_clear(frm);

	ff = fopen(fn, "r");
	if (!ff) {
		perror("opening form file for reading");
		frm_set_value_by_name(frm, FRM_ERROR, "Could not open form file");
		return 1;
	}

	while (fgets(buf, sizeof(buf), ff)) {
		char *colon = strchr(buf, ':');
		if (colon) {
			*colon = '\0';
			strncpy(key, buf, sizeof(key)-1);
			key[LONGEST_KEY] = '\0';
			val = colon;
			val++;

			/* Skip leading white space */
			while ((*val) && (isspace(*val)))
				val++;

			if (strlen(val) > 0) {
				frm_define_field(frm, key, val);
			}
		}
	}
	frm_print(frm);
	return 0;
}

/** Creates and returns a new form.	*/
Form *frm_new(const Form *frm) {
	Form *f = calloc(1, sizeof(Form));
	if (frm)
		frm_copy(f, frm);

	return f;
}

/** Sets the title for a form.	*/
Form *frm_set_title(Form *f, const char *ttl) {
	if (! f)
		return NULL;

	if (f->title) {
		free(f->title);
		f->title = NULL;
	}

	if (ttl)
		f->title = strdup(ttl);

	return f;	
}

/** Sets a field value.	*/
void frm_set_value_by_name(Form *frm, const char *nm, const char *val) {
	if (! frm)
		return;

	if (! fld_set_value_by_name(frm->fld, nm, val)) {
		/* didn't find field, so add it.	*/
		frm_add_field(frm, nm, val);
	}
}
