/** \addtogroup forms
 *@{
 */
/** \file forms.h	Form interface.
 * \author Darryl Luff	*/
/* GtkSQL -- an interactive graphical query tool for PostgreSQL
 * Copyright (C) 2003-2004 Darryl Luff
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifndef __FORMS_H
#define __FORMS_H

#include "buf.h"

/** Form Field type.	*/
typedef enum {
	FT_NONE,	/**<- Invalid field type.	*/
	FT_TEXT,	/**<- Field is a normal text field (default).	*/
	FT_PWD,		/**<- Field is a password field. Typed characters should be hidden in the user interface.	*/
	FT_FILENAME	/**<- Field is for a filename.	*/
} FieldType;
#define FT_MIN	FT_NONE
#define FT_MAX	FT_FILENAME

/** Field Relevance. Marks the field as optional or compulsory.	*/
typedef enum {
	FR_OPTIONAL,		/**<- Field is optional (default).	*/
	FR_REQUIRED		/**<- Field must be completed.	*/
} FieldRelevance;

/** The field name that holds form error messages.	*/
#define FRM_ERROR "frm_error"

/** Structure describing a single form field.	*/
typedef struct {
	char *name;		/**< Field name. Used for display.	*/
	FieldType type;	/**< Field type.	*/
	FieldRelevance rel;	/**< Field relevance.	*/
	char *value;		/**< Field value.	*/
	char *tip;		/**< Short tip for 'tooltips' type applications.	*/
	char *help;		/**< Longer help text.	*/
	void *next;		/**<- Link to next field in list.	*/
} FormField;

/** Structure describing a single form.	*/
typedef struct {
	/** Form title.	*/
	char *title;
	/** List of fields for this form.	*/
	FormField *fld;
} Form;

/* Field methods	*/
/** Frees the field list based at 'fld'.
 * \param fld	The base of the list.
 * \return Always returns NULL. This is so you can do 
 * "fields = fld_free(fields);" to NULL the pointer.	*/
FormField *fld_free(FormField *fld);
const char *fld_get_help(const FormField * fld);
const char *fld_get_name(const FormField * fld);
__attribute__((const)) FieldRelevance fld_get_relevance(const FormField * fld);
const char *fld_get_tip(const FormField * fld);
/** Returns the edit mode of the selected field in the list based at 'fld'.
 * \param fld	The field.
 * \return The field type.
 * \see FieldType
 * \see FT_NONE
 * \see FT_TEXT
 * \see FT_PWD
 * \see FT_FILENAME	*/
__attribute__((const)) FieldType fld_get_type(const FormField *fld);
 /** Create new field. Default properties are:
  * Type: FT_TEXT
  * Relevance: FR_OPTIONAL
  * \param nm Name for the new field (optional).
  */
FormField *fld_new(const char *nm);
void fld_set_relevance(FormField * fld, const FieldRelevance r);
void fld_set_tip(FormField * fld, const char *txt);
void fld_set_type(FormField * fld, const FieldType typ);
/** Sets the value of the specified field.
 * \param fld The field to set.
 * \param val The new value (may be NULL).	*/
void fld_set_value(FormField *fld, const char *val);

/* Form methods	*/
FormField *frm_add_field(Form * frm, const char *nm, const char *val);
ABuf *frm_check_fields(Form * frm);
void frm_clear_fields(Form * frm);
Form *frm_free(Form *frm);
const FormField *frm_get_field(const Form * frm, int n);
int frm_get_field_count(const Form * frm);
const char *frm_get_field_name(const Form * frm, int i);
const char *frm_get_title(const Form * frm);
const char *frm_get_value_by_name(const Form * frm, const char *nm);
Form *frm_new(const Form * frm);
Form *frm_set_title(Form * f, const char *ttl);
void frm_set_value_by_name(Form * frm, const char *nm, const char *val);

/*@}*/
#endif
