/** \file gtk_config.c	Configuration module.
 * \author Darryl Luff
 * \ingroup uigtk
 *
 * The appconfig module manages the gtksql configuration files.  At present
 * there are two different types of config files:
 * - The resource file ${HOME}/.gtksqlrc
 * - The config file ${HOME}/.gtksql/gtksql.conf
 *
 * The resource file is in the standard gtk resource format.
 * gtksql.conf is loaded every time the application runs.
 * Other scripts are loaded either from the command line, or from menus inside the application.
 */
/* GtkSQL -- an interactive graphical query tool for PostgreSQL
 * Copyright (C) 1998-2005  Lionel ULMER, Darryl Luff
 * $Id: gtk_config.c,v 1.8 2007/04/21 00:46:34 djluff Exp $
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Some config items are stored in a gtk resource file int he standard gtk format.
 * But gtk doesnt seem to allow you to put arbitrary things in there. So the 
 * remainder of things are stored in a simple text config file.
 *
 * This module handles the differences transparently, so the application doesn't need
 * to know about these seperate files.
 */

#include "../config.h"

#include <stdlib.h>
#include <string.h>
#include "appconfig.h"
#include "gtk_config.h"
#include "gtk_queries.h"
#include "gtk_support.h"
#include "utils.h"
#include "gtk_interface.h"
#include "gtksql_query_scintilla.h"
#include "gtkui-object.h"

/* Config strings	*/
#define BROWSE_AVAHI			"browse.avahi"
#define VIEW_TABLE_TREE			"view.tables.tree"
#define VIEW_TOOLBAR			"view.toolbar"
#define QUERY_NORMAL_COLOUR		"query.text.normal.colour"
#define QUERY_FIELD_COLOUR		"query.text.field.colour"
#define QUERY_SQL_COLOUR		"query.text.sql.colour"
#define QUERY_TABLE_COLOUR		"query.text.table.colour"
#define QUERY_FONT_NAME			"query.text.font.name"

/* Default settings	*/
#define DEF_BROWSE_AVAHI			TRUE
#define DEF_VIEW_TABLE_TREE			TRUE
#define DEF_VIEW_TOOLBAR			TRUE
#define DEF_VIEW_TOOLBAR_CONNECTION	TRUE
#define DEF_VIEW_TOOLBAR_QUERY		TRUE
#define DEF_VIEW_TOOLBAR_MISC		TRUE
#define DEF_QUERY_NORMAL_COLOUR		"black"
#define DEF_QUERY_FIELD_COLOUR		"green"
#define DEF_QUERY_SQL_COLOUR		"blue"
#define DEF_QUERY_TABLE_COLOUR		"darkred"

/** Data structure containing config items.	*/
typedef struct _gtkconfig {
	gchar *query_normal_colour;
	gchar *query_field_colour;
	gchar *query_sql_colour;
	gchar *query_table_colour;
	gchar *query_font_name;
	int browse_avahi;
} GtkConfig;

/** Global app config.	*/
GtkConfig gtkcfg;

void gtkcfg_set_browse_avahi(int ba) {
	//g_warning("XXX gtkcfg_set_browse_avahi(%d)\n", ba);
	gtkcfg.browse_avahi = ba;
}

static void gtkcfg_set_query_field_colour(const gchar *c) {
	/* Free the current value	*/
	if (gtkcfg.query_field_colour)
		free(gtkcfg.query_field_colour);

	/* Use strdup to set it to either the default or the supplied
	 * value, so i can be freed later either way	*/
	if (c == NULL)
		gtkcfg.query_field_colour = strdup(DEF_QUERY_FIELD_COLOUR);
	else	
		gtkcfg.query_field_colour = strdup(c);

	gtksql_query_scintilla_class_set_field_colour(gtkcfg.query_field_colour);
}

static void gtkcfg_set_query_normal_colour(const char *c) {
	if (gtkcfg.query_normal_colour)
		free(gtkcfg.query_normal_colour);

	if (c == NULL)
		gtkcfg.query_normal_colour = strdup(DEF_QUERY_NORMAL_COLOUR);
	else
		gtkcfg.query_normal_colour = strdup(c);

	gtksql_query_scintilla_class_set_normal_colour(gtkcfg.query_normal_colour);
}

static void gtkcfg_set_query_sql_colour(const char *c) {
	if (gtkcfg.query_sql_colour)
		free(gtkcfg.query_sql_colour);

	if (c == NULL)
		gtkcfg.query_sql_colour = strdup(DEF_QUERY_SQL_COLOUR);
	else
		gtkcfg.query_sql_colour = strdup(c);

	gtksql_query_scintilla_class_set_sql_colour(gtkcfg.query_sql_colour);
}

static void gtkcfg_set_query_font_name(const char *f) {
	if (gtkcfg.query_font_name)
		free(gtkcfg.query_font_name);

	if (f == NULL)
		gtkcfg.query_font_name = NULL;
	else
		gtkcfg.query_font_name = strdup(f);

	gtksql_query_scintilla_class_set_font_name(gtkcfg.query_font_name);
}

static void gtkcfg_set_query_table_colour(const char *c) {
	if (gtkcfg.query_table_colour)
		free(gtkcfg.query_table_colour);

	if (c == NULL)
		gtkcfg.query_table_colour = strdup(DEF_QUERY_TABLE_COLOUR);
	else
		gtkcfg.query_table_colour = strdup(c);

	gtksql_query_scintilla_class_set_table_colour(gtkcfg.query_table_colour);
}
	
const char *gtkcfg_get_query_field_colour(void) {
	if (! gtkcfg.query_field_colour)
		gtkcfg_set_query_field_colour(NULL);

	return gtkcfg.query_field_colour;
}

const char *gtkcfg_get_query_font_name(void) {
	return gtkcfg.query_font_name;
}

const char *gtkcfg_get_query_normal_colour(void) {
	if (! gtkcfg.query_normal_colour)
		gtkcfg_set_query_normal_colour(NULL);

	return gtkcfg.query_normal_colour;
}

const char *gtkcfg_get_query_sql_colour(void) {
	if (! gtkcfg.query_sql_colour)
		gtkcfg_set_query_sql_colour(NULL);

	return gtkcfg.query_sql_colour;
}

const char *gtkcfg_get_query_table_colour(void) {
	if (! gtkcfg.query_table_colour)
		gtkcfg_set_query_table_colour(NULL);

	return gtkcfg.query_table_colour;
}

static void gtkcfg_set_view_toolbar(gboolean t) {
	GtkWidget *mnuViewToolbar = gtkui_get_mnuViewToolbar(gtkapp);

	if (mnuViewToolbar)
		gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(mnuViewToolbar), t);
}

/** Set the default configuration values.	*/
static void gtkcfg_set_defaults(void)
{
	/* Integer settings. Set these directly	*/
	gtkcfg_set_browse_avahi(DEF_BROWSE_AVAHI);
	gtkcfg_set_view_table_tree(DEF_VIEW_TABLE_TREE);
	gtkcfg_set_view_toolbar(DEF_VIEW_TOOLBAR);

	/* Strings. Set them properly	*/
	gtkcfg_set_query_normal_colour(DEF_QUERY_NORMAL_COLOUR);
	gtkcfg_set_query_field_colour(DEF_QUERY_FIELD_COLOUR);
	gtkcfg_set_query_sql_colour(DEF_QUERY_SQL_COLOUR);
	gtkcfg_set_query_table_colour(DEF_QUERY_TABLE_COLOUR);
	gtkcfg_set_query_font_name(NULL);
}

/** Set the default configuration values.	*/
 void gtkcfg_clear(void)
{
	/* Free strings if set.	*/
	gtkcfg_set_query_normal_colour(NULL);
	gtkcfg_set_query_field_colour(NULL);
	gtkcfg_set_query_sql_colour(NULL);
	gtkcfg_set_query_table_colour(NULL);
	gtkcfg_set_query_font_name(NULL);
}

void gtkcfg_config_value(const char *key, const char *val) {
	if (! strncmp(BROWSE_AVAHI, key, strlen(BROWSE_AVAHI)))
		gtkcfg_set_browse_avahi(atoi(val));
	else if (! strncmp(VIEW_TABLE_TREE, key, strlen(VIEW_TABLE_TREE)))
		gtkcfg_set_view_table_tree(atoi(val));
	else if (! strncmp(VIEW_TOOLBAR, key, strlen(VIEW_TOOLBAR)))
		gtkcfg_set_view_toolbar(atoi(val));
	else if (! strncmp(QUERY_FIELD_COLOUR, key, strlen(QUERY_FIELD_COLOUR)))
		gtkcfg_set_query_field_colour(val);
	else if (! strncmp(QUERY_NORMAL_COLOUR, key, strlen(QUERY_NORMAL_COLOUR)))
		gtkcfg_set_query_normal_colour(val);
	else if (! strncmp(QUERY_SQL_COLOUR, key, strlen(QUERY_SQL_COLOUR)))
		gtkcfg_set_query_sql_colour(val);
	else if (! strncmp(QUERY_TABLE_COLOUR, key, strlen(QUERY_TABLE_COLOUR)))
		gtkcfg_set_query_table_colour(val);
	else if (! strncmp(QUERY_FONT_NAME, key, strlen(QUERY_FONT_NAME)))
		gtkcfg_set_query_font_name(val);
}

int gtkcfg_get_view_table_tree(void) {
	GtkWidget *mnuViewTreeview = gtkui_get_mnuViewTreeview(gtkapp);

	if (mnuViewTreeview)
		return gtk_check_menu_item_get_active(GTK_CHECK_MENU_ITEM(mnuViewTreeview));
	else
		return DEF_VIEW_TABLE_TREE;
}

int gtkcfg_get_browse_avahi(void) {
	//g_warning("gtkcfg_get_browse_avahi(): %d\n", gtkcfg.browse_avahi);
	return gtkcfg.browse_avahi;
}

static gboolean gtkcfg_get_view_toolbar(void) {
	GtkWidget *mnuViewToolbar = gtkui_get_mnuViewToolbar(gtkapp);

	if (mnuViewToolbar)
		return gtk_check_menu_item_get_active(GTK_CHECK_MENU_ITEM(mnuViewToolbar));
	else
		return DEF_VIEW_TOOLBAR;
}

/** Initialise the configuration.
 * Performs these steps:
 *  - Sets default values.	*/
void gtkcfg_init(void)
{
	gtkcfg_set_defaults();
}

/** Dump config settings.	*/
void gtkcfg_dump_config(void)
{
	g_message("%s: %d\n", BROWSE_AVAHI, gtkcfg_get_browse_avahi());
	g_message("VIEW_TABLE_TREE: %d\n", gtkcfg_get_view_table_tree());
	g_message("VIEW_TOOLBAR: %d\n", gtkcfg_get_view_toolbar());
	g_message("QUERY_FIELD_COLOUR: %s\n", gtkcfg_get_query_field_colour());
	g_message("QUERY_NORMAL_COLOUR: %s\n", gtkcfg_get_query_normal_colour());
	g_message("QUERY_SQL_COLOUR: %s\n", gtkcfg_get_query_sql_colour());
	g_message("QUERY_TABLE_COLOUR: %s\n", gtkcfg_get_query_table_colour());
	g_message("QUERY_FONT_NAME: %s\n", gtkcfg_get_query_font_name());
}

/** Save the config settings to a file.
 * \return	An integer error code. -1 = error opening file. 0 = success.	*/
void gtkcfg_save_file(void)
{
	//g_warning("XXX gtkcfg_save_file()\n");
	cfg_save_bool(BROWSE_AVAHI, gtkcfg_get_browse_avahi());
	cfg_save_bool(VIEW_TABLE_TREE, gtkcfg_get_view_table_tree());
	cfg_save_bool(VIEW_TOOLBAR, gtkcfg_get_view_toolbar());
	cfg_save_string(QUERY_FIELD_COLOUR, gtkcfg_get_query_field_colour());
	cfg_save_string(QUERY_NORMAL_COLOUR, gtkcfg_get_query_normal_colour());
	cfg_save_string(QUERY_SQL_COLOUR, gtkcfg_get_query_sql_colour());
	cfg_save_string(QUERY_TABLE_COLOUR, gtkcfg_get_query_table_colour());
	cfg_save_string(QUERY_FONT_NAME, gtkcfg_get_query_font_name());
}

void gtkcfg_set_view_table_tree(gboolean v) {
	GtkWidget *mnuViewTreeview = gtkui_get_mnuViewTreeview(gtkapp);
	if (mnuViewTreeview)
		gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(mnuViewTreeview), v);
}

/* where's the opposite to gdk_color_parse?? */
static gchar *colour_to_string(GdkColor *color) {
	const char *acColourFormat = "#%02X%02X%02X";
	static char buff[16];
	snprintf(buff, sizeof(buff), acColourFormat, color->red >> 8, color->green >> 8, color->blue >> 8);

	return buff;
}

static void gtkcfg_copy_from_dialog(void) {
	GdkColor colour;
	GtkWidget *autosaveChk;
	GtkWidget *browseAvahiChk;
	GtkWidget *configDlg = gtkui_get_configDlg(gtkapp);
	GtkWidget *fieldColourBtn;
	GtkWidget *normalColourBtn;
	GtkWidget *queryFontBtn;
	GtkWidget *sqlColourBtn;
	GtkWidget *tableColourBtn;

	if (! configDlg)
		return;

	autosaveChk = lookup_widget (configDlg, "autosaveChk");
	browseAvahiChk = lookup_widget (configDlg, "browseAvahiChk");
	fieldColourBtn = lookup_widget (configDlg, "fieldColourBtn");
	normalColourBtn = lookup_widget (configDlg, "normalColourBtn");
	sqlColourBtn = lookup_widget (configDlg, "sqlColourBtn");
	tableColourBtn = lookup_widget (configDlg, "tableColourBtn");
	queryFontBtn = lookup_widget (configDlg, "queryFontBtn");

	cfg_set_autosave(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(autosaveChk)));
	gtkcfg_set_browse_avahi(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(browseAvahiChk)));

	gtk_color_button_get_color(GTK_COLOR_BUTTON(fieldColourBtn), &colour);
	gtkcfg_set_query_field_colour(colour_to_string(&colour));

	gtk_color_button_get_color(GTK_COLOR_BUTTON(normalColourBtn), &colour);
	gtkcfg_set_query_normal_colour(colour_to_string(&colour));

	gtk_color_button_get_color(GTK_COLOR_BUTTON(sqlColourBtn), &colour);
	gtkcfg_set_query_sql_colour(colour_to_string(&colour));

	gtk_color_button_get_color(GTK_COLOR_BUTTON(tableColourBtn), &colour);
	gtkcfg_set_query_table_colour(colour_to_string(&colour));

	gtkcfg_set_query_font_name(gtk_font_button_get_font_name(GTK_FONT_BUTTON(queryFontBtn)));
}

static void gtkcfg_copy_to_dialog(void) {
	GtkWidget *autosaveChk;
	GtkWidget *browseAvahiChk;
	GtkWidget *normalColourBtn;
	GtkWidget *fieldColourBtn;
	GtkWidget *sqlColourBtn;
	GtkWidget *tableColourBtn;
	GtkWidget *queryFontBtn;
	GdkColor colour;
	GtkWidget *configDlg = gtkui_get_configDlg(gtkapp);
	const gchar *fontName;

	//g_warning("XXX gtkcfg_copy_to_dialog()\n");
	if (! configDlg)
		return;

	autosaveChk = lookup_widget (configDlg, "autosaveChk");
	browseAvahiChk = lookup_widget (configDlg, "browseAvahiChk");
	fieldColourBtn = lookup_widget (configDlg, "fieldColourBtn");
	normalColourBtn = lookup_widget (configDlg, "normalColourBtn");
	sqlColourBtn = lookup_widget (configDlg, "sqlColourBtn");
	tableColourBtn = lookup_widget (configDlg, "tableColourBtn");
	queryFontBtn = lookup_widget (configDlg, "queryFontBtn");

	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(autosaveChk), cfg_get_autosave());

#ifdef USE_AVAHI
	//g_warning("XXX setting browseAvahiChk\n");
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(browseAvahiChk), gtkcfg_get_browse_avahi());
#else
	//g_warning("XXX disabling browseAvahiChk\n");
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(browseAvahiChk), FALSE);
	gtk_widget_set_sensitive(GTK_WIDGET(browseAvahiChk), FALSE);
#endif

	if (gdk_color_parse(gtkcfg_get_query_field_colour(), &colour) == TRUE)
		gtk_color_button_set_color(GTK_COLOR_BUTTON(fieldColourBtn), &colour);

	if (gdk_color_parse(gtkcfg_get_query_normal_colour(), &colour) == TRUE)
		gtk_color_button_set_color(GTK_COLOR_BUTTON(normalColourBtn), &colour);

	if (gdk_color_parse(gtkcfg_get_query_sql_colour(), &colour) == TRUE)
		gtk_color_button_set_color(GTK_COLOR_BUTTON(sqlColourBtn), &colour);

	if (gdk_color_parse(gtkcfg_get_query_table_colour(), &colour) == TRUE)
		gtk_color_button_set_color(GTK_COLOR_BUTTON(tableColourBtn), &colour);

	fontName = gtkcfg_get_query_font_name();
	if (fontName)
		gtk_font_button_set_font_name(GTK_FONT_BUTTON(queryFontBtn), fontName);
}

void gtk_config_set_all_styles() {
	GtksqlQueryScintilla *query;
	int n;

	n = 0;
	while ((query = query_get_at(n))) {
		gtksql_query_scintilla_set_styles(query);
		n++;
	}
}

void gtkcfg_show_dialog(void) {
	GtkWidget *configDlg = gtkui_get_configDlg(gtkapp);

	if (! configDlg)
		configDlg = create_configDlg();

	if (! configDlg)
		return;

	gtkcfg_copy_to_dialog();

	gint result = gtk_dialog_run (GTK_DIALOG (configDlg));
	switch (result) {
		case GTK_RESPONSE_OK:
			gtkcfg_copy_from_dialog();
			cfg_save_file();
			gtk_config_set_all_styles();
			break;
		default:
			break;
	}
	gtk_widget_hide(configDlg);
}
