/** \file gtk_queries.c	Functions to handle queries.
 * \author Lionel ULMER
 * \author Darryl Luff
 * \addtogroup uigtk	*/
/*
 * GtkSQL -- an interactive graphical query tool for PostgreSQL and MySQL.
 * Copyright (C) 1998-2006  Lionel ULMER, Darryl Luff.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <string.h>
#include "gtk_callbacks.h"
#include "gtk_interface.h"
#include "gtk_queries.h"
#include "gtk_query_import_dlg.h"
#include "gtkui-object.h"
//#include "gtk_widgets.h"
#include "gui_gtk.h"
#include "utils.h"

#define STATUS_SIZE 64

#define IS_SEPARATOR(c) ((c == ' ')  || (c == '.')  || (c == '\r') || \
			 (c == '\n') || (c == '\t') || (c == ',')  || \
			 (c == '(')  || (c == '('))

/** Set the name of the query page. If name is NULL or an empty string,
 * the call is ignored. Otherwise, the name is set as specified.
 * If the query has been added to the queryNotebook, the notebook tabs
 * are updated as well. 
 * \param qry	The query to set.
 * \param name	The new name.	*/
void query_set_name(GtksqlQueryScintilla * qry, const gchar * name)
{
	if (!qry)
		return;

	if ((name != NULL) && (strlen(name) != 0)) {
		gtksql_query_scintilla_set_name(qry, name);
		gtkui_set_query_name(gtkapp, qry, name);
	}
}

/** Return the query text widget selected by 'n'. 'n' is zero-based.
 * \return	The result object from the current query.	*/
void *query_get_current_result(void)
{
	return gtksql_query_scintilla_get_result(query_get_current());
}

/** Return a copy of the text from the passed query.
 * \param query	The query widget.
 * \return A copy of the text. This must be freed with g_free when finished.	*/
const char *query_get_text(GtksqlQueryScintilla * query)
{
	GtkTextBuffer *buffer;
	GtkTextIter start;
	GtkTextIter end;
	gchar *txt;

	if (query == NULL)
		return NULL;

	buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(query));
	if (buffer == NULL)
		return NULL;

	gtk_text_buffer_get_bounds(buffer, &start, &end);
	txt = gtk_text_buffer_get_text(buffer, &start, &end, FALSE);

	return txt;
}

/** Save the query to a file.	*/
void query_save_to_file(GtksqlQueryScintilla * query, const char *fn)
{
	FILE *fp;
	const char *cquery;
	const char *name;

	if ((query == NULL) || (fn == NULL)) {
		g_warning("query_save_to_file(%p, %p)\n", query, fn);
		return;
	}

	if (strlen(fn) < 1) {
		g_warning("query_save_to_file(). Ignoring null filename\n");
		return;
	}

	cquery = gtksql_query_scintilla_get_text(query);

	fp = fopen(fn, "wb");
	if (fp != NULL) {
		fprintf(fp, "%s\n", cquery);
		fclose(fp);
		name = get_short_filename(fn);
		if (name == NULL)
			name = "Untitled";

		gtkgui_set_query_name(name);
		gtksql_query_scintilla_set_save_point(query);
	}
}

/** Close the current query.
 * \return The number of queries remaining.	*/
static int query_close_now(GtksqlQueryScintilla * qry)
{
	GtkWidget *queryNotebook;

	queryNotebook = gtkui_get_queryNotebook(gtkapp);\

	if (!qry) {
		g_warning("query_close_now(): qry is NULL.\n");
		return gtk_notebook_get_n_pages(GTK_NOTEBOOK(queryNotebook));
	}
	gtk_container_remove(GTK_CONTAINER(queryNotebook), GTK_WIDGET(qry));

	return gtk_notebook_get_n_pages(GTK_NOTEBOOK(queryNotebook));
}

/** Close the current query.
 * \return An appropriate GTK_RESPONSE code.	*/
gint query_close(void)
{
	GtksqlQueryScintilla *qry;

	qry = query_get_current();
	if (!qry) {
		g_warning("query_close(): qry is NULL.\n");
		return GTK_RESPONSE_CANCEL;
	}

	query_close_now(qry);
	return GTK_RESPONSE_CANCEL;
}

/** \return The GtkTextView of the current query.	*/
GtksqlQueryScintilla *query_get_current()
{
	GtkWidget *queryNotebook = gtkui_get_queryNotebook(gtkapp);

	if (! queryNotebook)
		return NULL;

	int cur_page = gtk_notebook_get_current_page(GTK_NOTEBOOK(queryNotebook));
	GtkWidget *query = gtk_notebook_get_nth_page(GTK_NOTEBOOK(queryNotebook), cur_page);

	return GTKSQL_QUERY_SCINTILLA(query);
}

/** Retrieve the name (title) for a query.
 * \param qry	The query.	*/
const gchar *query_get_name(GtksqlQueryScintilla * qry)
{
	if (qry)
		return gtksql_query_scintilla_get_name(qry);
	else
		return gtksql_query_scintilla_get_name(query_get_current());
}

/** Return the query text widget selected by 'n'. 'n' is zero-based.
 * \param n	The index of the entry to return.
 * \return	The selected query text widget.	*/
GtksqlQueryScintilla *query_get_at(int n)
{
	GtkWidget *queryNotebook = gtkui_get_queryNotebook(gtkapp);

	if (! queryNotebook)
		return NULL;

	/* The notebook page is the scroller. */
	return GTKSQL_QUERY_SCINTILLA(gtk_notebook_get_nth_page(GTK_NOTEBOOK(queryNotebook), n));
}

/** Initialise the query notebook. Removes the dummy page added in glade. Also
 * sets the widget name used with styles.	*/
void queries_init()
{
	GtkWidget *queryNotebook = gtkui_get_queryNotebook(gtkapp);

	if (! queryNotebook)
		return;

	/* The query notebook */
	gtk_widget_set_name(GTK_WIDGET(queryNotebook), "queries");

	while (gtk_notebook_get_n_pages(GTK_NOTEBOOK(queryNotebook)) > 0)
		gtk_notebook_remove_page(GTK_NOTEBOOK(queryNotebook), 0);

	gtk_widget_show(GTK_WIDGET(queryNotebook));
}

/** Import a query from a file.	*/
void query_import(void)
{
	gtkui_show_queryImportDlg(gtkapp);
}

/** \return	The number of queries.	*/
int query_count()
{
	GtkWidget *queryNotebook = gtkui_get_queryNotebook(gtkapp);

	if (queryNotebook == NULL)
		return 0;

	return gtk_notebook_get_n_pages(GTK_NOTEBOOK(queryNotebook));
}

/** Save a query to disk.
 * \param qry	The query to save. If NULL, the current query is saved.	*/
void query_save(GtksqlQueryScintilla * qry)
{
	GtksqlQueryScintilla *query;

	if (qry == NULL)
		query = query_get_current();
	else
		query = qry;

	if (query == NULL) {
		g_warning("query_save(). query not valid\n");
		return;
	}

	query_save_as(NULL);
}

/** Save the current query to disk.	*/
void query_save_as(GtksqlQueryScintilla * qry)
{
	GtkWidget *querySaveDlg = gtkui_get_querySaveDlg(gtkapp);

	if (querySaveDlg == NULL)
		g_warning("query_save_as(%p): Dialog not initialised\n", qry);
	else {
		g_object_set_data(G_OBJECT(querySaveDlg), QUERY_QRY_KEY, qry);
		gtk_widget_show(querySaveDlg);
	}
}

/** Display a query. Sets up the result etc.	*/
static void query_display(GtksqlQueryScintilla *qry)
{
	GtkWidget *label;
	const gchar *name;
	GtkWidget *queryNotebook = gtkui_get_queryNotebook(gtkapp);

	if (! queryNotebook)
		return;

	if (qry == NULL) {
		g_warning("query_display(%p): query is null\n", qry);
		return;
	}

	/* Add the query to the notebook */
	name = gtksql_query_scintilla_get_name(qry);
	if (name == NULL)
		name = "Query";

	label = gtk_label_new(name);
	gtk_widget_show(GTK_WIDGET(label));
	gtk_notebook_append_page(GTK_NOTEBOOK(queryNotebook), GTK_WIDGET(qry), label);
	gtk_notebook_set_current_page(GTK_NOTEBOOK(queryNotebook), -1);

	gtk_widget_show(GTK_WIDGET(qry));
	gtk_widget_show(GTK_WIDGET(queryNotebook));
	update_ui();
}

/** \return The scrolling view containing the new query text widget.	*/
void query_add(void)
{
	ABuf *buf = buf_new(64);
	GtksqlQueryScintilla *query = GTKSQL_QUERY_SCINTILLA(gtksql_query_scintilla_new());

	/* Set the query name.  */
	snprintf((char *)buf->b_dat, buf->b_len, "Query%d", query_count());
	gtksql_query_scintilla_set_name(query, (char *)buf->b_dat);
	gtk_widget_show_all(GTK_WIDGET(query));

	/* And display it.  */
	query_display(query);

	buf_free(buf);
}
