/** \file gtk_treeview.c	Table tree view.
 * \author Darryl Luff
 * \addtogroup uigtk
 *
 * The tree view uses three levels:
 *
 * database
 * |
 *  - table
 *  |
 *   - field
 *
 *
 * This is just for me to remember how to access columns:
 * static void tree_selection_changed_cb (GtkTreeSelection *selection, gpointer data) {
 *      GtkTreeIter iter;
 *      GtkTreeModel *model;
 *      gchar *author;
 *
 *      if (gtk_tree_selection_get_selected (selection, &model, &iter)) {
 *              gtk_tree_model_get (model, &iter, AUTHOR_COLUMN, &author, -1);
 *              g_print ("You selected a book by %s\n", author);
 *              g_free (author);
 *      }
 * }
 *
 */

/* GtkSQL -- an interactive graphical query tool for PostgreSQL, MySQL and SQLite.
 * Copyright (C) 2004 Darryl Luff
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "../config.h"

#include <string.h>
#include "gtkui-object.h"
#include "guiapi.h"
#include "gtksql_query_scintilla.h"

/** Columns for display	*/
enum {
	PIXBUF_COLUMN,
	FIELD_COLUMN,
	TYPE_COLUMN,
} TreeviewColumns;

GdkPixbuf *dbPixbuf = NULL;
GdkPixbuf *tablePixbuf = NULL;
GdkPixbuf *fieldPixbuf = NULL;


/* Treating buf as a comma-separated list, returns true if the list contains key */
static gboolean buf_contains(const ABuf *buf, gchar *key) {
	char *s;

	if ((! buf) || (! key))
		return FALSE;

	s = strtok((char *)buf->b_dat, ",");
	while (s) {
		if (strcmp(s, key) == 0)
			return TRUE;

		s = strtok(NULL, ",");
	}
	return FALSE;
}

static GdkPixbuf *load_pixbuf(gchar * fn)
{
	GdkPixbuf *pbuf = NULL;
	GError *error = NULL;

	pbuf = gdk_pixbuf_new_from_file(fn, &error);
	if (error)
		g_warning("Error opening pixmap %s: %s\n", fn, error->message);

	return pbuf;
}

static void load_pixmaps(void)
{
	dbPixbuf = load_pixbuf(PACKAGE_PIXMAPS_DIR "/gtksql_db.png");
	tablePixbuf = load_pixbuf(PACKAGE_PIXMAPS_DIR "/gtksql_table.png");
	fieldPixbuf = load_pixbuf(PACKAGE_PIXMAPS_DIR "/gtksql_field.png");
}

/** Create the display for a single table.
 * \param tbd	The table definition.
 * \param model The GtkTreeStore the data is in
 * \param tableiter The node in the tree view that holds the table.	*/
static void
add_tree_fields(DBTableDef * tbd, GtkTreeStore * model,
				GtkTreeIter * tableiter)
{
	const char *line[3];
	GtkTreeIter listiter;
	const DBFieldDef *fld;
	int fieldCount;
	int i;
	GtkWidget *treeView = gtkui_get_treeView(gtkapp);

	if (! treeView)
		return;

	fieldCount = tabledef_fieldcount(tbd);
	for (i = 0; i < fieldCount; i++) {
		fld = tabledef_getfield(tbd, i);

		line[0] = fielddef_getname(fld);
		line[1] = fielddef_gettype(fld);
		gtk_tree_store_append(model, &listiter, tableiter);
		gtk_tree_store_set(model, &listiter, PIXBUF_COLUMN,
						   fieldPixbuf, FIELD_COLUMN, line[0],
						   TYPE_COLUMN, line[1], -1);
	}
	gtk_tree_view_columns_autosize(GTK_TREE_VIEW(treeView));
}

/** Iterator function for all selected rows	*/
static void
add_selected_tree_fields(GtkTreeModel * model, GtkTreePath * path,
						 GtkTreeIter * iter, gpointer data)
{
	gchar *name;
	gchar *table = NULL;
	gint depth;
	GtkTreeIter parentIter;
	ABuf *buf = *((ABuf **) data);
	ABuf *thisfield = buf_new(100);

	thisfield = buf_strcpy(thisfield, "");

	depth = gtk_tree_path_get_depth(path);
	gtk_tree_model_get(model, iter, FIELD_COLUMN, &name, -1);
	switch (depth) {
		case 1:	/* db name. Just use it.	*/
			thisfield = buf_strcat(thisfield, name);
			break;
		case 2:	/* table name. Use 'table.*'	*/
			thisfield = buf_strcat(thisfield, name);
			thisfield = buf_strcat(thisfield, ".*");
			break;
		case 3:	/* field name. use 'table.field'	*/
			if (gtk_tree_model_iter_parent(model, &parentIter, iter))
				gtk_tree_model_get(model, &parentIter, FIELD_COLUMN, &table, -1);

			if (table) {
				thisfield = buf_strcpy(thisfield, table);
				thisfield = buf_strcat(thisfield, ".");
			}
			thisfield = buf_strcat(thisfield, name);
			break;
	}
	if (buf_strlen(buf) > 0)
		buf = buf_strcat(buf, ", ");

	buf = buf_strcat(buf, (const char *)thisfield->b_dat);
	thisfield = buf_free(thisfield);

	*((ABuf **) data) = buf;
}

/** Iterator function for all selected rows	*/
static void
add_selected_tree_tables(GtkTreeModel * model, GtkTreePath * path,
						 GtkTreeIter * iter, gpointer data)
{
	gint depth;
	GtkTreeIter pariter;
	gchar *name = NULL;
	ABuf *buf = *((ABuf **) data);

	depth = gtk_tree_path_get_depth(path);
	if (depth == 2) {	/* table name	*/
		gtk_tree_model_get(model, iter, FIELD_COLUMN, &name, -1);
	}
	else if (depth == 3) {	/* field record	*/
		/* Get the parent	*/
		if (gtk_tree_model_iter_parent(model, &pariter, iter)) {
			gtk_tree_model_get(model, &pariter, FIELD_COLUMN, &name, -1);
		}
	}

	if (name != NULL) {
		if (! buf_contains(buf, name)) {
			if (buf_strlen(buf) < 1)
				buf = buf_strcpy(buf, name);
			else {
				buf = buf_strcat(buf, ",");
				buf = buf_strcat(buf, name);
			}
		}
	}

	*((ABuf **) data) = buf;
}

/** \return A list of the selected table fields. The list is returned as a comma-separated text string.
 * If no fields are selected, returns "*". This list MUST BE FREED by the caller when finished.	*/
gchar *treeview_get_selected_fields(void)
{
	GtkTreeSelection *selection;
	char *ret;
	ABuf *buf = buf_new(512);
	GtkWidget *treeView = gtkui_get_treeView(gtkapp);

	if (!treeView) {
		g_warning("get_selected_tree_fields(): treeView is NULL\n");
		return g_strdup("*");
	}

	selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(treeView));
	gtk_tree_selection_selected_foreach(GTK_TREE_SELECTION(selection),
										add_selected_tree_fields, &buf);
	if (buf_strlen(buf) < 1)
		buf = buf_strcpy(buf, "*");

	ret = g_strdup((char *)buf->b_dat);
	buf = buf_free(buf);

	return ret;
}

/** \return A list of the selected table fields. The list is returned as a comma-separated text string.
 * If no fields are selected, returns "*". This list MUST BE FREED by the caller when finished.	*/
gchar *treeview_get_selected_tables(void)
{
	GtkTreeSelection *selection;
	char *ret;
	ABuf *buf = buf_new(512);
	GtkWidget *treeView = gtkui_get_treeView(gtkapp);

	if (!treeView) {
		g_warning("get_selected_tree_tables(): treeView is NULL\n");
		return g_strdup("*");
	}

	selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(treeView));
	gtk_tree_selection_selected_foreach(GTK_TREE_SELECTION(selection),
										add_selected_tree_tables, &buf);
	if (buf_strlen(buf) < 1)
		buf = buf_strcpy(buf, "*");

	ret = g_strdup((char *)buf->b_dat);
	buf = buf_free(buf);

	return ret;
}

static void
tree_selected_foreach_func(GtkTreeModel * model, GtkTreePath * path,
						   GtkTreeIter * iter, gpointer data)
{
	gchar *name;
	ABuf *buf;

	if (data == NULL) {
		g_warning("tree_selected_foreach_func(): userdata is NULL\n");
		return;
	}

	buf = *((ABuf **) data);

	gtk_tree_model_get(model, iter, FIELD_COLUMN, &name, -1);
	if (buf_strlen(buf) == 0)		   /* First entry */
		buf = buf_strcpy(buf, name);
	else {							   /* Current entries. Add a "," */
		buf = buf_strcat(buf, ", ");
		buf = buf_strcat(buf, name);
	}

	*((ABuf **) data) = buf;
}

static void
tree_popup_menu(GtkWidget * treeview, GdkEventButton * event,
				gpointer userdata)
{
	/* Note: event can be NULL here when called from view_onPopupMenu;
	 * gdk_event_get_time() accepts a NULL argument */
	GtkWidget *treePopup = gtkui_get_treePopup(gtkapp);
	if (! treePopup)
		return;

	gtk_menu_popup(GTK_MENU(treePopup), NULL, NULL, NULL, NULL,
				   (event != NULL) ? event->button : 0,
				   gdk_event_get_time((GdkEvent *) event));
}

/** 'click' event for table pane.
 * Pops up the popup menu on button 3 being pressed.	*/
static gboolean
on_tree_button_press_event(GtkWidget * widget, GdkEventButton * event,
						   gpointer user_data)
{
	GtkWidget *treePopup = NULL;

	if (event->button == 3) {
		treePopup = gtkui_get_treePopup(gtkapp);
		if (treePopup == NULL) {
			g_warning("treePopup not initialised.\n");
			return FALSE;
		}
		tree_popup_menu(widget, NULL, user_data);
		return TRUE;
	}
	return FALSE;
}

static gboolean on_tree_popup_menu(GtkWidget * treeview, gpointer userdata)
{
	tree_popup_menu(treeview, NULL, userdata);
	return TRUE;
}

/* \todo get_selection_string() needs to be modified to prepend table names
 * to selected field names.	*/
ABuf *get_selection_stringX(GtkWidget *widget) {
	GtkTreeSelection *selected;
	GtkTreeView *view;
	ABuf *fields = buf_new(1024);

	if (!fields) {
		g_warning("sendCallback() could not allocate memory for a buffer\n");
		return NULL;
	}

	view = GTK_TREE_VIEW(widget);
	if (view == NULL) {
		g_warning("send_callback(). no send widget\n");
		return NULL;
	}

	selected = gtk_tree_view_get_selection(view);
	if (selected == NULL) {
		g_warning("send_callback(). no send selection\n");
		return NULL;
	}

	/* Loop through the selections */
	fields = buf_strcpy(fields, "");
	gtk_tree_selection_selected_foreach(selected, tree_selected_foreach_func, &fields);

	return fields;
}

void treeview_do_copy(GtkWidget *widget) {
	gchar *fields;
	GtkClipboard *clipboard;

	/* Get the selected fields into a string.	*/
	//str = get_selection_string(widget);
	fields = treeview_get_selected_fields();

	/* and copy it to the clipboard.	*/
	clipboard = gtk_clipboard_get (GDK_SELECTION_CLIPBOARD);
	//gtk_clipboard_set_text (clipboard, (const gchar *)str->b_dat, -1);
	gtk_clipboard_set_text (clipboard, fields, -1);

	//XXXstr = buf_free(str);
	g_free(fields);
}

static void treeview_drag_data_get_callback(GtkWidget *widget, GdkDragContext *dc, GtkSelectionData *selection_data, guint info, guint t, gpointer data) {
	// ABuf *fields;
	gchar *fields;

	// fields = get_selection_string(widget);
	fields = treeview_get_selected_fields();

	//gtk_selection_data_set_text(selection_data, (char *)fields->b_dat, -1);
	gtk_selection_data_set_text(selection_data, fields, -1);

	//fields = buf_free(fields);
	g_free(fields);
}

void treeview_init(void)
{
	GtkTreeViewColumn *column;
	GtkCellRenderer *renderer;
	GtkCellRenderer *pixbuf;
	GtkTreeSelection *selection;
	GtkWidget *treeView = gtkui_get_treeView(gtkapp);

	if (!treeView) {
		g_warning("treeview_init(): treeView is NULL\n");
		return;
	}

	load_pixmaps();

	/* Setup button presses */
	g_signal_connect(G_OBJECT(treeView), "button_press_event",
					 G_CALLBACK(on_tree_button_press_event), treeView);
	g_signal_connect(G_OBJECT(treeView), "popup-menu",
					 G_CALLBACK(on_tree_popup_menu), NULL);

	/* And misc. flags.  */
	gtk_tree_view_set_rules_hint(GTK_TREE_VIEW(treeView), TRUE);
	gtk_tree_view_set_reorderable(GTK_TREE_VIEW(treeView), TRUE);

	selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(treeView));
	if (selection)
		gtk_tree_selection_set_mode(selection, GTK_SELECTION_MULTIPLE);

	/* Add the columns */
	pixbuf = gtk_cell_renderer_pixbuf_new();
	renderer = gtk_cell_renderer_text_new();
	column = gtk_tree_view_column_new();
	gtk_tree_view_column_set_title(GTK_TREE_VIEW_COLUMN(column), "Field");
	gtk_tree_view_column_pack_start(GTK_TREE_VIEW_COLUMN(column),
									pixbuf, FALSE);
	gtk_tree_view_column_pack_start(GTK_TREE_VIEW_COLUMN(column),
									renderer, FALSE);
	gtk_tree_view_column_set_attributes(GTK_TREE_VIEW_COLUMN(column),
										pixbuf, "pixbuf", PIXBUF_COLUMN,
										NULL);
	gtk_tree_view_column_set_attributes(GTK_TREE_VIEW_COLUMN(column),
										renderer, "text", FIELD_COLUMN, NULL);
	gtk_tree_view_append_column(GTK_TREE_VIEW(treeView), column);

	renderer = gtk_cell_renderer_text_new();
	column = gtk_tree_view_column_new_with_attributes("Type", renderer,
													  "text",
													  TYPE_COLUMN, NULL);
	gtk_tree_view_append_column(GTK_TREE_VIEW(treeView), column);

	/* Setup drag/drop. NOTE: This needs to be done AFTER the columns are
	 * added or dragging doesn't work.	*/
	gtk_drag_source_set(GTK_WIDGET(treeView), GDK_BUTTON1_MASK, dragTargetsScintilla, DRAG_TARGET_COUNT, GDK_ACTION_COPY);
	g_signal_connect(G_OBJECT(treeView), "drag_data_get", G_CALLBACK(treeview_drag_data_get_callback), treeView);
}

/** Initialises table display for a new connection.
 * The list of tables is recalled and field displays created.	*/
void treeview_new_connection(void)
{
	const char *text;
	GtkTreeStore *model;
	DBConnection *conn;
	int i;
	int tables;
	GtkTreeIter dbiter;
	GtkTreeIter tableiter;
	DBTableDef *tbd = NULL;
	GtkWidget *treeView = gtkui_get_treeView(gtkapp);

	if (! treeView)
		return;

	/* If it already exists, just clear the data, otherwise make a new one.	*/
	model = GTK_TREE_STORE(gtk_tree_view_get_model(GTK_TREE_VIEW(treeView)));
	if (model)
		gtk_tree_store_clear(model);
	else {
		model = gtk_tree_store_new(3, GDK_TYPE_PIXBUF, G_TYPE_STRING, G_TYPE_STRING);
		gtk_tree_view_set_model(GTK_TREE_VIEW(treeView), GTK_TREE_MODEL(model));

		/* Need to release 'our' reference because we wont keep a reference to
		 * the model */
		g_object_unref(model);
	}

	conn = ui_get_connection();
	if (conn == NULL) {
		/* Don't warn on this. It is normal for disconnections.
		 * g_warning("treeview_new_connection(). conn is NULL.\n"); */
		return;
	}

	if (dbconnection_get_table_def(conn) < 0) {
		g_warning("Error retrieving table list.\n");
		return;
	}

	/* The top level item */
	gtk_tree_store_append(model, &dbiter, NULL);
	text = ui_get_connection_url(FALSE);
	if (! text)
		text = dbconnection_get_db_name(conn);

	if (! text)
		text = "";

	gtk_tree_store_set(model, &dbiter, PIXBUF_COLUMN, dbPixbuf,
					   FIELD_COLUMN, text, -1);

	tables = dbconnection_get_table_count(conn);
	for (i = 0; i < tables; i++) {
		tbd = (DBTableDef *)dbconnection_get_table(conn, i);
		if (! tbd) {
			g_warning("Could not retrieve table %d\n", i);
		}
		else {
			gtk_tree_store_append(model, &tableiter, &dbiter);
			gtk_tree_store_set(model, &tableiter, PIXBUF_COLUMN,
						   tablePixbuf, FIELD_COLUMN, tbd->name, -1);
			add_tree_fields(tbd, model, &tableiter);
		}
	}
}
