/** \file gui_gtk.c Common GUI functions.
 * \author Lionel ULMER
 * \author Darryl Luff
 * \addtogroup uigtk GTK UI module	*/
/* GtkSQL -- an interactive graphical query tool for PostgreSQL
 * Copyright (C) 1998-2005 Lionel ULMER, Darryl Luff
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "../config.h"
#include <stdlib.h>
#include <gtk/gtk.h>
#include <string.h>
#include <sys/types.h>
#include <time.h>

#include "browser.h"
#include "gtk_config.h"
#include "gtk_queries.h"
#include "gtkui-object.h"
#include "gui_gtk.h"
#include "test_gtk.h"
#include "utils.h"

/** Name of the rc file.	*/
static ABuf *rcname = NULL;

/** Non-zero when gtk main loop is running.	*/
int running = 0;

/** Get the name of the rc file to use.
 * \return The filename.	*/
char *get_rcname(void)
{
	rcname = buf_strcpy(rcname, getenv("HOME"));
	rcname = buf_strcat(rcname, "/.gtksqlrc");
	return (char *)rcname->b_dat;
}

/** Add a new query.	*/
static void gtkgui_add_query(void)
{
	query_add();
}

/** Delete all open queries.	*/
static void gtkgui_delete_all_queries(void)
{
	gint result;

	do {
		result = query_close();
	} while ((result != GTK_RESPONSE_CANCEL) && (query_count() > 0));
}

/** Disconnects the GUI's database connection.	*/
static void gtkgui_disconnect(void)
{
	if (!gtkapp)
		return;

	gtkui_disconnect(gtkapp);
}

/** Dump the gui-specific to stdout.	*/
static void gtkgui_dump_config(void)
{
	gtkcfg_dump_config();
}

/** Find the number of rows in the current result.
 * \return The number of rows.	*/
static int gtkgui_get_query_count(void)
{
	return query_count();
}

/** Find the number of columns in the current result.
 * \return The number of columns.	*/
static int gtkgui_get_result_columns()
{
	void *result;
	DBConnection *conn;

	/* Find the number of rows in the result. */
	conn = ui_get_connection();
	if (conn == NULL)
		return 0;

	result = gtksql_query_scintilla_get_result(query_get_current());
	if (result == NULL)
		return 0;

	if (dbconnection_get_query_status(conn, result) != DB_QUERY_RECORDS_OK)
		return 0;

	return dbconnection_get_field_count(conn, result);
}

/** Lookup a field text given a column and row.
 * \param row	The field row
 * \param column	The field column
 * \return	The field value as a string.	*/
static const char *gtkgui_get_result_field(int row, int column)
{
	DBConnection *conn = ui_get_connection();
	void *result = query_get_current_result();
	int nb_tuples, nb_cols;
	int stat;

	stat = dbconnection_get_query_status(conn, result);
	if (stat == DB_QUERY_RECORDS_OK) {
		nb_tuples = dbconnection_get_record_count(conn, result);
		nb_cols = dbconnection_get_field_count(conn, result);

		if ((row >= nb_tuples) || (column >= nb_cols))
			return "";

		return dbconnection_get_field_value(conn, result, row, column);
	}
	else
		return "";
}

/** Find the number of rows in the current result.
 * \return The number of rows.	*/
static int gtkgui_get_result_rows(void)
{
	void *result;
	DBConnection *conn;

	/* Find the number of rows in the result. */
	conn = ui_get_connection();
	if (conn == NULL)
		return 0;

	result = gtksql_query_scintilla_get_result(query_get_current());
	if (result == NULL)
		return 0;

	if (dbconnection_get_query_status(conn, result) != DB_QUERY_RECORDS_OK)
		return 0;

	return dbconnection_get_record_count(conn, result);
}

static const char *gtkgui_ui_name(void)
{
	return "Gtk";
}

/** Initialise the GUI.
 * \param argc	The argc parameter passed to main.
 * \param argv	The argv parameter passed to main.	*/
static void gtkgui_init(int argc, char **argv)
{
	gtk_init(&argc, &argv);
	gtkui_object_create();

#ifdef USE_TESTS
	gtk_tests_init();
#endif
}

/** Set the text for the currently active query.
 * \param txt	The text to set.	*/
static void gtkgui_set_query_text(const char *txt)
{
	gtksql_query_scintilla_set_text(query_get_current(), txt);
}

/** Set the currently active connection.
 * \param	con	The connection pointer to set.	*/
static void gtkgui_set_connection(DBConnection * connection)
{
	gtkui_set_connection(gtkapp, connection);
}

/** Set the name of the current query notebook page. If name is NULL or an empty string,
 * the call is ignored. Otherwise, the name is set as specified.
 * \param name	The new name.	*/
void gtkgui_set_query_name(const char *name)
{
	query_set_name(query_get_current(), (gchar *) name);
}

/** Load the current query from a file.
 * \param filename	The file to load.	*/
static void gtkgui_load_query_from_file(const char *filename)
{
	const char *name;
	ABuf *buf = buf_new(256);
	FILE *fp;
	GtksqlQueryScintilla *cq;

	cq = query_get_current();

	gtkgui_set_query_text("");

	fp = fopen(filename, "r");
	if (fp != NULL) {
		while (fgets((char *)buf->b_dat, buf->b_len, fp) != NULL) {
		}

		fclose(fp);
		name = get_short_filename(filename);
		if (name == NULL)
			name = "Untitled";

		gtkgui_set_query_name(name);
	}
	buf_free(buf);
}

/* Update the progress bar	*/
static void gtkgui_progress_reset(void *instance)
{
	gtksql_query_scintilla_progress_reset(GTKSQL_QUERY_SCINTILLA(instance));
}

/* Update the progress bar	*/
static void gtkgui_progress_update(void *instance, float percentage)
{
	gtksql_query_scintilla_progress_update(GTKSQL_QUERY_SCINTILLA(instance),
										   percentage);
}

/** Run the GUI.
 * \return The return code to be used on exit.	*/
static int gtkgui_run(void)
{
	running++;
	gtk_main();
	gtkcfg_clear();
	running--;
	return 0;
}

/** Save the gui-specific config to the file.	*/
static void gtkgui_save_config(void)
{
	gtkcfg_save_file();
}

/** Send the query to the database.	*/
static void gtkgui_send_query()
{
#define ERROREXEC_S	" Error executing query \n %s "
	DBConnection *conn;
	GtksqlQueryScintilla *currentQuery;

	currentQuery = query_get_current();
	if (!currentQuery) {
		g_warning("Couldnt access current query\n");
		return;
	}

	conn = ui_get_connection();
	gtksql_query_scintilla_send(currentQuery, conn);
}

/** Displays an error dialog.	*/
static void gtkgui_show_error(const char *title, const char *message)
{
	GtkWidget *dialog = gtkui_create_error_dialog(title, message);

	if (dialog)
		gtk_widget_show(dialog);
}

void gtkgui_show_table_treeview(gboolean show)
{
	gtkcfg_set_view_table_tree(show);
	update_ui();
}

/** Pop the last message off the status stack.	*/
static void gtkgui_status_pop(void *instance)
{
	gtksql_query_scintilla_status_pop(GTKSQL_QUERY_SCINTILLA(instance));
}

/** Allocate a GUI structure for the GTK gui.
 * \return \see GuiAPI.	*/
GuiAPI *gtkgui_register(void)
{
	GuiAPI *gtkapi;

	gtkapi = calloc(1, sizeof(GuiAPI));
	if (gtkapi) {
		gtkapi->add_query = gtkgui_add_query;
		gtkapi->config_value = gtkcfg_config_value;
		gtkapi->delete_all_queries = gtkgui_delete_all_queries;
		gtkapi->disconnect = gtkgui_disconnect;
		gtkapi->dump_config = gtkgui_dump_config;
		gtkapi->load_query_from_file = gtkgui_load_query_from_file;
		gtkapi->get_query_count = gtkgui_get_query_count;
		gtkapi->get_result_columns = gtkgui_get_result_columns;
		gtkapi->get_result_field = gtkgui_get_result_field;
		gtkapi->get_result_rows = gtkgui_get_result_rows;
		gtkapi->ui_name = gtkgui_ui_name;
		gtkapi->init = gtkgui_init;
		gtkapi->progress_reset = gtkgui_progress_reset;
		gtkapi->progress_update = gtkgui_progress_update;
		gtkapi->run = gtkgui_run;
		gtkapi->save_config = gtkgui_save_config;
		gtkapi->send_query = gtkgui_send_query;
		gtkapi->set_connection = gtkgui_set_connection;
		gtkapi->set_query_name = gtkgui_set_query_name;
		gtkapi->set_query_text = gtkgui_set_query_text;
		gtkapi->show_error = gtkgui_show_error;
		gtkapi->status_pop = gtkgui_status_pop;
	}
	else
		g_warning("gtkgui_register(): Could not allocate memory\n");

	return gtkapi;
}

/** Update the displayed widgets to match the current state of the application.	*/
void update_ui(void)
{
	gtkui_update(gtkapp, FALSE);
}
