/** \file guiapi.c	Interface to menu module.
 * \author Lionel ULMER
 * \author Darryl Luff
 * \addtogroup ui	*/
/* GtkSQL -- an interactive graphical query tool for PostgreSQL
 * Copyright (C) 2002-2004 Darryl Luff
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "../config.h"
#include "gtk_support.h"
#include "gui_gtk.h"

/** The global GUI instance.	*/
static GuiAPI *ui = NULL;

/** Add a new blank query.	*/
void ui_add_query(void)
{
	if (!ui)
		return;

	ui->add_query();
}

/** Dump config to stdout.	*/
void ui_dump_config(void)
{
	if (!ui) {
		g_warning(_("ui_dump_config(): ui not initialised\n"));
		return;
	}

	if (ui->dump_config)
		ui->dump_config();
	else
		g_warning
			("ui_dump_config(): ui->dump_config() not initialised. ui = %p\n",
			 ui);
}

/** Get the current connection.
 * \return The connection pointer.	*/
DBConnection *ui_get_connection(void)
{
	if (!ui) {
		fprintf(stderr, "ui_get_connection(). ui not initialised\n");
		return NULL;
	}

	return ui->connection;
}

/** Update a configuration value.	*/
void ui_config_value(const char *key, const char *val)
{
	if (!ui) {
		fprintf(stderr, "ui_config_value(%s, %s). ui not initialised.\n", key,
				val);
		return;
	}

	if (ui->config_value != NULL)
		ui->config_value(key, val);
	else
		fprintf(stderr,
				"ui_config_value(%s, %s). ui->config_value not initialised.\n",
				key, val);
}

/** Returns true if the connection is good.
 * \return 1 if the connection is good, 0 otherwise.	*/
int ui_connection_ok(void)
{
	if (ui_get_connection() != NULL)
		return 1;
	else
		return 0;
}

/** Delete all open queries.	*/
void ui_delete_all_queries(void)
{
	if (!ui)
		return;

	if (ui->delete_all_queries)
		ui->delete_all_queries();
}

/** Disconnects the GUI's database connection.	*/
void ui_disconnect(void)
{
	if (!ui) {
		g_warning("ui_disconnect(): ui not initialised\n");
		return;
	}

	if (ui_get_connection() == NULL)
		return;

	if (ui->disconnect)
		ui->disconnect();
}

/** Get the URL for the current connection.
 * \return The url string. Don't free this.	*/
const char *ui_get_connection_url(int showPwd)
{
	const DBConnector *connector;
	const DBConnection *conn = ui_get_connection();
	if (conn == NULL)
		return NULL;

	connector = dbconnection_get_connector(conn);
	if (connector == NULL)
		return NULL;

	return dbconnector_form_to_url(connector, showPwd);
}

/** Find the number of rows in the current result.
 * \return The number of rows.	*/
int ui_get_query_count(void)
{
	if (!ui)
		return 0;

	if (ui->get_query_count)
		return ui->get_query_count();

	return 0;
}

/** Find the number of columns in the current result.
 * \return The number of columns.	*/
int ui_get_result_columns(void)
{
	if ((ui) && (ui->get_result_columns))
		return ui->get_result_columns();

	return 0;
}

/** Lookup a field text given a column and row.
 * \param row	The field row
 * \param column	The field column
 * \return	The field value as a string.	*/
const char *ui_get_result_field(int row, int column)
{
	if ((!ui) || (!ui->get_result_field))
		return ui->get_result_field(row, column);

	return NULL;
}

/** Find the number of rows in the current result.
 * \return The number of rows.	*/
int ui_get_result_rows(void)
{
	if ((!ui) && (!ui->get_result_rows))
		return ui->get_result_rows();

	return 0;
}

/** Initialise the GUI.
 * \param argc	The argc parameter passed to main.
 * \param argv	The argv parameter passed to main.	*/
void ui_init(int argc, char **argv)
{
	if (ui == NULL)
		ui = gtkgui_register();

	if (ui == NULL) {
		fprintf(stderr, "WARNING: ui_init(). GUI not initialised\n");
		return;
	}

	if (ui->init)
		ui->init(argc, argv);
}

/** Load the current query from a file.
 * \param filename	The file to load.	*/
void ui_load_query_from_file(const char *filename)
{
	if (!ui)
		return;

	if (ui->load_query_from_file)
		ui->load_query_from_file(filename);
}

/** Exit the GUI.	*/
void ui_quit(void)
{
	if (running != 0)
		gtk_main_quit();
}

/** Run the GUI.
 * \return The return code to be used on exit.	*/
void ui_progress_reset(void *instance)
{
	if (!ui)
		return;

	if (!ui->progress_reset)
		return;

	ui->progress_reset(instance);
}

/* Update the progress bar	*/
void ui_progress_update(void *instance, float percentage)
{
	if (!ui)
		return;

	if (!ui->progress_update)
		return;

	ui->progress_update(instance, percentage);
}

/** Run the GUI.
 * \return The return code to be used on exit.	*/
int ui_run(void)
{
	if (!ui)
		return 1;

	if (!ui->run)
		return 1;

	return ui->run();
}

/** Save the ui-specific config to the file.	*/
void ui_save_config(void)
{
	if (!ui) {
		g_warning(_("ui_save_config(): ui not initialised\n"));
		return;
	}

	if (ui->save_config)
		ui->save_config();
	else
		g_warning
			("ui_save_config(): ui->save_config() not initialised. ui = %p\n",
			 ui);
}

/** Send the query to the database.	*/
void ui_send_query(void)
{
	if (!ui) {
		g_warning(_("ui_send_query(): ui not initialised\n"));
		return;
	}

	if (ui->send_query)
		ui->send_query();
	else
		g_warning(_
				  ("ui_send_query(): ui->send_query() not initialised. ui = %p\n"),
				  ui);
}

/** Set the currently active connection.
 * \param	con	The connection pointer to set.	*/
void ui_set_connection(DBConnection * con)
{
	if (ui == NULL) {
		fprintf(stderr,
				"WARNING: ui_set_connection(). GUI not initialised\n");
		return;
	}

	ui->connection = con;
	if (ui->set_connection)
		ui->set_connection(con);
	else
		fprintf(stderr, "WARNING: ui->set_connection not defined\n");
}

/** Set the name of the current query notebook page. If name is NULL or an empty string,
 * the call is ignored. Otherwise, the name is set as specified.
 * \param name	The new name.	*/
void ui_set_query_name(const char *name)
{
	if (!ui)
		return;

	if (ui->set_query_name)
		ui->set_query_name(name);
}

/** Set the text for the currently active query.
 * \param txt	The text to set.	*/
void ui_set_query_text(const char *txt)
{
	if (!ui) {
		fprintf(stderr, "WARNING: ui is NULL\n");
		return;
	}

	if (ui->set_query_text)
		ui->set_query_text(txt);
	else
		fprintf(stderr, "WARNING: ui->set_query_text is NULL\n");
}

/** Set the text for the currently active query.
 * \param title	The title text to be displayed.
 * \param msg	The error message.	*/
void ui_show_error(const char *title, const char *msg)
{
	const char *t, *m;

	t = title;
	if (! t)
		t = "Error";

	m = msg;
	if (! m)
		m = "";

	if (ui && ui->show_error)
		ui->show_error(title, msg);
	else
		fprintf(stderr, "%s: %s\n", t, m);
}

/** Pop the last message off the status stack.	*/
void ui_status_pop(void *instance)
{
	if (!ui)
		return;

	if (ui->status_pop)
		ui->status_pop(instance);
}

/** Push the message onto the status stack.
 * \param msg	The text to set.	*/
void ui_status_push(const char *msg)
{
	if (!ui)
		return;

	if (ui->status_push)
		ui->status_push(msg);
}
