/** \addtogroup ui
 * @{
 */
/** \file guiapi.h	Interface to menu module.
 * \author Lionel ULMER
 * \author Darryl Luff	*/
/* GtkSQL -- an interactive graphical query tool for PostgreSQL
 * Copyright (C) 2002-2004 Darryl Luff
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifndef __GUIAPI_H
#define __GUIAPI_H

#include <stdio.h>
#include "dbapi.h"

/** Error callback.	*/
typedef void (*ErrorCallback) (char *, char *);

/** This structure defines the GUI interface.	*/
struct _uiapi {
	/** Get the name of this GUI	*/
	const char *(*ui_name) (void);

	/** The active GUI connection.	*/
	DBConnection *connection;

	/** Add a new blank query.	*/
	void (*add_query) (void);

	/** Update a configuration value.	*/
	void (*config_value) (const char *, const char *);

	/** Callback for error notification.	*/
	void (*show_error) (const char *, const char *);

	/** Delete all open queries.	*/
	void (*delete_all_queries) (void);

	/** Disconnects the GUI's database connection.	*/
	void (*disconnect) (void);

	/** Find the number of rows in the current result.	*/
	int (*get_query_count) (void);

	/** Find the number of columns in the current result.	*/
	int (*get_result_columns) (void);

	/** Lookup a field text given a column and row.	*/
	const char *(*get_result_field) (int row, int column);

	/** Load the current query from a file.	*/
	void (*load_query_from_file) (const char *filename);

	/** Find the number of rows in the current result.	*/
	int (*get_result_rows) (void);

	/** Initialise the GUI.	*/
	void (*init) (int argc, char **argv);

	/** Reset the progress display.	*/
	void (*progress_reset) (void *instance);

	/** Update the progress display.	*/
	void (*progress_update) (void *instance, float percentage);

	/** Run the GUI.	*/
	int (*run) (void);

	/** Dump ui config to stdout.	*/
	void (*dump_config) (void);

	/** Save ui config to a file.	*/
	void (*save_config) (void);

	/** Send the query to the database.	*/
	void (*send_query) ();

	/** Set the currently active connection.	*/
	void (*set_connection) (DBConnection * con);

	/** Set the name of the current query notebook page.	*/
	void (*set_query_name) (const char *name);

	/** Set the text for the currently active query.	*/
	void (*set_query_text) (const char *txt);

	/** Push the sext onto the status display.	*/
	void (*status_push) (const char *txt);

	/** Pop the last status off the status stack.	*/
	void (*status_pop) (void *instance);
};

typedef struct _uiapi GuiAPI;

void ui_add_query(void);
void ui_config_value(const char *key, const char *val);
int ui_connection_ok(void);
void ui_delete_all_queries(void);
void ui_disconnect(void);
void ui_dump_config(void);
DBConnection *ui_get_connection(void);
const char *ui_get_connection_url(int showPwd);
int ui_get_query_count(void);
int ui_get_result_columns(void);
const char *ui_get_result_field(int row, int column);
int ui_get_result_rows(void);
void ui_init(int argc, char **argv);
void ui_load_query_from_file(const char *filename);
void ui_progress_reset(void *instance);
void ui_progress_update(void *instance, float percentage);
int ui_run(void);
void ui_quit(void);
void ui_save_config(void);
void ui_send_query(void);
void ui_set_connection(DBConnection * con);
void ui_set_query_name(const char *name);
void ui_set_query_text(const char *txt);
void ui_show_error(const char *title, const char *msg);
void ui_status_pop(void *instance);
void ui_status_push(const char *msg);

/*@}*/
#endif
