/** \mainpage GtkSQL
 * This is the source documentation for the GtkSQL program. GtkSQL is a query
 * interface to a number of different databases. This version includes drivers
 * for PostgreSQL, MySQL and SQLite, and experimental support for querying XML 
 * files using XPath. The XPath driver is not complete., and will not be 
 * included in a build unless '--enable-xpath' is specified to ./configure.
 *
 * \todo File open dialog 'Cancel' button doesn't work.
 *
 * \section help Suggestions
 * If you have any suggestions for the development of GtkSQL, please let me 
 * know. There are a lot of SQL query interfaces out there. I'd like to make
 * GtkSQL a bit different. My immediate plans are:
 * - Implement XPath queries of XML files (partially done - see d_xpath_f.c)
 * - Implement a forms interface so that GtkSQL can be called to display a 
 *   form to the user, take some input, and run a query. So it can be used as
 *   a generic database application for small utilities. (forms interface is done.
 *   can't be scripted yet).
 * - Implement scripting (partially done - see luaif.c).
 *
 * Other things that I initially thought would be good, but then wasn't so
 * sure, are:
 * - Ability to generate code fragments to access the connected databases. So 
 *   you could connect to a database, then generate the C, C++, Perl etc code
 *   needed to access that database.
 *
 * \section general General Guidelines
 * Where a structure is defined, I usually provide access functions for the 
 * structure members and use those instead of accessing the members directly. 
 * This makes it easier to see where things are referenced from, and cal also
 * automatically take care of null-pointers etc.
 *
 * I also try to keep functions small enough to fit into a single screen where
 * it makes sense to do so. And I hate having duplicate code in different 
 * sections. So I tend to break things up fairly small. There are many places 
 * where this isn't done, and there are still some pretty big functions. That 
 * probably just means that I haven't had to deal with that part of the code
 * yet!
 *
 * \section standard Coding Standards
 * I like consistancy in code formatting. And you need it to make things easy 
 * to find - especially in code maintained by multiple people! So I've 
 * modified (I think) all the code to conform to these naming standards:
 * - Function names use the GTK type syntax with all lower case, and separate 
 *   words in the name seperated with underscores. For example a function to
 *   set the configuration filename would be called config_set_filename().
 * - Variables are named using 'camel-case' with an initial lower-case letter.
 *   For example a variable holding the configuration filename would be called
 *   configFilename.
 * - Typedefs are named the same as variables, except that the first character
 *   is upper-case. So a typedef for a file format would be called FileFormat.
 * 
 * I also run the code through 'indent' whenever I make changes, using the 
 * format string:
 * \code
 * -br -brs -c40 -cdw -fca -i4 -nbc -nce -npcs -npsl -sc -ss -ts4
 * \endcode
 *
 * If you're making any changes, please try and follow the naming standard
 * so things are easy to find. If you submit them in another format, they'll
 * eventually magically morph into the above format anyway!
 *
 * If you really don't like these formats, please suggest a better one. I don't
 * really care what the standard format is, as long as all of the code is 
 * consistant with it.
 *
 * \section docs Documentation
 * The source is documented for Doxygen. All functions should have basic documentation
 * in the .h file, and full docs in the .c file.
 *
 * \section modules Modules
 * I've partitioned the source into modules and try and reduce the number of calls
 * between the modules. There are two main modules "GUI" and "DB". The interfaces to
 * these modules are defined in the files "guiapi.h" and "dbapi.h". Code from 
 * one module should only call functions in other modules that are defined in 
 * these API files. There are also modules for scripting and configuration, 
 * which haven't properly modularised yet. Ideally, I would like it to be just as easy to
 * write a new UI or script driver as it currently is to write a database driver.
 * So writing a Qt interface or implementing perl scripting should just be a 
 * matter of implementing a driver.
 *
 * The program is divided into these main modules:
 * - The DB module, which handles the database drivers. This module provides an API (defined by dbapi.h) to
 * other modules.
 * - The GUI module, which handles the user interface. This module provides an API
 * (defined by guiapi.h) to other modules.
 * - The scripting module, which controls script engines.
 * - The config module, which handles the application configuration, and currently doesn't do much at all.
 *
 * All calls between modules are hopefully calls declared in these interface files.
 *
 * \section drivers Drivers
 * To add a new driver:
 * - Select a short tag for your driver. The tag is simply a short name ("Mysql" etc).
 * In the following '<tag>' represents the tag in all lowercase, "<Tag>" is the tag
 * in camel-case, "<TAG>" is the tag in uppercase.
 * - Pick one of the existing drivers that is the most like yours.
 * - Copy the sample .c file to "d_<tag>.c".
 * - Copy the sample .h file to "d_<tag>.h".
 * - Update and complete the documentation at the top of the file. Format isn't
 * really important, but you should specify the URL syntax for accessing your 
 * driver, and describe the fields it uses in the connect dialog.
 * - Do these search/replaces on both d_<tag>.c and d_<tag>.h:
 *  - Replace the old <Tag> with your new <Tag>
 *  - Replace the old <tag> with your new <tag>
 *  - Replace the old <TAG> with your new <TAG>
 * - Note the line "#ifdef USE_<TAG>". This ensures that your driver code is only included if the 
 * appropriate configure directive is given. Note that even if your driver is not included,
 * an empty module is still compiled and linked. This is because I am lazy and haven't changed it!
 * It also means other modules don't have to be changed when a new driver is added.
 * - In d_<tag>.c, look at the structure DBConnection_<Tag>. This structure should hold any 
 * connection-specific data you need to keep. Nothing is required in this structure. Only your driver uses it.
 * - In d_<Tag>.c, look at the <tag>Keywords structure. This structure needs to hold any words that need
 * keyword-completion. Normally, this will be the query keywords (SELECT etc) that your driver supports.
 * - Implement the functions defined in the db_conn structure in dbapi.h. You must
 * provide a definition for each of these functions. See the documentation in that file for details.
 * - Implement the register_<Tag> function. This registers your driver with the application, and defines the 
 * connection form presented to the user.
 * - Open register.c. In the includes, add "#include "d_<tag>.h".
 * - Still in register.c, find the "db_register_func registerDB[] = {" structure. Add this at the end of this structure,
 * just before the terminating NULL:
 *
 * #ifdef USE_<TAG>
 *	register_<Tag>,
 * #endif
 *
 * - Update the autoconf files to build your application. This is the most painful part. The basic idea is
 * that there is a "use_<tag>" variable for each driver. If that variable is set to "0", the driver will not be included. 
 * If it's set to '1', the driver will be included. Stable drivers should have this value set to '1' by default, and 
 * unstable drivers should have it set to '0' by default.
 * The selection of which drivers to include is normally automatic (for the drivers enabled by default). Or it can
 * be overridden by ./configure arguments.
 *
 * NOTE: I use Anjuta to manage these files (www.anjuta.org). If you use it, please
 * use the Anjuta options under "Project/Configure Project" to set this up. If you don't use Anjuta, 
 * just do it normally in configure.in and I'll update the Anjuta config later.
 * You will have to do the following:
 *  - In configure.in ("Programs" tab in Anjuta), look for the line "use_mysql=1". This says to include the mysql
 * driver by default. Add a new line that says "use_<tag>=1"
 *  Look for the section that starts "AC_ARG_WITH(mysql,      [  --with-mysql            MySQL support], \". Copy this section 
 * to a new section and change the mysql references to reflect your driver.
 *  - Look for the section that says:
 *
 * if test "$use_lua" = "1" ; then
 * 	AC_CHECK_LIB(lua, lua_open, , use_lua=)
 * fi
 *
 * This looks for the lua library, and if not found it turns lua support off. Add a section the same for your driver.
 *
 * - Add any AC_CHECK_HEADERS sections to look for any header files you need. Again, set the use_<tag> 
 * variable to '0' if you don't find something you need.
 *
 * - Near the end of configure.in, look for the section starting "AC_MSG_CHECKING(whether to include MySQL support)".
 * Copy this section for your driver. Ensure that it adds the "AC_DEFINE([USE_<TAG>],1,[Enable <Tag> support])"
 * section. This actually defines the variable that ends up in config.h, that your driver checks to see if it's
 * enabled.
 *
 */
/** \file main.c	GtkSQL -- an interactive SQL query tool.
 * \author Lionel ULMER
 * \author Darryl Luff
 */
/* Copyright (C) 1998-2004  Lionel ULMER, Darryl Luff.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if notload, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include "appconfig.h"
#include "browser.h"
#include "guiapi.h"
#include "scriptapi.h"
#include "utils.h"

int main(int argc, char *argv[])
{
	/* Setup environment */
#ifdef ENABLE_NLS
	setlocale(LC_ALL, "");
	bindtextdomain(GETTEXT_PACKAGE, PACKAGE_LOCALE_DIR);
	textdomain(GETTEXT_PACKAGE);
#endif

	/* Initialise configuration */
	cfg_init();

	/* Setup database drivers */
	db_init();

	/* Setup scripting */
	script_init();

	/* Setup user interfaces */
	ui_init(argc, argv);

	/* Load configuration */
	cfg_parse_cmdline(argc, argv);
	cfg_load_file();

	/* browser setup	*/
	browse_init();
	browse_refresh();

	/* Run the program */
	ui_run();

	return 0;
}
