/** \file plugins.c	Plugin loader.
 * \author Darryl Luff
 * \addtogroup plugins	*/
/* GtkSQL -- an interactive graphical query tool for PostgreSQL
 * Copyright (C) 2002-2007 Darryl Luff
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "../config.h"

#ifdef USE_PILS

#include <dlfcn.h>
#include <sys/dir.h>
#include <sys/stat.h>
#include <stdio.h>
#include <string.h>
#include <pils/generic.h>
#include <pils/plugin.h>

#include "buf.h"
#include "dbapi.h"
#include "guiapi.h"
#include "register.h"

// Temporary hardcoded plugindir
#define DEFAULTPLUGINDIR	"/usr/local/gtksql"

PILPluginUniv *PluginLoadingSystem = NULL;
GHashTable *DBAPIFuncs = NULL;
char result[64];
struct DBAPIOps *dbapiops;
PILGenericIfMgmtRqst RegisterRqsts[] = {
	{ "DBAPI", &DBAPIFuncs, NULL, NULL, NULL},
	{ NULL, NULL, NULL, NULL, NULL},
};

struct DBAPIImports dbapi_imports = {
	ui_show_error	// display_error
};

char *errorstr(PIL_rc err) {
		switch (err) {
			case PIL_OK:
				return "Success"; break;
			case PIL_INVAL:
				return "Invalid Parameters"; break;
			case PIL_BADTYPE:
				return "Bad plugin/interface type"; break;
			case PIL_EXIST:
				return "Duplicate Plugin/Interface name"; break;
			case PIL_OOPS:
				return "Internal Error"; break;
			case PIL_NOPLUGIN:
				return "No such plugin or Interface"; break;
			default:
				return "Unknown error code";
		}
}

void pilplugins_init(char *path) {
	PIL_rc err;

	printf("Initialising pils plugins\n");
	PluginLoadingSystem = NewPILPluginUniv(path);
	err = PILLoadPlugin(PluginLoadingSystem, "InterfaceMgr", "generic", &RegisterRqsts);
	if (err != PIL_OK) {
		printf("Load 'generic' returned %d: %s\n", err, errorstr(err));
		return;
	}

	err = PILLoadPlugin(PluginLoadingSystem, "DBAPI", "xpath", NULL);
	if (err != PIL_OK) {
		printf("Load 'xpath' returned %d: %s\n", err, errorstr(err));
		return;
	}

	if (! DBAPIFuncs) {
		printf("DBAPIFuncs = NULL\n");
		return;
	}

	dbapiops = g_hash_table_lookup(DBAPIFuncs, "xpath");
	if (! dbapiops) {
		printf("dbapiops = NULL\n");
		return;
	}

	if (dbapiops->set_imports)
		dbapiops->set_imports(&dbapi_imports);
	else
		fprintf(stderr, "dbapiops->set_imports is NULL\n");
}

#ifdef XXXOLD

void check_plugin(char *fn) {
	void *driver;
	void *func;

	if (! fn)
		return;

	printf("[Loader] Trying to load %s\n", fn);
	driver = dlopen(fn, RTLD_LAZY);
	if (! driver) {
		fprintf(stderr, "[Loader] Could not open %s: %s\n", fn, dlerror());
		return;
	}
	dlerror();
	func = dlsym(driver, "register_DBAPI");
	if (func) {
		registered_drivers_add(func);
		printf("Added DBAPI driver\n");
	}
	else {
		printf("Not a DBAPI driver: %s\n", dlerror());
		dlclose(driver);
	}
}

void plugins_load(char *fn) {
	DIR *dp;
	struct direct *dir;
	char *c;
	unsigned int namelen;
	ABuf *buf = NULL;
	char *f;

	if (fn)
		f = fn;
	else
		f = DEFAULTPLUGINDIR;

	// Check directory
	if ((dp = opendir(f)) == NULL) {
		fprintf(stderr, "cannot open plugin directory %s.\n", f);
		return;
	}

	/* Read entries	*/
	while ((dir = readdir(dp)) != NULL) {
		/* Skip removed files	*/
		if ((dir->d_ino == 0) || (dir->d_type != 0))
			continue;

		printf("%s\n", dir->d_name);
		/* For each file, check for signature functions	*/
		namelen = strlen(dir->d_name);
		if (namelen > 3) {
			c = dir->d_name + namelen - 3;
			if (strcmp(c, ".so") == 0) {
				/* looks like a plugin.	*/
				buf = buf_check(buf, 1000);
				buf = buf_strcpy(buf, f);
				buf = buf_strcat(buf, "/");
				buf = buf_strcat(buf, dir->d_name);
				check_plugin((char *)buf->b_dat);
			}
		}
	}
	buf = buf_free(buf);
	closedir(dp);
}
#endif

/* endif for ifdef USE_PILS */
#endif 
