/** \file register.c	Plugin registry. This module maintains the lists
 * of available drivers and provides functions to manipulate them. 
 * In the long run, it will hopefully be replaced by a dynamic loading 
 * mechanism for db, UI and script drivers.
 * \author Lionel ULMER
 * \author Darryl Luff
 * \addtogroup register	*/
/* $Id: register.c,v 1.8 2005/12/31 00:03:04 djluff Exp $
 * GtkSQL -- an interactive graphical query tool for PostgreSQL
 * Copyright (C) 1998-2005 Lionel ULMER, Darryl Luff
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "../config.h"
#include <stdlib.h>

#include "d_mysql_f.h"
#include "d_postgres_f.h"
#include "d_unixodbc.h"
#include "d_sqlite.h"

typedef struct _adriver {
	db_register_func registerDB;
	struct _adriver *next;
} ADriver;

ADriver *drivers = NULL;

/** Holds references to the registration functions for the supported
 * databases.	*/
db_register_func XXXregisterDB[] = {
#ifdef USE_POSTGRESQL
	/** Registers the PostgreSQL driver.	*/
	register_PostgreSQL,
#endif
#ifdef USE_MYSQL
	/** Registers the MySQL driver.	*/
	register_MySQL,
#endif
#ifdef USE_UNIXODBC
	/** Registers the unixODBC driver.	*/
	register_unixODBC,
#endif
#ifdef USE_SQLITE
	/** Registers the SQLite driver.	*/
	register_SQLite,
#endif
	/* Make sure this NULL stays on the end.	*/
	NULL
};

int registered_drivers_count(void)
{
	int nb_db = 0;
	ADriver *d;

	if (! drivers)
		return 0;

	d = drivers;
	nb_db = 1;
	while (d->next) {
		d = d->next;
		nb_db++;
	}

	return nb_db;
}

void registered_drivers_add(db_register_func driver) {
	ADriver *drv;
	ADriver *d;

	if (! driver)
		return;

	drv = calloc(1, sizeof(ADriver));
	if (! drv)
		return;

	drv->registerDB = driver;
	if (! drivers) {
		drivers = drv;
		return;
	}

	d = drivers;
	while (d->next)
		d = d->next;

	d->next = drv;
}

void registered_drivers_init_static(void) {
#ifdef USE_POSTGRESQL
	registered_drivers_add(register_PostgreSQL);
#endif
#ifdef USE_MYSQL
	registered_drivers_add(register_MySQL);
#endif
#ifdef USE_UNIXODBC
	registered_drivers_add(register_unixODBC);
#endif
#ifdef USE_SQLITE
	registered_drivers_add(register_SQLite);
#endif
}

db_register_func registered_drivers_get_first(void) {
	if (! drivers)
		return NULL;

	return drivers->registerDB;
}

static ADriver *registered_drivers_find(db_register_func last) {
	ADriver *d;
	if (! drivers)
		return NULL;

	d = drivers;
	while ((d->registerDB != last) && (d->next))
		d = d->next;

	if (d->registerDB == last)
		return d;

	return NULL;	
}

db_register_func registered_drivers_get_next(db_register_func last) {
	ADriver *d;
	d = registered_drivers_find(last);
	if (! d)
		return NULL;

	d = d->next;
	if (! d)
		return NULL;

	return d->registerDB;
}
