/** \file test_gtk.c Common GUI functions.
 * \author Darryl Luff
 * \addtogroup test Test module	*/
/* GtkSQL -- an interactive graphical query tool for PostgreSQL
 * Copyright (C) 2004-2005 Darryl Luff
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "../config.h"

#ifdef USE_TESTS
#include <gtk/gtk.h>

#ifndef create_testDialog
#include "gtk_interface.h"
#endif
#include "gtk_support.h"

#include "gtkui-object.h"
#include "test_gtk.h"

GtkWidget *testlog = NULL;
GtkWidget *testtext = NULL;

static void show_testlog(void) {
	if (! testlog)
		testlog = create_testDialog();

	if (! testlog)
		return;

	if (! testtext)
		testtext = lookup_widget(testlog, "testText");

	gtk_widget_show(testlog);
}

static void log_message(const char *msg1, const char *msg2, const char *msg3) {
	GtkTextBuffer *buffer;
	ABuf *buf = buf_new(1024);

	if (! buf)
		return;

	buf = buf_strcpy(buf, "");
	buf = buf_strcat(buf, msg1);
	buf = buf_strcat(buf, msg2);
	buf = buf_strcat(buf, msg3);

	if (! testtext)
		return;

	buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(testtext));
	if (! buffer)
		return;

	gtk_text_buffer_insert_at_cursor(buffer, buf->b_dat, -1);
	buf_free(buf);
}

/** Tests that all widgets are in their correct state.	*/
static int gtk_test_activations(void) {
	GtkuiObject *ui;
	int errors;

	log_message("\nTesting Activations...\n", NULL, NULL);
	ui = gtkui_object_get();
	errors = gtkui_update(ui, TRUE);
	if (errors < 0)	/* Some internal error	*/
		return 1;

	return errors;
}

static int widget_visible(GtkWidget *widget) {
	return (GTK_WIDGET_FLAGS(widget) & GTK_VISIBLE) >> 8;
}

static int test_toggle_view(GtkCheckMenuItem *item, GtkWidget *widget, const gchar *msg) {
	int oldstate;
	int errors = 0;

	log_message(". Testing view toggle for ", msg, "\n");
	if (! item) {
		errors++;
		log_message(". FAIL: ", msg, " menu item is NULL\n\n");
		return errors;
	}

	if (! widget) {
		errors++;
		log_message(". FAIL: ", msg, " widget is NULL\n\n");
		return errors;
	}

	log_message(". Testing correct visible state...\n", NULL, NULL);
	oldstate = gtk_check_menu_item_get_active(item);
	if (oldstate != widget_visible(widget)) {
		errors++;
		log_message(". FAIL: ", msg, " visible state is wrong\n\n");
		return errors;
	}

	/* Now, toggle the state and make sure the state changes!	*/
	log_message(". Testing state change...\n", NULL, NULL);
	gtk_menu_item_activate(GTK_MENU_ITEM(item));
	if (oldstate == gtk_check_menu_item_get_active(GTK_CHECK_MENU_ITEM(item))) {
		errors++;
		log_message(". FAIL: ", msg, " state didn't change when it should have\n\n");
		return errors;
	}

	if (oldstate == widget_visible(widget)) {
		errors++;
		log_message(". FAIL: ", msg, " visible state is wrong after change\n\n");
		return errors;
	}

	/* Now, set the state back to the original	*/
	log_message(". Restoring to original state...\n", NULL, NULL);
	oldstate = gtk_check_menu_item_get_active(item);
	gtk_menu_item_activate(GTK_MENU_ITEM(item));
	if (oldstate == gtk_check_menu_item_get_active(item)) {
		errors++;
		log_message("FAIL: ", msg, " state didn't change (2)\n\n");
		return errors;
	}

	if (oldstate == widget_visible(widget)) {
		errors++;
		log_message(". FAIL: ", msg, " visible state is wrong after change (2)\n\n");
		return errors;
	}

	/* Now back to the start state.	*/
	return errors;
}

static int test_toolbar_view(void) {
	GtkCheckMenuItem *menuitem;
	GtkWidget *toolbar;
	int errors = 0;

	menuitem = GTK_CHECK_MENU_ITEM(gtkui_get_toolbar_menuitem(gtkapp));
	toolbar = gtkui_get_toolbar(gtkapp);
	errors = test_toggle_view(menuitem, toolbar, "Toolbar");

	return errors;
}

static int test_treeview(void) {
	GtkCheckMenuItem *menuitem;
	GtkWidget *toolbar;
	int errors = 0;

	menuitem = GTK_CHECK_MENU_ITEM(gtkui_get_treeview_menuitem(gtkapp));
	toolbar = gtkui_get_treeview(gtkapp);
	errors = test_toggle_view(menuitem, toolbar, "TreeView");

	return errors;
}

static int test_views(void) {
	log_message("\nTesting views...", NULL, NULL);
	return test_toolbar_view() + test_treeview();
}

static void gtk_tests_run(void) {
	show_testlog();
	log_message("\nStarting new test ===\n", NULL, NULL);

	if (gtk_test_activations() == 0)
		log_message("Activations: OK\n", NULL, NULL);
	else
		log_message("Activations: Fail\n", NULL, NULL);

	if (test_views() == 0)
		log_message("Views: OK\n", NULL, NULL);
	else
		log_message("Views: Fail\n", NULL, NULL);
}

static void on_selftest_activate(GtkMenuItem * menuitem, gpointer user_data) {
	gtk_tests_run();
}

void gtk_tests_init(void) {
	/* Add the test option to the test menu	*/
	GtkWidget *item;

	item = gtk_separator_menu_item_new ();
	gtk_widget_set_sensitive (item, FALSE);
	gtkui_append_filemenu_item(gtkapp, item);

	item = gtk_menu_item_new_with_label("Self test");
	g_signal_connect ((gpointer) item, "activate", G_CALLBACK (on_selftest_activate), NULL);
	gtkui_append_filemenu_item(gtkapp, item);
}

#endif
